Package: BLSloadR
Type: Package
Title: Download Time Series Data from the U.S. Bureau of Labor
        Statistics
Version: 0.4.5
Authors@R: c(
    person(
    given = "Nevada Department of Employment, Training, and Rehabilitation",
    role = "cph"
    ),
    person(
      given = "David", family = "Schmidt",
      email = "deschmidt@detr.nv.gov",
      role = c("aut", "cre")
    ),
    person(
      given = "Mark", family = "Rembert",
      email = "Mark.Rembert@mass.gov",
      role = c("aut")
    )
  )
Description: These functions provide a convenient interface for downloading data from the U.S. Bureau of Labor Statistics <https://www.bls.gov>.  The functions in this package utilize flat files produced by the Bureau of Labor Statistics, which contain full series history.  These files include employment, unemployment, wages, prices, industry and occupational data at a national, state, and sub-state level, depending on the series.  Individual functions are included for those programs which have data available at the state level.  The core functions provide direct access to the Current Employment Statistics (CES) <https://www.bls.gov/ces/>, Local Area Unemployment Statistics (LAUS) <https://www.bls.gov/lau/>, Occupational Employment and Wage Statistics (OEWS) <https://www.bls.gov/oes/> and Alternative Measures of Labor Underutilization (SALT) <https://www.bls.gov/lau/stalt.htm> data produced by the Bureau of Labor Statistics.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: data.table (>= 1.16), httr (>= 1.4.7), dplyr (>= 1.1), stringr
        (>= 1.5), lubridate (>= 1.9), rvest (>= 1.0.4), readxl (>=
        1.4.5), sf (>= 1.0), tidyselect (>= 1.2), tigris (>= 2.0), zoo
        (>= 1.8), htmltools, rstudioapi
RoxygenNote: 7.3.3
Suggests: ggplot2, knitr, mapgl, rmarkdown, tidyr, gt
VignetteBuilder: knitr
URL: https://schmidtdetr.github.io/BLSloadR/
Config/Needs/website: rmarkdown
LazyData: true
NeedsCompilation: no
Packaged: 2026-04-22 16:44:19 UTC; deschmidt
Author: Nevada Department of Employment, Training, and Rehabilitation [cph],
  David Schmidt [aut, cre],
  Mark Rembert [aut]
Maintainer: David Schmidt <deschmidt@detr.nv.gov>
Repository: CRAN
Date/Publication: 2026-04-23 05:10:59 UTC

# Generated by roxygen2: do not edit by hand

export(bls_get_cache_dir)
export(bls_overview)
export(create_bls_object)
export(download_bls_files)
export(fread_bls)
export(get_bls_data)
export(get_bls_diagnostics)
export(get_bls_summary)
export(get_ces)
export(get_jolts)
export(get_laus)
export(get_national_ces)
export(get_oews)
export(get_oews_areas)
export(get_qcew)
export(get_salt)
export(has_bls_issues)
export(list_ces_industries)
export(list_ces_states)
export(list_national_ces_options)
export(load_bls_dataset)
export(print_bls_warnings)
export(read_bls_excel)
export(read_bls_text)
export(show_ces_options)
export(show_national_ces_options)
export(smart_bls_download)
importFrom(data.table,":=")
importFrom(data.table,data.table)
importFrom(data.table,fcase)
importFrom(data.table,fread)
importFrom(data.table,rbindlist)
importFrom(dplyr,across)
importFrom(dplyr,arrange)
importFrom(dplyr,case_when)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,lag)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,rename_with)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,ungroup)
importFrom(htmltools,HTML)
importFrom(htmltools,htmlEscape)
importFrom(httr,GET)
importFrom(httr,HEAD)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,headers)
importFrom(httr,http_status)
importFrom(httr,progress)
importFrom(httr,status_code)
importFrom(httr,stop_for_status)
importFrom(httr,write_disk)
importFrom(lubridate,ym)
importFrom(lubridate,yq)
importFrom(readxl,read_excel)
importFrom(rstudioapi,isAvailable)
importFrom(rvest,html_attr)
importFrom(rvest,html_elements)
importFrom(rvest,read_html)
importFrom(sf,st_as_sf)
importFrom(sf,st_union)
importFrom(stats,setNames)
importFrom(stringr,str_detect)
importFrom(stringr,str_length)
importFrom(stringr,str_pad)
importFrom(stringr,str_remove)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_to_lower)
importFrom(tidyselect,everything)
importFrom(tidyselect,matches)
importFrom(tidyselect,starts_with)
importFrom(tigris,counties)
importFrom(tigris,shift_geometry)
importFrom(tigris,states)
importFrom(tools,R_user_dir)
importFrom(utils,head)
importFrom(zoo,as.yearqtr)

# BLSloadR
Functions for downloading BLS flat files into R

# Overview
BLSloadR is a packages designed to streamline access to the time series database downloads from the U.S. Bureau of Labor Statistics, made available at https://download.bls.gov/pub/time.series/. It is focused on accessing series that are frequently used by states to get state-level estimates, but includes the `load_bls_dataset()` and `bls_overview()` functions to provide generalized access to the other databases at this website within an R environment.

# Basic Usage
BLSloadR can be installed from CRAN by running `install.packages("BLSloadR")`. The development version can be found on Github at https://github.com/schmidtDETR/BLSloadR. 

The primary functions in this package all begin with get_ and are listed below:

-`get_ces()` - This accesses data from the Current Employment Statistics (CES) program at the state and metropolitan area levels. This provides employer-based estimates of employment, wages, and hours worked. This is the “SM” database.

-`get_national_ces()` - This accesses national data from the CES program at the national level, which does not include state-level breakouts. This is the “CE” database.

-`get_laus()` - This accesses data from the Local Area Unemployment Statistics (LAUS) program at a regional, state, and several sub-state levels. This is a localized version of the Current Population Survey (CPS) which is used to drive household-based estiamtes of employment and unemployment. This is the “LA” database. Note that because of the volume of data here, there are several different geographies that may be specified to pull the appropriate data file from BLS.

-`get_oews()` - This access the Occupational Employment and Wage Statistics (OEWS) data. This data provides survey-based estimates of employmen and wages by occupation at state and sub-state levels. This is the “OE” database. Note that only current-year data is available for OEWS in this database, as it is not built as a time series.

-`get_salt()` - This data is not actually located within the time.series folder, but instead is sourced from https://www.bls.gov/lau/stalt.htm. These Alternative Measures of Labor Underutilization for States are 12-month averages built from CPS data which provide more expansive or restrictive definitions of unemployment to measure the labor force, known as U1 through U6. This function also includes the optional geometry argument. If set to TRUE, this will use `tigris::states()` and `tigris::shift_geometry()` to provide state polygons for convenient mapping of the output.

-`get_jolts()` - This accesses data from the Job Openings and Labor Turnover Survey (JOLTS) which has both national and state data.  NOTE: Beginning in 2026, the BLS will discontinue publishing state JOLTS data on a monthly basis, moving to an annual release of data. This function will still pull data, but the frequency of state data will change. This is the "JT" database.

-`get_qcew()` - This accesses data from the Quarterly Census of Employment and Wages (QCEW).  This is a very large data set, so access is filtered by area or industry.  This function iterates requesting single-quarter files via the BLS QCEW Data Slices tool at https://www.bls.gov/cew/additional-resources/open-data/csv-data-slices.htm.  This function was included beginning in version 0.3.1.

# Configuring Your User Profile
BLSloadR will typically work by default without any cusomization.  However, there are some options you can use that may improve your experience.  These options are managed with *environment variables* in your R session that enable the following:

-`BLS_USER_AGENT` - setting this environment variable to your e-mail address will use your e-mail address when downloading data from the BLS. In case of errors with your downloads, this may help the BLS to identify you as an individual user. Setting this environment variable to a character string passes that character string to the BLS as the User-Agent HTML header.

-`USE_BLS_CACHE` - Setting this environment variable to "TRUE" will enable a local file cache of your BLS downloads which will download new files for supported functions only when the underlying data has changed.

-`BLS_CACHE_DIR` - If you want to use the file cache, you may wish to specify a location.  Setting this environment variable will specify a different path for the file cache than the default.

To permanently set these environment variables, you can edit your .Renviron file (such as with `usethis::edit_r_environ()`). To do so for a single session, you can set your environment variables with `Sys.setenv(USE_BLS_CACHE="TRUE")`.

# General BLS Time Series Functions
These optional helper functions can aid the user of this package by providing ways to summarize and explore all the time.series databases. These functions are a bit different than the specific functions above, as they implement a general way to merge and import BLS time.series databases, but do not manually specify the data, series, and lookup files to be joined.  As such, they return a bls_data_collection object which includes the joined data as well as diagnostic results including dropped columns, unexpected join results, and other tools to help review the data before use.  Further, when multiple data or series files are present, the user is prompted to choose one, so these tools are not suitable for a typical piped script.

`bls_overview()` - this function utilizes the standard structure of the time.series databases, which has a simple text file explaining the database structure that always follows the structure id.txt where id is the two-character database identification code.

`load_bls_dataset()` - this function attempts to read and join all the relevant files in a BLS database, and will sometimes prompt the user for additional input. For example, many databases have multiple data files available (such as “AllItems” and “Current”) and may have old series files as well (to manage historical coding changes). Because these joins are performed automatically, the object returned by this function is a more robust diagnostic object included the joined data table as well as information about the joins. Use Caution! BLS data structures are not always consistent. There may be anomalies in the structure of individual databases, such as missing column headers, that will degrade the ability of this function to read the data.

<!-- badges: start -->
  [![R-CMD-check](https://github.com/schmidtDETR/BLSloadR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/schmidtDETR/BLSloadR/actions/workflows/R-CMD-check.yaml)
  <!-- badges: end -->

#' List Available National CES Dataset Options
#'
#' This function displays the available dataset filtering options for the
#' get_national_ces() function, helping users understand what specialized
#' datasets are available for download.
#'
#' @param show_descriptions Logical. If TRUE, shows detailed descriptions
#'   of each dataset option. If FALSE (default), shows only the filter names.
#'
#' @return A data frame with dataset filter options and their descriptions.
#'
#' @examples
#' # Show available dataset filters
#' list_national_ces_options()
#'
#' # Show detailed descriptions
#' list_national_ces_options(show_descriptions = TRUE)
#'
#' @export
list_national_ces_options <- function(show_descriptions = FALSE) {

  datasets <- data.frame(
    filter = c(
      "all_data",
      "current_seasonally_adjusted",
      "real_earnings_all_employees",
      "real_earnings_production"
    ),
    description = c(
      "Complete national CES dataset - all series and full history",
      "Seasonally adjusted all-employee series only (faster download)",
      "Real earnings data (1982-84 dollars) for all employees",
      "Real earnings data (1982-84 dollars) for production employees"
    ),
    stringsAsFactors = FALSE
  )

  if (show_descriptions) {
    return(datasets)
  } else {
    return(datasets$filter)
  }
}

#' Show National CES Dataset Options and Usage Examples
#'
#' This function provides a comprehensive overview of the national CES dataset
#' filtering options available in get_national_ces(), including examples
#' of how to use each option.
#'
#' @return Prints formatted information to the console.
#'
#' @examples
#' show_national_ces_options()
#'
#' @export
show_national_ces_options <- function() {
  cat("=== BLS National Current Employment Statistics (CES) Dataset Options ===\n\n")

  cat("AVAILABLE DATASETS (4 options):\n")
  datasets <- list_national_ces_options(show_descriptions = TRUE)

  for(i in 1:nrow(datasets)) {
    cat("  ", datasets$filter[i], ": ", datasets$description[i], "\n")
  }

  cat("\nUSAGE EXAMPLES:\n")
  cat("  # Complete dataset (largest file, ~340MB)\n")
  cat("  ces_complete <- get_national_ces(dataset_filter = 'all_data')\n\n")

  cat("  # Seasonally adjusted data only (faster download)\n")
  cat("  ces_seasonal <- get_national_ces(dataset_filter = 'current_seasonally_adjusted')\n\n")

  cat("  # Real earnings for all employees\n")
  cat("  ces_earnings_all <- get_national_ces(dataset_filter = 'real_earnings_all_employees')\n\n")

  cat("  # Real earnings for production employees\n")
  cat("  ces_earnings_prod <- get_national_ces(dataset_filter = 'real_earnings_production')\n\n")

  cat("Performance: Specialized datasets reduce download time significantly!\n")
  cat("Note: All options include metadata files for context and labels.\n")
}

#' List Available States for CES Data
#'
#' @description
#' Lists all available U.S. states and territories that can be used with the
#' `states` parameter in `get_ces()` function.
#'
#' @return A character vector of available state/territory abbreviations
#' @export
#' @examples
#' # See all available states
#' list_ces_states()
#'
#' # Use with get_ces
#' # ces_data <- get_ces(states = c("MA", "NY"))  # All industries for these states
list_ces_states <- function() {
  states <- c(
    "AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DE", "DC", "FL",
    "GA", "HI", "ID", "IL", "IN", "IA", "KS", "KY", "LA", "ME",
    "MD", "MA", "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH",
    "NJ", "NM", "NY", "NC", "ND", "OH", "OK", "OR", "PA", "PR",
    "RI", "SC", "SD", "TN", "TX", "UT", "VT", "VA", "VI", "WA",
    "WV", "WI", "WY"
  )
  return(states)
}

#' List Available Industry Filters for CES Data
#'
#' @description
#' Lists all available industry categories that can be used with the
#' `industry_filter` parameter in `get_ces()` function. These filters allow
#' you to download specific industry data instead of the complete dataset.
#'
#' @param show_descriptions Logical. If TRUE, returns a data frame with
#'   filter names and descriptions. If FALSE, returns just the filter names.
#' @return A character vector of industry filter names, or a data frame
#'   with names and descriptions if show_descriptions = TRUE
#' @export
#' @examples
#' # See all available industry filters
#' list_ces_industries()
#'
#' # See filters with descriptions
#' list_ces_industries(show_descriptions = TRUE)
#'
#' # Use with get_ces
#' # manufacturing_data <- get_ces(industry_filter = "manufacturing")  # All states
list_ces_industries <- function(show_descriptions = FALSE) {

  industry_filters <- c(
    "current_year", "total_nonfarm", "total_nonfarm_statewide", "total_private",
    "goods_producing", "service_providing", "private_service_providing",
    "mining_logging", "mining_logging_construction", "construction",
    "manufacturing", "durable_goods", "nondurable_goods", "trade_trans_utilities",
    "wholesale_trade", "retail_trade", "trans_utilities", "information",
    "financial_activities", "prof_business_services", "edu_health_services",
    "leisure_hospitality", "other_services", "government"
  )

  if (!show_descriptions) {
    return(industry_filters)
  }

  # Create descriptions for each filter
  descriptions <- c(
    "Recent data across all industries (2006-present)",
    "Total non-farm employment (all industries, all years)",
    "Total non-farm employment (statewide level)",
    "Total private sector employment",
    "Goods-producing industries",
    "Service-providing industries",
    "Private service-providing industries",
    "Mining and logging",
    "Mining, logging, and construction",
    "Construction",
    "Manufacturing",
    "Durable goods manufacturing",
    "Non-durable goods manufacturing",
    "Trade, transportation, and utilities",
    "Wholesale trade",
    "Retail trade",
    "Transportation and utilities",
    "Information services",
    "Financial activities",
    "Professional and business services",
    "Education and health services",
    "Leisure and hospitality",
    "Other services",
    "Government"
  )

  result <- data.frame(
    filter = industry_filters,
    description = descriptions,
    stringsAsFactors = FALSE
  )

  return(result)
}

#' Show CES Data Filtering Options
#'
#' @description
#' Displays a comprehensive overview of all filtering options available
#' for the `get_ces()` function, including states, industries, and usage examples.
#'
#' @export
#' @examples
#' # See all filtering options
#' show_ces_options()
show_ces_options <- function() {
  cat("=== BLS Current Employment Statistics (CES) Filtering Options ===\n\n")

  cat("AVAILABLE STATES (", length(list_ces_states()), " total):\n", sep = "")
  states <- list_ces_states()
  # Print states in rows of 10
  for (i in seq(1, length(states), 10)) {
    end_idx <- min(i + 9, length(states))
    cat("  ", paste(states[i:end_idx], collapse = ", "), "\n")
  }

  cat("\nAVAILABLE INDUSTRY FILTERS (", length(list_ces_industries()), " total):\n", sep = "")
  industries <- list_ces_industries(show_descriptions = TRUE)
  for (i in 1:nrow(industries)) {
    cat("  ", sprintf("%-25s", industries$filter[i]), ": ", industries$description[i], "\n", sep = "")
  }

  cat("USAGE EXAMPLES:\n")
  cat("  # Specific states (all industries, all years)\n")
  cat("  ces_states <- get_ces(states = c('MA', 'NY', 'CT'))\n\n")

  cat("  # Specific industry (all states, 2007-present)\n")
  cat("  ces_manufacturing <- get_ces(industry_filter = 'manufacturing')\n\n")

  cat("  # Current year data (all states and industries, 2006-present)\n")
  cat("  ces_current <- get_ces(current_year_only = TRUE)\n\n")

  cat("  # Complete dataset (all states, industries, and years - slowest)\n")
  cat("  ces_all <- get_ces()\n\n")

  cat("Performance: Filtering reduces download time by 50-90% vs. full dataset!\n")
  cat("Note: Parameters are mutually exclusive - choose only one filtering option.\n")
}

#' Download Current Employment Statistics (CES) Data
#'
#' This function downloads Current Employment Statistics data from the Bureau of Labor Statistics.
#' The data includes national, regional, state, and substate employment statistics.
#' By default, all available areas, data types, and periods are included.
#'
#' @param states Character vector of state abbreviations to download (e.g., c("MA", "NY", "CA")).
#'   If specified, downloads only these states (all industries, all years).
#'   Cannot be combined with industry_filter or current_year_only.
#'   Use `list_ces_states()` to see all available states.
#' @param industry_filter Character string specifying industry category to download.
#'   If specified, downloads this industry for all states (2007-present).
#'   Cannot be combined with states or current_year_only.
#'   Use `list_ces_industries()` to see all available industry filters.
#' @param current_year_only Logical. If TRUE, downloads the current year file
#'   which contains all states and industries for recent years (2006-present).
#'   Cannot be combined with states or industry_filter. If FALSE (default), uses other parameters.
#' @param transform Logical. If TRUE (default), converts employment values from thousands
#'   to actual counts by multiplying by 1000 for specific data types (codes 1, 6, 26)
#'   and removes ", In Thousands" from data type labels.
#' @param monthly_only Logical. If TRUE (default), filters out annual data (period M13).
#' @param simplify_table Logical. If TRUE (default), removes excess columns and creates
#'   a date column from Year and Period in the original data.
#' @param suppress_warnings Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
#'   for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.
#' @param return_diagnostics Logical. If FALSE (default), returns only the data. If TRUE,
#'   returns the full bls_data_collection object with diagnostics.
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'
#' @return By default, returns a data.table with CES data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing data and comprehensive diagnostics.
#'
#' @details
#' **Performance Notes:** The default behavior downloads a very large file (~500MB+) containing
#' all states and industries, which can take several minutes. For faster downloads, consider:
#' \itemize{
#'   \item Use \code{states = c("MA", "NY")} to download only specific states
#'   \item Use \code{industry_filter = "total_nonfarm"} for summary employment data only
#'   \item Use \code{current_year_only = TRUE} for recent data only (2006-present)
#' }
#'
#' **State Codes:** Use standard two-letter state abbreviations (e.g., "MA", "CA", "NY").
#' Puerto Rico = "PR", Virgin Islands = "VI", District of Columbia = "DC".
#'
#' **Industry Filters:** Available options include:
#' \itemize{
#'   \item "total_nonfarm" - Total non-farm employment summary
#'   \item "total_private" - Private sector totals (2007-present)
#'   \item "manufacturing" - Manufacturing sector (2007-present)
#'   \item "construction" - Construction sector (2007-present)
#'   \item "retail_trade" - Retail trade sector (2007-present)
#'   \item "government" - Government sector (2007-present)
#'   \item And others - see BLS documentation for full list
#' }
#'
#' @seealso
#' \code{\link{list_ces_states}()} to see available states,
#' \code{\link{list_ces_industries}()} to see available industry filters,
#' \code{\link{show_ces_options}()} for a comprehensive overview of filtering options.
#'
#' @export
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @importFrom dplyr if_else
#' @importFrom stringr str_remove
#' @importFrom lubridate ym
#' @examples
#' \dontrun{
#' # Fast download: Massachusetts and Connecticut data only (all industries)
#' ces_states <- get_ces(states = c("MA", "CT"))
#'
#' # Fast download: Manufacturing data for all states
#' ces_manufacturing <- get_ces(industry_filter = "manufacturing")
#'
#' # Fast download: Current year data for all states and industries
#' ces_current <- get_ces(current_year_only = TRUE)
#'
#' # Complete dataset (slower - all states, industries, and years)
#' ces_all <- get_ces()
#'
#' # Download with full diagnostics if needed
#' ces_result <- get_ces(states = "MA", return_diagnostics = TRUE)
#' ces_data <- get_bls_data(ces_result)
#'
#' # Check for download issues
#' if (has_bls_issues(ces_result)) {
#'   print_bls_warnings(ces_result)
#' }
#' }
get_ces <- function(states = NULL, industry_filter = NULL, current_year_only = FALSE,
                    transform = TRUE, monthly_only = TRUE, simplify_table = TRUE,
                    suppress_warnings = TRUE, return_diagnostics = FALSE, cache = check_bls_cache_env()) {


  # Define state-specific URLs mapping
  state_urls <- list(
    "AL" = "https://download.bls.gov/pub/time.series/sm/sm.data.1.Alabama",
    "AK" = "https://download.bls.gov/pub/time.series/sm/sm.data.2.Alaska",
    "AZ" = "https://download.bls.gov/pub/time.series/sm/sm.data.3.Arizona",
    "AR" = "https://download.bls.gov/pub/time.series/sm/sm.data.4.Arkansas",
    "CA" = "https://download.bls.gov/pub/time.series/sm/sm.data.5c.California",
    "CO" = "https://download.bls.gov/pub/time.series/sm/sm.data.6.Colorado",
    "CT" = "https://download.bls.gov/pub/time.series/sm/sm.data.7.Connecticut",
    "DE" = "https://download.bls.gov/pub/time.series/sm/sm.data.8.Delaware",
    "DC" = "https://download.bls.gov/pub/time.series/sm/sm.data.9.DC",
    "FL" = "https://download.bls.gov/pub/time.series/sm/sm.data.10b.Florida",
    "GA" = "https://download.bls.gov/pub/time.series/sm/sm.data.11.Georgia",
    "HI" = "https://download.bls.gov/pub/time.series/sm/sm.data.12.Hawaii",
    "ID" = "https://download.bls.gov/pub/time.series/sm/sm.data.13.Idaho",
    "IL" = "https://download.bls.gov/pub/time.series/sm/sm.data.14.Illinois",
    "IN" = "https://download.bls.gov/pub/time.series/sm/sm.data.15.Indiana",
    "IA" = "https://download.bls.gov/pub/time.series/sm/sm.data.16.Iowa",
    "KS" = "https://download.bls.gov/pub/time.series/sm/sm.data.17.Kansas",
    "KY" = "https://download.bls.gov/pub/time.series/sm/sm.data.18.Kentucky",
    "LA" = "https://download.bls.gov/pub/time.series/sm/sm.data.19.Louisiana",
    "ME" = "https://download.bls.gov/pub/time.series/sm/sm.data.20.Maine",
    "MD" = "https://download.bls.gov/pub/time.series/sm/sm.data.21.Maryland",
    "MA" = "https://download.bls.gov/pub/time.series/sm/sm.data.22.Massachusetts",
    "MI" = "https://download.bls.gov/pub/time.series/sm/sm.data.23b.Michigan",
    "MN" = "https://download.bls.gov/pub/time.series/sm/sm.data.24.Minnesota",
    "MS" = "https://download.bls.gov/pub/time.series/sm/sm.data.25.Mississippi",
    "MO" = "https://download.bls.gov/pub/time.series/sm/sm.data.26.Missouri",
    "MT" = "https://download.bls.gov/pub/time.series/sm/sm.data.27.Montana",
    "NE" = "https://download.bls.gov/pub/time.series/sm/sm.data.28.Nebraska",
    "NV" = "https://download.bls.gov/pub/time.series/sm/sm.data.29.Nevada",
    "NH" = "https://download.bls.gov/pub/time.series/sm/sm.data.30.NewHampshire",
    "NJ" = "https://download.bls.gov/pub/time.series/sm/sm.data.31.NewJersey",
    "NM" = "https://download.bls.gov/pub/time.series/sm/sm.data.32.NewMexico",
    "NY" = "https://download.bls.gov/pub/time.series/sm/sm.data.33b.NewYork",
    "NC" = "https://download.bls.gov/pub/time.series/sm/sm.data.34.NorthCarolina",
    "ND" = "https://download.bls.gov/pub/time.series/sm/sm.data.35.NorthDakota",
    "OH" = "https://download.bls.gov/pub/time.series/sm/sm.data.36.Ohio",
    "OK" = "https://download.bls.gov/pub/time.series/sm/sm.data.37.Oklahoma",
    "OR" = "https://download.bls.gov/pub/time.series/sm/sm.data.38.Oregon",
    "PA" = "https://download.bls.gov/pub/time.series/sm/sm.data.39b.Pennsylvania",
    "PR" = "https://download.bls.gov/pub/time.series/sm/sm.data.40.PuertoRico",
    "RI" = "https://download.bls.gov/pub/time.series/sm/sm.data.41.RhodeIsland",
    "SC" = "https://download.bls.gov/pub/time.series/sm/sm.data.42.SouthCarolina",
    "SD" = "https://download.bls.gov/pub/time.series/sm/sm.data.43.SouthDakota",
    "TN" = "https://download.bls.gov/pub/time.series/sm/sm.data.44.Tennessee",
    "TX" = "https://download.bls.gov/pub/time.series/sm/sm.data.45c.Texas",
    "UT" = "https://download.bls.gov/pub/time.series/sm/sm.data.46.Utah",
    "VT" = "https://download.bls.gov/pub/time.series/sm/sm.data.47.Vermont",
    "VA" = "https://download.bls.gov/pub/time.series/sm/sm.data.48.Virginia",
    "VI" = "https://download.bls.gov/pub/time.series/sm/sm.data.49.VirginIslands",
    "WA" = "https://download.bls.gov/pub/time.series/sm/sm.data.50.Washington",
    "WV" = "https://download.bls.gov/pub/time.series/sm/sm.data.51.WestVirginia",
    "WI" = "https://download.bls.gov/pub/time.series/sm/sm.data.52.Wisconsin",
    "WY" = "https://download.bls.gov/pub/time.series/sm/sm.data.53.Wyoming"
  )

  # Define industry-specific URLs mapping (current data, 2007-present)
  industry_urls <- list(
    "current_year" = "https://download.bls.gov/pub/time.series/sm/sm.data.0.Current",
    "total_nonfarm" = "https://download.bls.gov/pub/time.series/sm/sm.data.54.TotalNonFarm.All",
    "total_nonfarm_statewide" = "https://download.bls.gov/pub/time.series/sm/sm.data.55.TotalNonFarmStatewide.All",
    "total_private" = "https://download.bls.gov/pub/time.series/sm/sm.data.56.TotalPrivate.Current",
    "goods_producing" = "https://download.bls.gov/pub/time.series/sm/sm.data.57.GoodsProducing.Current",
    "service_providing" = "https://download.bls.gov/pub/time.series/sm/sm.data.58.ServiceProviding.Current",
    "private_service_providing" = "https://download.bls.gov/pub/time.series/sm/sm.data.59.PrivateServiceProviding.Current",
    "mining_logging" = "https://download.bls.gov/pub/time.series/sm/sm.data.60.MiningandLogging.Current",
    "mining_logging_construction" = "https://download.bls.gov/pub/time.series/sm/sm.data.61.MiningLoggingConstr.Current",
    "construction" = "https://download.bls.gov/pub/time.series/sm/sm.data.62.Construction.Current",
    "manufacturing" = "https://download.bls.gov/pub/time.series/sm/sm.data.63.Manufacturing.Current",
    "durable_goods" = "https://download.bls.gov/pub/time.series/sm/sm.data.64.DurableGoods.Current",
    "nondurable_goods" = "https://download.bls.gov/pub/time.series/sm/sm.data.65.NonDurableGoods.Current",
    "trade_trans_utilities" = "https://download.bls.gov/pub/time.series/sm/sm.data.66.TradeTransUtilities.Current",
    "wholesale_trade" = "https://download.bls.gov/pub/time.series/sm/sm.data.67.WholesaleTrade.Current",
    "retail_trade" = "https://download.bls.gov/pub/time.series/sm/sm.data.68.RetailTrade.Current",
    "trans_utilities" = "https://download.bls.gov/pub/time.series/sm/sm.data.69.TransUtilities.Current",
    "information" = "https://download.bls.gov/pub/time.series/sm/sm.data.70.Information.Current",
    "financial_activities" = "https://download.bls.gov/pub/time.series/sm/sm.data.71.FinancialActivities.Current",
    "prof_business_services" = "https://download.bls.gov/pub/time.series/sm/sm.data.72.ProfBusSrvc.Current",
    "edu_health_services" = "https://download.bls.gov/pub/time.series/sm/sm.data.73.EduHealthSrvc.Current",
    "leisure_hospitality" = "https://download.bls.gov/pub/time.series/sm/sm.data.74.LeisureandHospitality.Current",
    "other_services" = "https://download.bls.gov/pub/time.series/sm/sm.data.75.OtherServices.Current",
    "government" = "https://download.bls.gov/pub/time.series/sm/sm.data.76.Government.Current"
  )

  # Validate inputs
  if (!is.null(states)) {
    states <- toupper(states)
    invalid_states <- states[!states %in% names(state_urls)]
    if (length(invalid_states) > 0) {
      stop("Invalid state codes: ", paste(invalid_states, collapse = ", "),
           ". Use standard two-letter abbreviations like 'MA', 'NY', 'CA'.")
    }
  }

  if (!is.null(industry_filter)) {
    if (!industry_filter %in% names(industry_urls)) {
      stop("Invalid industry_filter: '", industry_filter, "'. ",
           "Valid options: ", paste(names(industry_urls), collapse = ", "))
    }
  }

  # Check for conflicting parameters (mutually exclusive)
  param_count <- sum(!is.null(states), !is.null(industry_filter), current_year_only)
  if (param_count > 1) {
    stop("Parameters 'states', 'industry_filter', and 'current_year_only' are mutually exclusive. ",
         "Choose only one filtering option.")
  }

  # Build data URLs based on filters (mutually exclusive)
  data_urls <- c()

  if (current_year_only) {
    # Current year file contains all states/industries for current year
    data_urls <- c("Main Data" = industry_urls[["current_year"]])
    if(!suppress_warnings){
      message("Using current year data file (2006-present, all states and industries)")
    }
  } else if (!is.null(industry_filter)) {
    # Use industry-specific file (all states for this industry)
    data_urls <- c("Main Data" = industry_urls[[industry_filter]])
    if(!suppress_warnings){
      message("Using industry filter: ", industry_filter, " (2007-present data, all states)")
    }
  } else if (!is.null(states)) {
    # Use state-specific files (all industries for these states)
    state_data_urls <- state_urls[states]
    names(state_data_urls) <- paste("State Data -", names(state_data_urls))
    data_urls <- state_data_urls
    if(!suppress_warnings){
      message("Downloading data for states: ", paste(states, collapse = ", "), " (all industries)")
    }
  } else {
    # Use the large AllData file (original behavior)
    data_urls <- c("Main Data" = "https://download.bls.gov/pub/time.series/sm/sm.data.1.AllData")
    if(!suppress_warnings){
      message("Downloading complete dataset (all states, all industries, all years - this may take several minutes)...")
    }
  }

  # Define URLs for CES metadata files (always needed)
  ces_urls <- c(
    data_urls,  # Add the data URLs we determined above
    "Series Metadata" = "https://download.bls.gov/pub/time.series/sm/sm.series",
    "Industry Codes" = "https://download.bls.gov/pub/time.series/sm/sm.industry",
    "State Codes" = "https://download.bls.gov/pub/time.series/sm/sm.state",
    "Area Codes" = "https://download.bls.gov/pub/time.series/sm/sm.area",
    "Data Types" = "https://download.bls.gov/pub/time.series/sm/sm.data_type",
    "Supersector Codes" = "https://download.bls.gov/pub/time.series/sm/sm.supersector"
  )

  # Download all files
  if(!suppress_warnings){message("Starting CES data download...\n")}
  downloads <- download_bls_files(ces_urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Exit function if download failed.
  if(is.null(downloads) | length(downloads) == 0 | length(ces_urls) != length(downloads)){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
  }

  # Extract data from downloads - handle multiple data files when downloading by states
  if (!is.null(states) && !current_year_only && is.null(industry_filter)) {
    # Multiple state files - combine them
    state_data_list <- list()
    for (state in states) {
      state_file_name <- paste("State Data -", state)
      if (state_file_name %in% names(downloads)) {
        state_data_list[[state]] <- get_bls_data(downloads[[state_file_name]])
      }
    }
    data_main <- do.call(rbind, state_data_list)
  } else {
    # Single data file (industry filter, current year, or all data)
    main_data_name <- names(downloads)[grepl("Data", names(downloads))][1]
    data_main <- get_bls_data(downloads[[main_data_name]])
  }

  # Extract metadata (always needed)
  data_series <- get_bls_data(downloads$`Series Metadata`)
  data_industry <- get_bls_data(downloads$`Industry Codes`)
  data_state <- get_bls_data(downloads$`State Codes`)
  data_area <- get_bls_data(downloads$`Area Codes`)
  data_types <- get_bls_data(downloads$`Data Types`)
  data_supersector <- get_bls_data(downloads$`Supersector Codes`)

  # Track processing steps
  processing_steps <- character(0)

  # Combine all data
  if(!suppress_warnings){message("Combining datasets...\n")}
  ces_data <- data_main |>
    dplyr::select(-footnote_codes) |>
    dplyr::left_join(data_series, by = "series_id") |>
    #dplyr::select(-footnote_codes) |>
    dplyr::left_join(data_industry, by = "industry_code") |>
    dplyr::left_join(data_state, by = "state_code") |>
    dplyr::left_join(data_area, by = "area_code") |>
    dplyr::left_join(data_types, by = "data_type_code") |>
    dplyr::left_join(data_supersector, by = "supersector_code") |>
    dplyr::mutate(value = as.numeric(value),
                  industry_code = substr(series_id,11,18)) |>
    dplyr::filter(!is.na(value))

  processing_steps <- c(processing_steps, "joined_metadata", "converted_values", "removed_na")

  if(transform){
    if(!suppress_warnings){message("Applying value transformations...\n")}
    ces_data <- ces_data |>
      dplyr::mutate(
        value = if_else(
          data_type_code %in% c("01","06","26"),
          value * 1000,
          value
        ),
        data_type_text = stringr::str_remove(data_type_text, ", In Thousands")
      )
    processing_steps <- c(processing_steps, "transformed_values")
  }

  if(monthly_only){
    if(!suppress_warnings){message("Filtering to monthly data only...\n")}
    ces_data <- ces_data |>
      dplyr::filter(period != "M13")
    processing_steps <- c(processing_steps, "monthly_only")
  }

  if(simplify_table){
    if(!suppress_warnings){message("Simplifying table structure...\n")}
    ces_data <- ces_data |>
      dplyr::mutate(date = lubridate::ym(paste0(year,period))) |>
      dplyr::select(-c(benchmark_year:end_period,year,period)) |>
      dplyr::filter(state_code != "00")
    processing_steps <- c(processing_steps, "simplified_table", "added_date_column")
  }

  # Create BLS data object with diagnostics
  result <- create_bls_object(
    data = ces_data,
    downloads = downloads,
    data_type = "CES",
    processing_steps = processing_steps
  )

  # Print summary
  if(!suppress_warnings){
    message("CES data download complete!\n")
    message("Final dataset dimensions: ", paste(dim(ces_data), collapse = " x "), "\n")

    if (!is.null(states)) {
      message("States: ", paste(states, collapse = ", "), "\n")
    }
    if (!is.null(industry_filter)) {
      message("Industry filter: ", industry_filter, "\n")
    }
    if (current_year_only) {
      message("Dataset: Current year data (2006-present, all states and industries)\n")
    }
    if (is.null(states) && is.null(industry_filter) && !current_year_only) {
      message("Dataset: Complete CES data (all states, industries, and years)\n")
    }
  }

  if (has_bls_issues(result)) {
    if (!suppress_warnings) {
      message("\nDownload Issues Summary:\n")
      message("Total warnings:", result$summary$total_warnings, "\n")
      message("Files with issues:", result$summary$files_with_issues, "of", result$summary$files_downloaded, "\n")
      message("Run with return_diagnostics = TRUE and use print_bls_warnings() for details\n")
    }
  } else {
    message("No download issues detected.\n")
  }

  # Return based on user preference
  if (return_diagnostics) {
    return(result)
  } else {
    # Store diagnostics as attributes for later access if needed
    attr(ces_data, "bls_diagnostics") <- result
    return(ces_data)
  }
}

#' Download Occupational Employment and Wage Statistics (OEWS) Data
#'
#' This function downloads and joins together occupational employment and wage data
#' from the Bureau of Labor Statistics OEWS program. The data includes employment
#' and wage estimates by occupation and geographic area. Note that OEWS is a large data set (over 6 million rows), so it will require longer to download.
#'
#' @param simplify_table Logical.  If TRUE (default), remove columns from the result that are internal BLS references or can be derived from other elements in the table.
#' @param suppress_warnings Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
#'   for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.
#' @param return_diagnostics Logical. If TRUE, returns a bls_data_collection object
#'   with full diagnostics. If FALSE (default), returns just the data table.
#' @param fast_read Logical.  If TRUE (default), derives lookup values directly from series_id to avoid reading the series file, to speed download process. With fast_read, the data can download in 17 seconds (depending on bandwidth).  Without fast_read, the same download takes 57 seconds.
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'
#' @return By default, returns a data.table with OEWS data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing data and comprehensive diagnostics. The columns in the returned data frame when `simplify_table = TRUE` are listed below.  Unless otherwise specified, all data is returned as a character string to preserve the value of leading and trailing zeroes.
#'  \itemize{
#'     \item series_id - The unique OEWS series identifier.
#'     \item year - The year to which the estimate refers.  Because OEWS is not time series data, this is always the most recent year.
#'     \item value - Numeric. The value of the given data type, for the given area, in the given industry and occupation.
#'     \item seasonal - Whether or not the data is seasonally adjusted.
#'     \item areatype_code - Code representing the type of area (National ("N"), Statewide ("S"), or Local ("M")).
#'     \item industry_code - NAICS code of the industry.
#'     \item occupation_code - SOC code of the occupation. Description given by occupation_name.
#'     \item datatype_code - Lookup code for the data type of a given row.  Description given by datatype_name.
#'     \item state_code - Two-digit FIPS code for the state.
#'     \item area_code - The unique OEWS code for a substate area. Description given by area_name.
#'     \item series_title - Descriptive title of the full series ID.
#'     \item occupation_name - The text description of the occupation.
#'     \item occupation_description - More detailed description of the tasks associated with the occupation.
#'     \item area_name - The text description of the area.
#'     \item datatype_name - The text description of the type of data represented by `value`.
#'     
#'   }
#' @export
#'
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @examples
#' \dontrun{
#' # Download current OEWS data
#' oews_data <- get_oews()
#'
#' # View available occupations
#' unique(oews_data$occupation_name)
#'
#' # Filter for specific occupation
#' software_devs <- oews_data[grepl("Software", occupation_name)]
#' 
#' # Get full diagnostic object if needed
#' oews_with_diagnostics <- get_oews(return_diagnostics = TRUE)
#' print_bls_warnings(oews_with_diagnostics)
#'}
#'

get_oews <- function(simplify_table = TRUE, suppress_warnings = TRUE, return_diagnostics = FALSE, fast_read = TRUE, cache = check_bls_cache_env()) {
  
  if(fast_read){
    download_urls <- c(
      "data" = "https://download.bls.gov/pub/time.series/oe/oe.data.0.Current",
      "occupation" = "https://download.bls.gov/pub/time.series/oe/oe.occupation",
      "area" = "https://download.bls.gov/pub/time.series/oe/oe.area",
      "datatype" = "https://download.bls.gov/pub/time.series/oe/oe.datatype"
    )
    
  } else {
  
  # Define all URLs we need to download
  download_urls <- c(
    "data" = "https://download.bls.gov/pub/time.series/oe/oe.data.0.Current",
    "series" = "https://download.bls.gov/pub/time.series/oe/oe.series",
    "occupation" = "https://download.bls.gov/pub/time.series/oe/oe.occupation",
    "area" = "https://download.bls.gov/pub/time.series/oe/oe.area",
    "datatype" = "https://download.bls.gov/pub/time.series/oe/oe.datatype"
  )
  }
  
  # Download all files
  downloads <- download_bls_files(download_urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Exit function if download failed.
  if(is.null(downloads) | length(downloads) == 0){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
  }
  
  # Extract data from downloads
  oews_current <- get_bls_data(downloads$data)
  if(!fast_read){
    oews_series <- get_bls_data(downloads$series)
  }
  oews_occupation <- get_bls_data(downloads$occupation)
  oews_area <- get_bls_data(downloads$area)
  oews_datatype <- get_bls_data(downloads$datatype)
  
  if(fast_read){
    oews <- oews_current |> 
      dplyr::select(-footnote_codes) |>
      dplyr::mutate(
        seasonal = substr(series_id,3,3),
        areatype_code = substr(series_id,4,4),
        area_code = substr(series_id,5,11),
        industry_code = substr(series_id,12,17),
        occupation_code = substr(series_id,18,23),
        datatype_code = substr(series_id,24,25)
      ) |> 
      dplyr::left_join(oews_occupation, by = "occupation_code") |>
      dplyr::left_join(oews_area, by = c("areatype_code", "area_code")) |>
      dplyr::left_join(oews_datatype, by = "datatype_code") |>
      dplyr::mutate(value = as.numeric(value))

      } else {
  
  
  # Join all the data together -- No fast_read
  oews <- oews_current |> 
    dplyr::select(-footnote_codes) |>
    dplyr::left_join(oews_series, by = "series_id") |>
    dplyr::left_join(oews_occupation, by = "occupation_code") |>
    dplyr::left_join(oews_area, by = c("areatype_code", "state_code", "area_code")) |>
    dplyr::left_join(oews_datatype, by = "datatype_code") |>
    dplyr::mutate(value = as.numeric(value))
  
  }
  
  # Track processing steps
  if(fast_read){
    processing_steps <- c(
      "Derived join columns from series_id.",
      "Joined occupation, area, and datatype metadata",
      "Converted values to numeric"
    )
    } else {
      processing_steps <- c(
    "Joined series, occupation, area, and datatype metadata",
    "Converted values to numeric"
  )
    }
  
  if(simplify_table){
    
    if(fast_read){
      oews <- oews |> 
        dplyr::select(-c(period, selectable, sort_sequence, display_level))
      
      processing_steps <- c(processing_steps, "Removed columns per simplify_table.")
      
    } else {
      oews <- oews |> 
        dplyr::select(-c(period, sector_code, footnote_codes, begin_year, begin_period, end_year, end_period, selectable, sort_sequence, display_level))
    
    processing_steps <- c(processing_steps, "Removed columns per simplify_table.")
    }
  }
  
  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = oews,
    downloads = downloads,
    data_type = "OEWS",
    processing_steps = processing_steps
  )

  # Return either the collection object or just the data
  if (return_diagnostics) {
    return(bls_collection)
  } else {
    return(oews)
  }
  
  return(result)
}

#' Download OEWS Area Definitions
#'
#' @param ref_year Four-digit year (converted to integer). The year for which to retrieve OEWS area definitions. Valid values are 2024 through current release year. Prior years included Township codes, which change the structure of the file.
#' @param silent Logical. If TRUE (default), suppress console output
#' @param geometry Logical.  If TRUE (default), downloads shapefiles for OEWS area definitions using `tigris::counties()` and `tigris::shift_geometry()` to render Alaska, Hawaii, and Puerto Rico with a focus on the area of the continental United States.
#' 
#' @return Data table which maps individual counties to OEWS area definitions.
#'   \itemize{
#'     \item fips_code - The State FIPS code
#'     \item state_name - The state name
#'     \item state_abb - The state two-character postal abbreviation
#'     \item oews_area_code - The OEWS area code defining the metropolitan area or nonmetropolitan area the county belongs to.
#'     \item oews_area_name - The OEWS area name
#'     \item county_code - The FIPS code for the county
#'     \item county_name - The county name
#'     }
#' 
#' @export
#' 
#' @importFrom httr GET
#' @importFrom httr write_disk
#' @importFrom httr add_headers
#' @importFrom readxl read_excel
#' @importFrom tigris counties
#' @importFrom sf st_union
#' @importFrom dplyr select
#' @importFrom dplyr left_join
#' @importFrom dplyr group_by
#' @importFrom dplyr summarize
#' 
#' @examples
#' \dontrun{
#'  # Get OEWS area definitions without shapefiles and with processing messages.
#'  test <- get_oews_areas(ref_year = 2024, geometry = FALSE, silent = FALSE)
#'  
#' }
#' 
get_oews_areas <- function(ref_year, silent = TRUE, geometry = TRUE){
  
  # Validate ref_year input
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  min_year <- 2024
  max_year <- current_year - 1
  
  if (is.na(as.integer(ref_year)) || length(ref_year) != 1) {
    stop("`ref_year` must be coercable to a single integer value.")
  }
  
  dl_year <- as.integer(ref_year)
  
  if (dl_year < min_year || dl_year > max_year) {
    stop(sprintf("`ref_year` must be between %d and %d. Estimates are generally released in April for the prior year.", min_year, max_year))
  }
  
  # Create download URL
  oews_url <- paste0("https://www.bls.gov/oes/",dl_year,"/may/area_definitions_m",dl_year,".xlsx")
  
  headers <- get_bls_excel_headers(
    refer = "https://www.bls.gov/oes/"
  )
  
  # Download Excel file
  if(!silent){
    message("Downloading OEWS area definitions from BLS.")
  }
  response <- httr::GET(oews_url, 
                        httr::write_disk(tf <- tempfile(fileext = ".xlsx")), 
                        httr::add_headers(.headers = headers))
  
  # Check for successful response
  httr::stop_for_status(response)
  
  # Track processing steps
  processing_steps <- character(0)
  
  # Read and process Excel file
  if(!silent){
    message(paste0("Processing OEWS area definition Excel file for ",dl_year,"."))
  }
  oews_areas <- readxl::read_excel(
    tf,
    skip = 1,
    col_types = c("text", "text", "text", "text", "text", "text", "text"),
    col_names = c("fips_code", "state_name", "state_abb", "oews_area_code", "oews_area_name", "county_code", "county_name")
  ) |> 
    dplyr::mutate(
      oews_area_code = stringr::str_pad(oews_area_code, width = 7, side = "left", pad = "0")
    )
  
  # Clean up temporary file
  unlink(tf)
  
  if(geometry == TRUE){
    
    if(!silent){
      message("Creating OEWS shapefiles")
    }
    
    oews_area <- oews_areas |> 
      dplyr::mutate(GEOID = paste0(fips_code,county_code)) |>
      dplyr::select(GEOID, oews_area_code, oews_area_name)
    
    area_shapes <- tigris::counties(year = dl_year, progress_bar = FALSE) |> 
      tigris::shift_geometry() |>
      dplyr::select(GEOID, geometry)
    
    oews_areas <- area_shapes |>
      dplyr::left_join(oews_area, by = "GEOID") |>
      dplyr::group_by(oews_area_name, oews_area_code) |>
      dplyr::summarize(geometry = sf::st_union(geometry), .groups = "drop")
  }
  
  
  return(oews_areas)
  
  
}

#' Download State Alternative Labor Market Measures (SALT) Data
#'
#' This function downloads detailed alternative unemployment measures data from BLS,
#' including U-1 through U-6 measures. The data provides a more comprehensive view
#' of labor market conditions beyond the standard unemployment rate (U-3).
#'
#' @param only_states Logical. If TRUE (default), includes only state-level data.
#'   If FALSE, includes sub-state areas like New York City where available.
#' @param geometry Logical. If TRUE, uses tigris::states() to download shapefiles for the states 
#'   to include in the data. If FALSE (default), only returns data table.
#' @param suppress_warnings Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
#'   for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.
#' @param return_diagnostics Logical. If TRUE, returns a bls_data_collection object
#'   with full diagnostics. If FALSE (default), returns just the data table.
#'
#' @return By default, returns a data.table with Alternative Measures of Labor Underutilization data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing data and comprehensive diagnostics.
#'   The function also adds derived measures and quartile comparisons across states.
#'
#' @export
#' @importFrom httr GET
#' @importFrom httr write_disk
#' @importFrom httr add_headers
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @importFrom dplyr across
#' @importFrom dplyr case_when
#' @importFrom dplyr rename_with
#' @importFrom dplyr lag
#' @importFrom dplyr group_by
#' @importFrom dplyr ungroup
#' @importFrom dplyr arrange
#' @importFrom sf st_as_sf
#' @importFrom stringr str_remove
#' @importFrom stringr str_length
#' @importFrom stringr str_to_lower
#' @importFrom stringr str_replace_all
#' @importFrom stringr str_pad
#' @importFrom lubridate yq
#' @importFrom tidyselect matches
#' @importFrom tidyselect starts_with
#' @importFrom tidyselect everything
#' @importFrom tigris states
#' @importFrom tigris shift_geometry
#' @importFrom zoo as.yearqtr
#' @importFrom readxl read_excel
#' @examples
#' \dontrun{
#' # Download state-level SALT data
#' salt_data <- get_salt()
#'
#' # Include sub-state areas
#' salt_all <- get_salt(only_states = FALSE)
#'  
#' # Download SALT with geometry included
#' get_salt(geometry = TRUE)
#' 
#' # Get full diagnostic object if needed
#' data_with_diagnostics <- get_salt(return_diagnostics = TRUE)
#' }
#' 

get_salt <- function(only_states = TRUE, geometry = FALSE, suppress_warnings = TRUE, return_diagnostics = FALSE) {
  
  salt_url <- "https://www.bls.gov/lau/stalt-moave.xlsx"
  
  # Downloading BLS Alternative Measures file
  if(!suppress_warnings){
    message("Downloading Alternative Measures from Excel file from BLS...")
  }
  salt_data <- read_bls_excel(salt_url, verbose = !suppress_warnings, skip = 1)

  # Track processing steps
  processing_steps <- character(0)
  
  if(is.null(salt_data)){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE to see status messages.")
  }
  
  # Read and process Excel file
  message("Processing SALT Excel file...\n")
  salt_data <- salt_data |> 
    dplyr::rename_with(.fn = stringr::str_to_lower) |>
    dplyr::mutate(date = lubridate::yq(paste0(`end year`, `end quarter`))) |>
    dplyr::select(-c(record, `start year`, `start quarter`, `end year`, `end quarter`, `unique period`)) |>
    dplyr::mutate(dplyr::across(tidyselect::starts_with("u-"), function(x){x = x/100})) |>
    dplyr::rename_with(.cols = tidyselect::starts_with("u-"), .fn = stringr::str_remove, pattern = "-") |>
    dplyr::rename_with(.cols = tidyselect::everything(), .fn = stringr::str_replace_all, pattern = " ", replacement = "_") |>
    dplyr::mutate(not_job_losers = unemployed - job_losers,
                  unemployed_under_14_weeks = unemployed - `unemployed_15+_weeks`,
                  losers_notlosers_ratio = job_losers / not_job_losers,
                  u1b = u3-u1,
                  u2b = u3-u2,
                  u4b = discouraged_workers / (civilian_labor_force + discouraged_workers),
                  u4c = u4 - u4b,
                  marginally_attached_not_discouraged = all_marginally_attached - discouraged_workers,
                  u5b = marginally_attached_not_discouraged / (civilian_labor_force + marginally_attached_not_discouraged),
                  u5c = u5 - (discouraged_workers / (civilian_labor_force + discouraged_workers + marginally_attached_not_discouraged)) - u5b,
                  u6b = involuntary_part_time_employed / civilian_labor_force,
                  period_name = zoo::as.yearqtr(date))
  
  processing_steps <- c(processing_steps, "read_excel", "standardized_columns", "calculated_derived_measures")
  
  # Filter to states only if requested
  if (only_states | geometry) {
    salt_data <- salt_data |>
      dplyr::mutate(fips_len = stringr::str_length(fips)) |>
      dplyr::filter(fips_len == 2) |>
      dplyr::select(-fips_len)
    processing_steps <- c(processing_steps, "filtered_states_only")
  }
  
  # Add quartile comparisons and lagged values
  salt_data <- salt_data |>
    dplyr::group_by(date) |>
    dplyr::mutate(u1_25 = quantile(u1, probs = c(0.25), na.rm = TRUE),
                  u1_50 = median(u1, na.rm = TRUE),
                  u1_75 = quantile(u1, probs = c(0.75), na.rm = TRUE),
                  u2_25 = quantile(u2, probs = c(0.25), na.rm = TRUE),
                  u2_50 = median(u2, na.rm = TRUE),
                  u2_75 = quantile(u2, probs = c(0.75), na.rm = TRUE),
                  u3_25 = quantile(u3, probs = c(0.25), na.rm = TRUE),
                  u3_50 = median(u3, na.rm = TRUE),
                  u3_75 = quantile(u3, probs = c(0.75), na.rm = TRUE),
                  u4b_25 = quantile(u4b, probs = c(0.25), na.rm = TRUE),
                  u4b_50 = median(u4b, na.rm = TRUE),
                  u4b_75 = quantile(u4b, probs = c(0.75), na.rm = TRUE),
                  u5b_25 = quantile(u5b, probs = c(0.25), na.rm = TRUE),
                  u5b_50 = median(u5b, na.rm = TRUE),
                  u5b_75 = quantile(u5b, probs = c(0.75), na.rm = TRUE)
    ) |>
    dplyr::ungroup() |>
    dplyr::group_by(state) |>
    dplyr::mutate(
      dplyr::across(tidyselect::matches("^u[0-9]"),
                    .fns = function(x){dplyr::lag(x, 4)},
                    .names = "py_{.col}")
    ) |>
    dplyr::mutate(
      dplyr::across(tidyselect::matches("^u[0-9]"),
                    .fns = function(x){dplyr::lag(x, 1)},
                    .names = "pq_{.col}")
    ) |>
    dplyr::ungroup()
  
  processing_steps <- c(processing_steps, "added_quartile_comparisons", "added_lagged_values")
  
  if(geometry){
    shapes <- tigris::states() |>
      tigris::shift_geometry() |>
      select(NAME, geometry)
    
    salt_data <- salt_data |>
      dplyr::left_join(shapes, by = c("state"="NAME")) |>
      sf::st_as_sf()
    
    processing_steps <- c(processing_steps, "added U.S. state geometry")
  }
  
  # Clean up temporary file
  unlink(tf)
  
  # Create simple download info (since this is Excel, not using fread_bls)
  download_info <- list(
    "salt_excel" = list(
      url = salt_url,
      original_dimensions = dim(salt_data),
      final_dimensions = dim(salt_data),
      phantom_columns_detected = 0,
      phantom_column_names = character(0),
      cleaning_applied = FALSE,
      header_data_mismatch = FALSE,
      original_header_count = ncol(salt_data),
      final_data_count = ncol(salt_data),
      empty_columns_removed = 0,
      final_column_names = names(salt_data),
      warnings = character(0)
    )
  )
  
  # # Create BLS data collection object
  # result <- create_bls_object(
  #   data = salt_data,
  #   downloads = list("salt_excel" = list(diagnostics = download_info$salt_excel)),
  #   data_type = "SALT",
  #   processing_steps = processing_steps
  # )
  
  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = salt_data,
    downloads = list("salt_excel" = list(diagnostics = download_info$salt_excel)),
    data_type = "SALT",
    processing_steps = processing_steps
  )
  
    # Display warnings if requested
  if (!suppress_warnings) {
    print_bls_warnings(bls_collection, detailed = FALSE)
  }
  
  # Return either the collection object or just the data
  if (return_diagnostics) {
    return(bls_collection)
  } else {
    return(salt_data)
  }
  
  # Print download complete, if warnings not disabled.
  if (!suppress_warnings) {
  message("SALT data download and processing complete.\n")
  message("Final dimensions:", paste(dim(salt_data), collapse = " x "), "\n")
  }
  
  return(result)
}

#' Download Job Openings and Labor Turnover Survey (JOLTS) Data
#'
#' This function downloads Job Openings and Labor Turnover data from the U.S. Bureau
#' of Labor Statistics. JOLTS data provides insights into job market dynamics including
#' job openings, hires, separations, quits, and layoffs. Data is available at national,
#' regional, and state levels with various industry and size class breakdowns.
#'
#' @param monthly_only Logical. If TRUE (default), excludes annual data (period M13)
#'   and includes only monthly observations.
#' @param remove_regions Logical. If TRUE (default), excludes regional aggregates
#'   (Midwest, Northeast, South, West) identified by state codes MW, NE, SO, WE.
#' @param remove_national Logical. If TRUE (default), excludes national-level data
#'   (state code 00). Set to FALSE to include national data with industry and
#'   size class breakdowns.
#' @param suppress_warnings Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
#'   for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.
#' @param return_diagnostics Logical. If TRUE, returns a bls_data_collection object
#'   with full diagnostics. If FALSE (default), returns just the data table.
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'
#' @return By default, returns a data.table with JOLTS data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing JOLTS data with the following key columns:
#'   \describe{
#'     \item{series_id}{BLS series identifier}
#'     \item{year}{Year of observation}
#'     \item{period}{Time period (M01-M12 for months)}
#'     \item{value}{JOLTS statistic value (transformed based on data type)}
#'     \item{date}{Date of observation}
#'     \item{state_text}{State name}
#'     \item{dataelement_text}{Type of JOLTS measure (job openings, hires, separations, etc.)}
#'     \item{area_text}{Geographic area description}
#'     \item{sizeclass_text}{Establishment size class}
#'     \item{industry_text}{Industry classification}
#'     \item{ratelevel_code}{Whether the value is a "Level" (count) or "Rate" (percentage)}
#'     \item{periodname}{Month name}
#'   }
#'
#' @details The function performs several data transformations:
#'   \itemize{
#'     \item Converts rate values to proportions (divides by 100) except for Unemployed to Job Opening ratio.
#'     \item Converts level values to actual counts (multiplies by 1000)
#'     \item Creates a proper date column from year and period
#'     \item Adds readable month names
#'   }
#'
#' @export
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @importFrom dplyr case_when
#' @importFrom lubridate ym
#' @examples
#' \dontrun{
#' # Download state-level JOLTS data (default - returns data directly)
#' jolts_data <- get_jolts()
#'
#' # Include national data with industry breakdowns
#' jolts_national <- get_jolts(remove_national = FALSE)
#'
#' # Get full diagnostic object if needed
#' jolts_with_diagnostics <- get_jolts(return_diagnostics = TRUE)
#' print_bls_warnings(jolts_with_diagnostics)
#'
#' # View job openings by state for latest period
#' job_openings <- jolts_data[dataelement_text == "Job openings" & 
#'                           date == max(date)]
#' }

get_jolts <- function(monthly_only = TRUE, remove_regions = TRUE, remove_national = TRUE, 
                      suppress_warnings = TRUE, return_diagnostics = FALSE, cache = check_bls_cache_env()) {
  
  # Define all URLs we need to download
  download_urls <- c(
    "data" = "https://download.bls.gov/pub/time.series/jt/jt.data.1.AllItems",
    "series" = "https://download.bls.gov/pub/time.series/jt/jt.series",
    "states" = "https://download.bls.gov/pub/time.series/jt/jt.state",
    "elements" = "https://download.bls.gov/pub/time.series/jt/jt.dataelement",
    "area" = "https://download.bls.gov/pub/time.series/jt/jt.area",
    "sizeclass" = "https://download.bls.gov/pub/time.series/jt/jt.sizeclass",
    "industry" = "https://download.bls.gov/pub/time.series/jt/jt.industry"
  )
  
  # Download all files
  downloads <- download_bls_files(download_urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Exit function if download failed.
  if(is.null(downloads) | length(downloads) == 0){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
  }
  
  # Extract data from downloads
  jolts_import <- get_bls_data(downloads$data)
  jolts_series <- get_bls_data(downloads$series)
  jolts_states <- get_bls_data(downloads$states)
  jolts_elements <- get_bls_data(downloads$elements)
  jolts_area <- get_bls_data(downloads$area)
  jolts_sizeclass <- get_bls_data(downloads$sizeclass)
  jolts_industry <- get_bls_data(downloads$industry)
  
  # Join all the data together
  jolts <- jolts_import |>
    dplyr::select(-c(footnote_codes)) |>
    dplyr::left_join(jolts_series, by = "series_id") |>
    dplyr::left_join(jolts_states |> dplyr::select(-c(display_level:sort_sequence)), by = "state_code") |>
    dplyr::left_join(jolts_elements |> dplyr::select(-c(display_level:sort_sequence)), by = "dataelement_code") |>
    dplyr::left_join(jolts_area |> dplyr::select(-c(display_level:sort_sequence)), by = "area_code") |>
    dplyr::left_join(jolts_sizeclass |> dplyr::select(-c(display_level:sort_sequence)), by = "sizeclass_code") |>
    dplyr::left_join(jolts_industry |> dplyr::select(-c(display_level:sort_sequence)), by = "industry_code")
  
  # Track processing steps
  processing_steps <- c(
    "Joined series, states, elements, area, sizeclass, and industry metadata"
  )
  
  # Apply filters
  if (monthly_only) {
    jolts <- jolts |>
      dplyr::filter(period != "M13")
    
    processing_steps <- c(processing_steps, "Filtered to monthly data only")
  }
  
  if (remove_regions) {
    jolts <- jolts |>
      dplyr::filter(!(state_code %in% c("MW", "NE", "SO", "WE")))
    
    processing_steps <- c(processing_steps, "Removed regional aggregates")
  }
  
  if (remove_national) {
    jolts <- jolts |>
      dplyr::filter(!(state_code %in% c("00")))
    
    processing_steps <- c(processing_steps, "Removed national-level data")
  }
  
  # Apply transformations
  jolts <- jolts |>
    dplyr::mutate(date = ym(paste(year, stringr::str_remove(period, "M"), sep = "-"))) |>
    dplyr::mutate(
      value = as.numeric(value),
      ratelevel_code = case_when(
        ratelevel_code == "L" ~ "Level",
        ratelevel_code == "R" ~ "Rate",
        TRUE ~ "Other"
      ),
      periodname = format(date, "%B"),
      value = if_else(dataelement_code == "UO", value * 100, value),
      value = if_else(ratelevel_code == "Rate", value / 100, value * 1000)
    )
  
  processing_steps <- c(processing_steps, 
                        "Created date column",
                        "Converted values to numeric",
                        "Transformed rate/level codes to text",
                        "Added month names",
                        "Applied value transformations (rates to proportions, levels to counts)")
  
  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = jolts,
    downloads = downloads,
    data_type = "JOLTS",
    processing_steps = processing_steps
  )
  
  # Show warnings if any issues were detected
  if (has_bls_issues(bls_collection) && !suppress_warnings) {
    print_bls_warnings(bls_collection)
  }
  
  # Return either the collection object or just the data
  if (return_diagnostics) {
    return(bls_collection)
  } else {
    return(jolts)
  }
}

#' Generate headers for BLS requests
#' 
#' Returns a named character vector of HTTP headers required for BLS API requests.
#' These headers mimic a standard browser to ensure compatibility with BLS servers.
#' 
#' @param host The URL to use in the Host header (default: "download.bls.gov")
#' @return A named character vector of HTTP headers
#' @keywords internal
get_bls_headers <- function(host = "download.bls.gov"
                            ) {
  # 1. Check for a local environment variable first
  # This allows users to set their email/identity via .Renviron or Sys.setenv()
  ua <- Sys.getenv("BLS_USER_AGENT")
  
  # 2. If the variable is empty, use a list of plausible headers to rotate
  if (ua == "") {
    plausible_agents <- c(
      "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
      "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:123.0) Gecko/20100101 Firefox/123.0",
      "Mozilla/5.0 (R; BLSloadR Package)"
    )
    # Select one at random for this session/call
    ua <- sample(plausible_agents, 1)
  }
  
  # 3. Generate dynamic headers
  
  c(
    "Accept" = "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding" = "gzip, deflate, br",
    "Accept-Language" = "en-US,en;q=0.9",
    "Connection" = "keep-alive",
    "Host" = host,
    "Referer" = refer,
    "Sec-Ch-Ua" = 'Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
    "Sec-Ch-Ua-Mobile" = "?0",
    "Sec-Ch-Ua-Platform" = '"Windows"',
    "Sec-Fetch-Dest" = "document",
    "Sec-Fetch-Mode" = "navigate",
    "Sec-Fetch-Site" = "same-origin",
    "Sec-Fetch-User" = "?1",
    "Upgrade-Insecure-Requests" = "1",
    "User-Agent" = ua
  )
}
#' Generate headers for BLS requests to download Excel files
#' 
#' Returns a named character vector of HTTP headers required for BLS API requests.
#' These headers mimic a standard browser to ensure compatibility with BLS servers.
#' This function returns a more limited set of headers used to download an Ecel file.
#' 
#' @param refer The URL to use in the Referer header (default: "https://www.bls.gov/lau/stalt-archived.htm")
#' @return A named character vector of HTTP headers
#' @keywords internal
get_bls_excel_headers <- function(refer = "https://www.bls.gov/lau/stalt-archived.htm") {
  # 1. Check for a local environment variable first
  # This allows users to set their email/identity via .Renviron or Sys.setenv()
  ua <- Sys.getenv("BLS_USER_AGENT")
  
  # 2. If the variable is empty, use a list of plausible headers to rotate
  if (ua == "") {
    plausible_agents <- c(
      "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
      "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:123.0) Gecko/20100101 Firefox/123.0",
      "Mozilla/5.0 (R; BLSloadR Package)"
    )
    # Select one at random for this session/call
    ua <- sample(plausible_agents, 1)
  }
  
  # 3. Generate dynamic headers
  
  c(
    "Referer" = refer,
    "Sec-Ch-Ua" = 'Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
    "Sec-Ch-Ua-Mobile" = "?0",
    "Sec-Ch-Ua-Platform" = '"Windows"',
    "Upgrade-Insecure-Requests" = "1",
    "User-Agent" = ua
  )
}
#' Create a BLS data object with diagnostics
#'
#' This is a helper function to create a list with the additional class 'bls_data_collection' containing data downloaded form the U.S. Bureau of Labor Statistics as well as diagnostic details about the download. It is used invisibly in the package to bundle information about file downloads.
#'
#' @param data The processed data (data.table/data.frame)
#' @param downloads List of download results from fread_bls()
#' @param data_type Character string describing the type of BLS data (e.g., "CES", "JOLTS", "CPS")
#' @param processing_steps Character vector describing processing steps applied
#' @return A bls_data_collection object
#' @export
create_bls_object <- function(data, downloads, data_type = "BLS", processing_steps = character(0)) {

  # Extract diagnostics from downloads
  download_diagnostics <- lapply(downloads, function(x) {
    if (inherits(x, "bls_data")) {
      return(get_bls_diagnostics(x))
    } else {
      return(NULL)
    }
  })

  # Collect all warnings
  all_warnings <- character(0)
  for (name in names(download_diagnostics)) {
    diag <- download_diagnostics[[name]]
    if (!is.null(diag) && length(diag$warnings) > 0) {
      prefixed_warnings <- paste(name, ":", diag$warnings)
      all_warnings <- c(all_warnings, prefixed_warnings)
    }
  }

  # Create summary
  summary_info <- list(
    data_type = data_type,
    files_downloaded = length(downloads),
    files_with_issues = sum(sapply(download_diagnostics, function(x) !is.null(x) && length(x$warnings) > 0)),
    total_warnings = length(all_warnings),
    processing_steps = processing_steps,
    final_dimensions = dim(data),
    download_timestamp = Sys.time()
  )

  # Create result object
  result <- list(
    data = data,
    download_diagnostics = download_diagnostics,
    warnings = all_warnings,
    summary = summary_info
  )

  class(result) <- c("bls_data_collection", "list")

  return(result)
}

#' Extract data from BLS data object
#'
#' This is a helper function to extract the data element of a 'bls_data_collection' object.
#'
#' @param bls_obj A bls_data_collection object or raw data
#' @return The data component of a 'bls_data_collection' object as a data frame.
#' @export

get_bls_data <- function(bls_obj) {
  if (inherits(bls_obj, "bls_data_collection")) {
    return(bls_obj$data)
  } else if (inherits(bls_obj, "bls_data")) {
    return(bls_obj$data)
  } else {
    # For backward compatibility
    return(bls_obj)
  }
}

#' Get download diagnostics from BLS data object
#'
#' This is a helper function to extract the download diagnostics element of a 'bls_data_collection' object.
#'
#' @param bls_obj A bls_data_collection object
#' @return List of download diagnostics from a bls_data_collection object.
#' @export
get_bls_diagnostics <- function(bls_obj) {
  if (inherits(bls_obj, "bls_data_collection")) {
    return(bls_obj$download_diagnostics)
  } else if (inherits(bls_obj, "bls_data")) {
    return(bls_obj$diagnostics)
  } else {
    return(NULL)
  }
}
#' Check and Download BLS File with Local Caching
#'
#' This function manages the downloading of files from the BLS server with a 
#' local caching layer. It uses HTTP HEAD requests to compare the server's 
#' `Content-Length` and `Last-Modified` headers with local file attributes. 
#' The file is only downloaded if it does not exist locally, or if the remote 
#' version is newer or a different size.
#'
#' @param url A character string representing the URL of the BLS file (e.g., a `.txt` or `.gz` file from download.bls.gov).
#' @param cache_dir A character string specifying the local directory to store cached files. May also be set with the  enviroment variable `BLS_CACHE_DIR`
#'   Defaults to a persistent user data directory managed by \code{tools::R_user_dir}.
#' @param verbose Logical. Defaults to FALSE.  If TRUE, returns status messages for download.
#'
#' @return A character string containing the local path to the downloaded (or cached) file.
#'
#' @details 
#' The function uses a specific set of browser-like headers to ensure compatibility 
#' with BLS server security policies. Upon a successful download, the local file's 
#' modification time is synchronized with the server's `Last-Modified` header using 
#' \code{Sys.setFileTime} to ensure accurate future comparisons.
#'
#' @importFrom httr HEAD GET add_headers headers status_code write_disk progress
#' @importFrom tools R_user_dir
#' @export
#'
#' @examples
#' \dontrun{
#' url <- "https://download.bls.gov/pub/time.series/ce/ce.data.0.AllCESSeries"
#' local_path <- smart_bls_download(url)
#' data <- data.table::fread(local_path)
#' }
smart_bls_download <- function(url, cache_dir = NULL, verbose = FALSE) {
  
  # 1. Define specific headers required by BLS servers
  bls_headers <- httr::add_headers(.headers = get_bls_headers())
  
  # 2. Establish cache directory
  if (is.null(cache_dir)) {
    cache_dir <- bls_get_cache_dir()
  }
  
  if (!dir.exists(cache_dir)) {
    dir.create(cache_dir, recursive = TRUE)
  }
  
  local_path <- file.path(cache_dir, basename(url))
  
  # 3. Fetch Remote Metadata
  response_head <- httr::HEAD(url, bls_headers)
  
  if (httr::status_code(response_head) != 200) {
    if (file.exists(local_path)) {
      warning("Could not reach BLS server; using existing cached file.")
      return(local_path)
    } else {
      stop("Could not connect to BLS server and no local cache exists. Status: ", 
           httr::status_code(response_head))
    }
  }
  
  res_headers <- httr::headers(response_head)
  
  # Correctly parse the remote modification time
  # BLS uses RFC 1123 format: "Thu, 18 Dec 2025 13:30:00 GMT"
  remote_mtime_raw <- res_headers[["last-modified"]]
  remote_mtime <- as.POSIXct(remote_mtime_raw, 
                             format = "%a, %d %b %Y %H:%M:%S GMT", 
                             tz = "GMT")
  
  remote_size <- as.numeric(res_headers[["content-length"]])
  
  # 4. Caching Logic
  needs_download <- TRUE
  
  if (file.exists(local_path)) {
    local_info <- file.info(local_path)
    
    # Check if size matches and if local file is NOT older than remote
    size_matches <- is.na(remote_size) || (local_info$size == remote_size)
    
    # Use a small tolerance (1s) to avoid floating point issues with POSIXct
    is_current <- (local_info$mtime >= (remote_mtime - 1))
    
    if (size_matches && is_current) {
      if(verbose){
      message(paste0("For ", url,": Cached local file is up to date."))
      }
      needs_download <- FALSE
    }
  }
  
  # 5. Perform download if necessary
  if (needs_download) {
    if(verbose){
    message(paste0("For ", url, ": Cached file missing or outdated: Downloading new file."))
      
    res <- httr::GET(
      url, 
      bls_headers,
      httr::write_disk(local_path, overwrite = TRUE),
      httr::progress()
    ) } else {
      res <- httr::GET(
        url, 
        bls_headers,
        httr::write_disk(local_path, overwrite = TRUE)
      )
    }
    
    # Sync file time to the server's Last-Modified time exactly
    Sys.setFileTime(local_path, remote_mtime)
  }
  
  return(local_path)
}

#' Get summary information from BLS data object
#'
#' This is a helper function to extract the summary element of a 'bls_data_collection' object. This containes the number of files downloaded, the number of files with potential warnings, and the total number of warnings.
#'
#' @param bls_obj A bls_data_collection object
#' @return List of summary information
#' @export
get_bls_summary <- function(bls_obj) {
  if (inherits(bls_obj, "bls_data_collection")) {
    return(bls_obj$summary)
  } else {
    return(NULL)
  }
}

#' Print warnings for BLS data object
#'
#' @param bls_obj A bls_data_collection object
#' @param detailed Logical. If TRUE, shows detailed diagnostics for each file
#' @param silent Logical. If TRUE, suppress console output
#'
#' @return Character vector of warnings (invisibly)
#'
#' @export
print_bls_warnings <- function(bls_obj, detailed = FALSE, silent = FALSE) {

  # Handle different object types
  if (inherits(bls_obj, "bls_data_collection")) {
    warnings <- bls_obj$warnings
    summary <- bls_obj$summary
    diagnostics <- bls_obj$download_diagnostics
  } else if (inherits(bls_obj, "bls_data")) {
    diag <- bls_obj$diagnostics
    warnings <- if (!is.null(diag)) diag$warnings else character(0)
    summary <- list(data_type = "Single File", files_downloaded = 1,
                    files_with_issues = length(warnings) > 0,
                    total_warnings = length(warnings))
    diagnostics <- list("Single File" = diag)
  } else {
    if (!silent) message("No diagnostic information available\n")
    return(invisible(character(0)))
  }

  if (length(warnings) == 0) {
    if (!silent) message("No warnings for", summary$data_type, "data download\n")
    return(invisible(character(0)))
  }

  if (!silent) {
    message(summary$data_type, "Data Download Warnings:\n")
    message(paste(rep("=", nchar(summary$data_type) + 25), collapse = ""), "\n")
    message("Total files downloaded:", summary$files_downloaded, "\n")
    message("Files with issues:", summary$files_with_issues, "\n")
    message("Total warnings:", summary$total_warnings, "\n")
    if (!is.null(summary$final_dimensions)) {
      message("Final data dimensions:", paste(summary$final_dimensions, collapse = " x "), "\n")
    }
    message("\n")

    if (detailed && length(diagnostics) > 0) {
      message("Detailed Diagnostics:\n")
      for (file_name in names(diagnostics)) {
        diag <- diagnostics[[file_name]]
        if (!is.null(diag) && length(diag$warnings) > 0) {
          message("\n", file_name, ":\n")
          message("  URL:", diag$url, "\n")
          message("  Original dimensions:", paste(diag$original_dimensions, collapse = " x "), "\n")
          message("  Final dimensions:", paste(diag$final_dimensions, collapse = " x "), "\n")
          message("  Issues:\n")
          for (warning in diag$warnings) {
            message("    -", warning, "\n")
          }
        }
      }
    } else {
      message("Summary of warnings:\n")
      for (i in seq_along(warnings)) {
        message("  ", i, ". ", warnings[i], "\n")
      }
    }

    if (!detailed && length(diagnostics) > 1) {
      message("\nRun with return_diagnostics=TRUE and print_bls_warnings(data, detailed = TRUE) for file-by-file details\n")
    }
  }

  return(invisible(warnings))
}

#' Check if BLS data object has potential issues with import.
#'
#' @param bls_obj A BLS data object
#'
#' @return Logical indicating if there were any import issues detected.
#'
#' @export
has_bls_issues <- function(bls_obj) {
  if (inherits(bls_obj, "bls_data_collection")) {
    return(length(bls_obj$warnings) > 0)
  } else if (inherits(bls_obj, "bls_data")) {
    diag <- bls_obj$diagnostics
    return(!is.null(diag) && length(diag$warnings) > 0)
  } else {
    return(FALSE)
  }
}


#' Helper function for downloading and tracking BLS files
#'
#' This function is used to pass multiple URLs at the Bureau of Labor Statistics into 'fread_bls()'
#'
#' @param urls Named character vector of URLs to download
#' @param suppress_warnings Logical. If TRUE, suppress individual download warnings
#' @param cache Logical. If TRUE, download and cache local copy of files.
#'
#' @return Named list of bls_data objects
#'
#' @export
download_bls_files <- function(urls, suppress_warnings = TRUE, cache = FALSE) {

  results <- list()

  for (name in names(urls)) {
    if (!suppress_warnings) message("Downloading", name, "...\n")

    result <- fread_bls(urls[[name]], cache = cache)

    # Check for issues
    if (has_bls_issues(result)) {
      if (!suppress_warnings) {
        print_bls_warnings(result, silent = FALSE)
      }
    }

    results[[name]] <- result
  }

  return(results)
}

#' Get the current BLSloadR Cache Directory
#' 
#' Displays the path currently being used for caching. This will check the
#' `BLS_CACHE_DIR` environment variable, falling back to the default
#' system cache directory if the variable is not set.
#'
#' @return A character string of the cache path.
#' @export
bls_get_cache_dir <- function() {
  env_path <- Sys.getenv("BLS_CACHE_DIR")
  
  if (env_path != "") {
    return(normalizePath(env_path, mustWork = FALSE))
  }
  
  # Fallback to the same logic used in smart_bls_download
  return(normalizePath(tools::R_user_dir("BLSloadR", which = "cache"), mustWork = FALSE))
}
#' Check if Global Caching is Enabled via Environment Variable
#' @keywords internal
#' @return Logical value indicating if the environment variable USE_BLS_CACHE is one of TRUE, YES, or 1
check_bls_cache_env <- function() {
  val <- Sys.getenv("USE_BLS_CACHE", unset = "FALSE")
  # Returns TRUE if user set it to "TRUE", "true", "1", or "yes"
  return(toupper(val) %in% c("TRUE", "1", "YES"))
}

#' Download Local Area Unemployment Statistics (LAUS) Data
#'
#' @description This function downloads Local Area Unemployment Statistics data from the U.S. Bureau
#'  of Labor Statistics. Due to the large size of some LAUS datasets (county and city
#'  files are >300MB), users must specify which geographic level to download. The function
#'  provides access to both seasonally adjusted and unadjusted data at various geographic levels. 
#'  Additional datasets provide comprehensive non-seasonally-adjusted data for all areas broken out 
#'  in 5-year increments
#'
#' @param geography Character string specifying the geographic level and adjustment type.
#'   Default is "state_adjusted". Valid options are:
#'   \itemize{
#'     \item "state_current_adjusted" - Current seasonally adjusted state data
#'     \item "state_unadjusted" - All historical unadjusted state data
#'     \item "state_adjusted" - All historical seasonally adjusted state data (default)
#'     \item "region_unadjusted" - Unadjusted regional and division data
#'     \item "region_adjusted" - Seasonally adjusted regional and division data
#'     \item "metro" - Metropolitan statistical area data
#'     \item "division" - Division-level data
#'     \item "micro" - Micropolitan statistical area data
#'     \item "combined" - Combined statistical area data
#'     \item "county" - County-level data (large file >300MB)
#'     \item "city" - City and town data (large file >300MB)
#'     \item "1990-1994" - Comprehensive unadjusted data for 1990-1994
#'     \item "1995-1999" - Comprehensive unadjusted data for 1995-1999
#'     \item "2000-2004" - Comprehensive unadjusted data for 2000-2004
#'     \item "2005-2009" - Comprehensive unadjusted data for 2005-2009
#'     \item "2010-2014" - Comprehensive unadjusted data for 2010-2014
#'     \item "2015-2019" - Comprehensive unadjusted data for 2015-2019
#'     \item "2020-2024" - Comprehensive unadjusted data for 2020-2024
#'     \item "2025-2029" - Comprehensive unadjusted data for 2025-2029
#'     \item "ST" - Any state two-character USPS abbreviation, plus DC and PR
#'   }
#' @param monthly_only Logical. If TRUE (default), excludes annual data (period M13)
#'   and creates a date column from year and period.
#' @param transform Logical. If TRUE (default), converts rate and ratio measures from
#'   percentages to proportions by dividing by 100. Unemployment rates will be expressed
#'   as decimals (e.g., 0.05 for 5\% unemployment) rather than as whole numbers (e.g. 5).
#' @param suppress_warnings Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
#'   for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.
#' @param return_diagnostics Logical. If TRUE, returns a bls_data_collection object
#'   with full diagnostics. If FALSE (default), returns just the data table.
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'
#' @return By default, returns a data.table with LAUS data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing LAUS data with the following key columns:
#'   \describe{
#'     \item{series_id}{BLS series identifier}
#'     \item{year}{Year of observation}
#'     \item{period}{Time period (M01-M12 for months, M13 for annual)}
#'     \item{value}{Employment statistic value (transformed if transform = TRUE)}
#'     \item{date}{Date of observation (if monthly_only = TRUE)}
#'     \item{area_text}{Geographic area name}
#'     \item{area_type_code}{Code indicating area type}
#'     \item{measure_text}{Type of measure (unemployment rate, labor force, employment, etc.)}
#'     \item{seasonal}{Seasonal adjustment status}
#'   }
#'
#' @details The function joins data from multiple BLS files:
#'   \itemize{
#'     \item Main data file (varies by geography selection)
#'     \item Series definitions (la.series)
#'     \item Area codes and names (la.area)
#'     \item Measure definitions (la.measure)
#'   }
#' @export
#' 
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @importFrom dplyr if_else
#' @importFrom stringr str_detect
#' @importFrom lubridate ym
#' 
#' @examples
#' \dontrun{
#' # Download state-level seasonally adjusted data (default operation)
#' laus_states <- get_laus()
#' 
#' # View unemployment rates by state for latest period
#' unemployment <- laus_states[grepl("unemployment rate", measure_text) & date == max(date)]
#'
#' # Download unadjusted state data
#' laus_states_unadjusted <- get_laus(geography = "state_unadjusted")
#'
#' # Get full diagnostic object if needed
#' laus_with_diagnostics <- get_laus(return_diagnostics = TRUE)
#' print_bls_warnings(laus_with_diagnostics)
#'
#' 
#' }

get_laus <- function(geography = "state_adjusted", monthly_only = TRUE, transform = TRUE, 
                     suppress_warnings = TRUE, return_diagnostics = FALSE, cache = check_bls_cache_env()) {
  
  # Define the URL mapping
  laus_urls <- list(
    "state_current_adjusted" = "https://download.bls.gov/pub/time.series/la/la.data.1.CurrentS",
    "state_unadjusted" = "https://download.bls.gov/pub/time.series/la/la.data.2.AllStatesU",
    "state_adjusted" = "https://download.bls.gov/pub/time.series/la/la.data.3.AllStatesS",
    "region_unadjusted" = "https://download.bls.gov/pub/time.series/la/la.data.4.RegionDivisionU",
    "region_adjusted" = "https://download.bls.gov/pub/time.series/la/la.data.5.RegionDivisionS",
    "metro" = "https://download.bls.gov/pub/time.series/la/la.data.60.Metro",
    "division" = "https://download.bls.gov/pub/time.series/la/la.data.61.Division",
    "micro" = "https://download.bls.gov/pub/time.series/la/la.data.62.Micro",
    "combined" = "https://download.bls.gov/pub/time.series/la/la.data.63.Combined",
    "county" = "https://download.bls.gov/pub/time.series/la/la.data.64.County",
    "city" = "https://download.bls.gov/pub/time.series/la/la.data.65.City",
    "2025-2029" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU25-29",
    "2020-2024" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU20-24",
    "2015-2019" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU15-19",
    "2010-2014" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU10-14",
    "2005-2009" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU05-09",
    "2000-2004" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU00-04",
    "1995-1999" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU95-99",
    "1990-1994" = "https://download.bls.gov/pub/time.series/la/la.data.0.CurrentU90-94",
    'AR' = 'https://download.bls.gov/pub/time.series/la/la.data.10.Arkansas',
    'CA' = 'https://download.bls.gov/pub/time.series/la/la.data.11.California',
    'CO' = 'https://download.bls.gov/pub/time.series/la/la.data.12.Colorado',
    'CT' = 'https://download.bls.gov/pub/time.series/la/la.data.13.Connecticut',
    'DE' = 'https://download.bls.gov/pub/time.series/la/la.data.14.Delaware',
    'DC' = 'https://download.bls.gov/pub/time.series/la/la.data.15.DC',
    'FL' = 'https://download.bls.gov/pub/time.series/la/la.data.16.Florida',
    'GA' = 'https://download.bls.gov/pub/time.series/la/la.data.17.Georgia',
    'HI' = 'https://download.bls.gov/pub/time.series/la/la.data.18.Hawaii',
    'ID' = 'https://download.bls.gov/pub/time.series/la/la.data.19.Idaho',
    'IL' = 'https://download.bls.gov/pub/time.series/la/la.data.20.Illinois',
    'IN' = 'https://download.bls.gov/pub/time.series/la/la.data.21.Indiana',
    'IA' = 'https://download.bls.gov/pub/time.series/la/la.data.22.Iowa',
    'KS' = 'https://download.bls.gov/pub/time.series/la/la.data.23.Kansas',
    'KY' = 'https://download.bls.gov/pub/time.series/la/la.data.24.Kentucky',
    'LA' = 'https://download.bls.gov/pub/time.series/la/la.data.25.Louisiana',
    'ME' = 'https://download.bls.gov/pub/time.series/la/la.data.26.Maine',
    'MD' = 'https://download.bls.gov/pub/time.series/la/la.data.27.Maryland',
    'MA' = 'https://download.bls.gov/pub/time.series/la/la.data.28.Massachusetts',
    'MI' = 'https://download.bls.gov/pub/time.series/la/la.data.29.Michigan',
    'MN' = 'https://download.bls.gov/pub/time.series/la/la.data.30.Minnesota',
    'MS' = 'https://download.bls.gov/pub/time.series/la/la.data.31.Mississippi',
    'MO' = 'https://download.bls.gov/pub/time.series/la/la.data.32.Missouri',
    'MT' = 'https://download.bls.gov/pub/time.series/la/la.data.33.Montana',
    'NE' = 'https://download.bls.gov/pub/time.series/la/la.data.34.Nebraska',
    'NV' = 'https://download.bls.gov/pub/time.series/la/la.data.35.Nevada',
    'NH' = 'https://download.bls.gov/pub/time.series/la/la.data.36.NewHampshire',
    'NJ' = 'https://download.bls.gov/pub/time.series/la/la.data.37.NewJersey',
    'NM' = 'https://download.bls.gov/pub/time.series/la/la.data.38.NewMexico',
    'NY' = 'https://download.bls.gov/pub/time.series/la/la.data.39.NewYork',
    'NC' = 'https://download.bls.gov/pub/time.series/la/la.data.40.NorthCarolina',
    'ND' = 'https://download.bls.gov/pub/time.series/la/la.data.41.NorthDakota',
    'OH' = 'https://download.bls.gov/pub/time.series/la/la.data.42.Ohio',
    'OK' = 'https://download.bls.gov/pub/time.series/la/la.data.43.Oklahoma',
    'OR' = 'https://download.bls.gov/pub/time.series/la/la.data.44.Oregon',
    'PA' = 'https://download.bls.gov/pub/time.series/la/la.data.45.Pennsylvania',
    'PR' = 'https://download.bls.gov/pub/time.series/la/la.data.46.PuertoRico',
    'RI' = 'https://download.bls.gov/pub/time.series/la/la.data.47.RhodeIsland',
    'SC' = 'https://download.bls.gov/pub/time.series/la/la.data.48.SouthCarolina',
    'SD' = 'https://download.bls.gov/pub/time.series/la/la.data.49.SouthDakota',
    'TN' = 'https://download.bls.gov/pub/time.series/la/la.data.50.Tennessee',
    'TX' = 'https://download.bls.gov/pub/time.series/la/la.data.51.Texas',
    'UT' = 'https://download.bls.gov/pub/time.series/la/la.data.52.Utah',
    'VT' = 'https://download.bls.gov/pub/time.series/la/la.data.53.Vermont',
    'VA' = 'https://download.bls.gov/pub/time.series/la/la.data.54.Virginia',
    'WA' = 'https://download.bls.gov/pub/time.series/la/la.data.56.Washington',
    'WV' = 'https://download.bls.gov/pub/time.series/la/la.data.57.WestVirginia',
    'WI' = 'https://download.bls.gov/pub/time.series/la/la.data.58.Wisconsin',
    'WY' = 'https://download.bls.gov/pub/time.series/la/la.data.59.Wyoming',
    'AL' = 'https://download.bls.gov/pub/time.series/la/la.data.7.Alabama',
    'AK' = 'https://download.bls.gov/pub/time.series/la/la.data.8.Alaska',
    'AZ' = 'https://download.bls.gov/pub/time.series/la/la.data.9.Arizona'
  )
  
  # Validate geography argument
  if (!geography %in% names(laus_urls)) {
    stop("Invalid geography. Valid options are: ",
         paste(names(laus_urls), collapse = ", "))
  }
  
  # Warn about large files
  if (geography %in% c("city", "county")) {
    message("Warning: ", geography, " data file is very large (>300MB).")
  }
  
  # Define all URLs we need to download
  download_urls <- c(
    "data" = laus_urls[[geography]],
    "series" = "https://download.bls.gov/pub/time.series/la/la.series",
    "area" = "https://download.bls.gov/pub/time.series/la/la.area",
    "measure" = "https://download.bls.gov/pub/time.series/la/la.measure"
  )
  
  # Download all files
  downloads <- download_bls_files(download_urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Exit function if download failed.
  if(is.null(downloads) | length(downloads) == 0){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
  }
  
  # Extract data from downloads
  laus_import <- get_bls_data(downloads$data)
  laus_series <- get_bls_data(downloads$series)
  laus_area <- get_bls_data(downloads$area)
  laus_measure <- get_bls_data(downloads$measure)
  
  # Join all the data together
  laus <- laus_import |>
    dplyr::select(-footnote_codes) |>
    dplyr::left_join(laus_series |> dplyr::select(-footnote_codes), by = c("series_id")) |>
    dplyr::left_join(laus_area, by = c("area_code", "area_type_code")) |>
    dplyr::left_join(laus_measure, by = "measure_code") |>
    dplyr::mutate(value = as.numeric(value)) |>
    dplyr::filter(!is.na(value)) |>
    dplyr::select(-c(display_level:sort_sequence)) |>
    dplyr::select(-c(series_title:end_period))
  
  # Track processing steps
  processing_steps <- c(
    "Joined series, area, and measure metadata",
    "Converted values to numeric",
    "Removed rows with missing values"
  )
  
  # Handle monthly filtering and date creation
  if (monthly_only) {
    laus <- laus |>
      dplyr::filter(period != "M13") |>
      dplyr::mutate(date = lubridate::ym(paste(as.character(year), substr(period, 2, 3), sep = "-")))
    
    processing_steps <- c(processing_steps, "Filtered to monthly data only", "Created date column")
  }
  
  # Handle transformation
  if (transform) {
    laus <- laus |>
      dplyr::mutate(value = if_else(str_detect(measure_text, "rate")|str_detect(measure_text, "ratio"), value/100, value))
    
    processing_steps <- c(processing_steps, "Converted rates and ratios to proportions")
  }
  
  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = laus,
    downloads = downloads,
    data_type = "LAUS",
    processing_steps = processing_steps
  )
  
  # Show warnings if any issues were detected
  if (has_bls_issues(bls_collection) && !suppress_warnings) {
    print_bls_warnings(bls_collection)
  }
  
  # Return either the collection object or just the data
  if (return_diagnostics) {
    return(bls_collection)
  } else {
    return(laus)
  }
}

#' Generic BLS Dataset Download
#' 
#' This function generalizes a method to download all BLS data for a given time series database.
#' These files are accessed from https://download.bls.gov/pub/time.series/ and several datasets
#' are available. A summary of an identified database can be generated using the `bls_overiew()`
#' function. When multiple potential data files exist (common in large data sets), the function
#' will prompt for an input of which file to use.
#' 
#' @param database_code This is the two digit character identifier for the desired database.
#'   Some Valid options are:
#'   \itemize{
#'     \item "ce" - National Current Employment Statistics Data
#'     \item "sm" - State and Metro area Current Employment Statistics Data
#'     \item "mp" - Major Sector Total Factor Productivity
#'     \item "ci" - Employment Cost Index
#'     \item "eb" - Employee Benefits Survey
#'   }
#'   
#' @param return_full This argument defaults to FALSE. If set to TRUE it will return
#'   a list of the elements of data retrieved from the BLS separating the data, series, and
#'   mapping values downloaded.
#'  
#' @param simplify_table This parameter defaults to TRUE. When TRUE it will remove all
#'  columns from the date with "_code" in the column name, as well as a series of internal
#'  identifiers which provide general information about the series but which are not needed for
#'  performing time series analysis. This parameter also converts the column "value" to numeric
#'  and generates a date column from the year and period columns in the data.
#'  
#' @param suppress_warnings Logical. If TRUE, suppress individual download warnings during processing.
#' 
#' @param which_data Character string or NULL. Defaults to NULL.
#'   \itemize{
#'     \item "all" - Automatically selects the data file containing ".1.All" (e.g., "bd.data.1.AllItems" or "le.data.1.AllData").
#'     \item "current" - Automatically selects the data file containing "Current" (e.g., "ce.data.0.Current").
#'     \item NULL - Default behavior. Prompts the user to select a file if multiple exist, or selects the single available file.
#'   }
#'   If the requested pattern is not found, the function falls back to the default behavior, prompting the user to select a file.
#'   
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'  
#' @returns This function will return either a bls_data_collection object (if return_full is FALSE or not provided)
#'  or a named list of the returned data including the bls_data_collection object.
#'  
#' @export
#' @importFrom data.table data.table
#' @importFrom data.table fcase
#' @importFrom data.table rbindlist
#' @importFrom data.table :=
#' @importFrom rvest read_html html_elements html_attr
#' @importFrom httr GET add_headers stop_for_status content
#' @importFrom dplyr left_join
#' @importFrom stats setNames
#' @importFrom utils head
#' 
#' @examples
#' \dontrun{
#' # Import All Data
#' fm_import <- load_bls_dataset("fm", which_data = "all")
#' 
#' # Get $data element
#' fm_data <- fm_import$data
#' 
#' # Filter to a Series
#' # Families with Children Under 6 and No Employed Parent
#' 
#' u6_no_emp <- fm_data |> 
#'   dplyr::filter(series_title == "Total families with children under 6 - with no parent employed") |> 
#'   dplyr:: select(year, value, fchld_text, fhlf_text, tdat_text)
#' 
#' 
#' head(u6_no_emp)
#' }
#' 
#' \dontrun{
#' # Examples requiring manual intervention in the console
#' # Download Employer Cost Index Data
#' cost_index <- load_bls_dataset("ci")
#'
#' # Download separated data, series, and mapping columns
#' benefits <- load_bls_dataset("eb", return_full = TRUE)
#'
#' # Download data without removing excess columns and value conversions
#' productivity <- load_bls_dataset("mp", simplify_table = FALSE)
#' 
#' # Check for download issues
#' if (has_bls_issues(cost_index)) {
#'   print_bls_warnings(cost_index, detailed = TRUE)
#' }
#' 
#' }

load_bls_dataset <- function(database_code, return_full = FALSE, simplify_table = TRUE, suppress_warnings = FALSE, which_data = NULL, cache = check_bls_cache_env()) {
  
  # Validate inputs
  if (!is.character(database_code) || length(database_code) != 1) {
    stop("database_code must be a single character string")
  }
  
  # Validate which_data input
  if (!is.null(which_data) && (!is.character(which_data) || length(which_data) != 1 || !which_data %in% c("all", "current"))) {
    stop("which_data must be NULL, 'all', or 'current'")
  }
  
  base_url <- sprintf("https://download.bls.gov/pub/time.series/%s/", database_code)
  
  # Function to scrape directory contents with proper headers
  get_directory_files <- function(url, prefix) {
    tryCatch({
      # Set up headers to avoid 403 errors
      headers <- get_bls_headers()
      
      # Make request with headers
      response <- httr::GET(url, httr::add_headers(.headers = headers))
      httr::stop_for_status(response)
      
      # Exit function if download failed.
      if(is.null(downloads)){
        stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
      }
      
      # Parse HTML content
      page <- rvest::read_html(httr::content(response, as = "text"))
      links <- rvest::html_elements(page, "a")
      hrefs <- rvest::html_attr(links, "href")
      
      # Extract just the filename from the full path
      # hrefs will be like "/pub/time.series/ce/ce.data.0.AllCESSeries"
      filenames <- basename(hrefs)
      
      # Filter for files that start with the prefix and exclude unwanted extensions
      valid_files <- filenames[
        grepl(paste0("^", prefix, "\\."), filenames) & 
          !grepl("\\.(contacts|txt|footnote)$", filenames) &
          !is.na(filenames) &
          filenames != ""
      ]
      
      return(valid_files)
    }, error = function(e) {
      stop("Could not access BLS directory: ", url, "\nError: ", e$message)
    })
  }
  
  # Get all valid files from the directory
  file_names <- get_directory_files(base_url, database_code)
  
  if (length(file_names) == 0) {
    stop("No valid files found in the BLS database directory for code: ", database_code)
  }
  
  # Create file table and classify by pattern
  file_table <- data.table(file_name = file_names)
  file_table[, file_type := fcase(
    grepl("\\.data\\.", file_name), "data",
    grepl("\\.series($|\\.)", file_name), "series",
    grepl("\\.aspect($|\\.)", file_name), "aspect",
    default = "mapping"
  )]
  
  # Identify files
  mapping_files <- file_table[file_type == "mapping", file_name]
  data_files <- file_table[file_type == "data", file_name]
  series_file <- file_table[file_type == "series", file_name]
  aspect_files <- file_table[file_type == "aspect", file_name]
  
  if (length(series_file) == 0) {
    stop("Could not find a series file in the BLS database directory.")
  }
  
  # Handle multiple series files (prompt user to choose)
  if (length(series_file) > 1) {
    message("Multiple series files found. Please select a file to load:\n")
    for (i in seq_along(series_file)) {
      message(i, ": ", series_file[i], "\n")
    }
    
    # Get user input for series file selection
    selected_series_index <- as.integer(readline(prompt = "Enter the number of the series file you want to load: "))
    
    # Validate the input
    if (is.na(selected_series_index) || selected_series_index < 1 || selected_series_index > length(series_file)) {
      stop("Invalid selection. Please run the function again and enter a valid number.")
    }
    
    # Get the selected series file name
    series_file <- series_file[selected_series_index]
    message("Loading series file:", series_file, "\n")
    
  } else if (length(series_file) == 1) {
    message("Loading series file:", series_file, "\n")
  }
  
  # --- Logic for data file selection ---
  selected_data_file <- NULL
  
  # 1. Attempt Auto-selection if requested
  if (!is.null(which_data)) {
    pattern <- NULL
    if (which_data == "all") {
      pattern <- ".1.All"
    } else if (which_data == "current") {
      pattern <- "Current"
    }
    
    if (!is.null(pattern)) {
      # Look for files containing the pattern (case insensitive)
      # We check if pattern is in the filename
      matches <- grep(pattern, data_files, ignore.case = TRUE, value = TRUE)
      
      if (length(matches) > 0) {
        # Use the first match
        selected_data_file <- matches[1]
        message("Auto-selected '", which_data, "' file: ", selected_data_file, "\n")
      } else {
        message("Warning: No file matching '", pattern, "' found. Falling back to manual selection/default.\n")
      }
    }
  }
  
  # 2. Fallback to Prompt or Single File Logic if not auto-selected
  if (is.null(selected_data_file)) {
    if (length(data_files) > 1) {
      # If there are multiple data files and auto-select didn't find a match, prompt the user
      message("Multiple data files found. Please select a file to load:\n")
      for (i in seq_along(data_files)) {
        message(i, ": ", data_files[i], "\n")
      }
      
      # Get user input for file selection
      selected_index <- as.integer(readline(prompt = "Enter the number of the file you want to load: "))
      
      # Validate the input
      if (is.na(selected_index) || selected_index < 1 || selected_index > length(data_files)) {
        stop("Invalid selection. Please run the function again and enter a valid number.")
      }
      
      # Get the selected file name
      selected_data_file <- data_files[selected_index]
      message("Loading:", selected_data_file, "\n")
      
    } else if (length(data_files) == 1) {
      # If there is only one data file, use it directly
      selected_data_file <- data_files[1]
      message("Loading:", selected_data_file, "\n")
    } else {
      stop("No data files found in the BLS database directory.")
    }
  }
  
  # --- Logic for aspect file selection ---
  selected_aspect_file <- NULL
  if (length(aspect_files) > 1) {
    message("Multiple aspect files found. Please select a file to load:\n")
    for (i in seq_along(aspect_files)) {
      message(i, ": ", aspect_files[i], "\n")
    }
    
    # Get user input for aspect file selection
    selected_aspect_index <- as.integer(readline(prompt = "Enter the number of the aspect file you want to load: "))
    
    # Validate the input
    if (is.na(selected_aspect_index) || selected_aspect_index < 1 || selected_aspect_index > length(aspect_files)) {
      stop("Invalid selection. Please run the function again and enter a valid number.")
    }
    
    # Get the selected aspect file name
    selected_aspect_file <- aspect_files[selected_aspect_index]
    message("Loading aspect file:", selected_aspect_file, "\n")
    
  } else if (length(aspect_files) == 1) {
    # If there is only one aspect file, use it directly
    selected_aspect_file <- aspect_files[1]
    message("Loading aspect file:", selected_aspect_file, "\n")
  } else if (length(aspect_files) == 0) {
    if (!suppress_warnings) message("No aspect files found in the BLS database directory.")
  }
  
  # Create URLs for downloading
  urls <- c(
    setNames(paste0(base_url, selected_data_file), selected_data_file),
    setNames(paste0(base_url, series_file), series_file)
  )
  
  # Add aspect file URL if it exists
  if (!is.null(selected_aspect_file)) {
    aspect_url <- setNames(paste0(base_url, selected_aspect_file), selected_aspect_file)
    urls <- c(urls, aspect_url)
  }
  
  # Add mapping file URLs
  if (length(mapping_files) > 0) {
    mapping_urls <- setNames(paste0(base_url, mapping_files), mapping_files)
    urls <- c(urls, mapping_urls)
  }
  
  # Download all files using the new system
  downloads <- download_bls_files(urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Extract data from downloads
  data_dt <- get_bls_data(downloads[[selected_data_file]])
  series_dt <- get_bls_data(downloads[[series_file]])
  
  # Remove unwanted columns from all files
  columns_to_remove <- c("display_level", "sort_sequence", "selectable", "footnote_codes")
  data_dt <- data_dt |> dplyr::select(-tidyselect::any_of(columns_to_remove))
  series_dt <- series_dt |> dplyr::select(-tidyselect::any_of(columns_to_remove))
  
  # Track processing steps
  processing_steps <- character(0)
  
  # STEP 1: Join data to series first to get lookup codes
  if (!suppress_warnings) message("Joining data to series file...")
  full_dt <- left_join(data_dt, series_dt, by = "series_id")
  processing_steps <- c(processing_steps, "joined_data_to_series")
  
  # STEP 2: Join aspect file if it exists (after series, before mapping files)
  if (!is.null(selected_aspect_file) && selected_aspect_file %in% names(downloads)) {
    tryCatch({
      if (!suppress_warnings) message("Joining aspect file...")
      aspect_dt <- get_bls_data(downloads[[selected_aspect_file]])
      
      # Remove unwanted columns from aspect file
      aspect_dt <- aspect_dt |> dplyr::select(-tidyselect::any_of(columns_to_remove))
      
      # Rename the value column in aspect file to aspect_value to avoid conflicts
      if ("value" %in% names(aspect_dt)) {
        aspect_dt <- aspect_dt |>
          dplyr::rename(aspect_value = value)
      }
      
      # Join aspect file on series_id, year, and period
      join_cols <- c("series_id", "year", "period")
      available_join_cols <- intersect(join_cols, names(aspect_dt))
      
      if (length(available_join_cols) > 0) {
        full_dt <- left_join(full_dt, aspect_dt, by = available_join_cols)
        processing_steps <- c(processing_steps, "joined_aspect_file")
        if (!suppress_warnings) message("Aspect file joined successfully on: ", paste(available_join_cols, collapse = ", "))
      } else {
        if (!suppress_warnings) message("Warning: Could not join aspect file - no matching columns found")
      }
      
    }, error = function(e) {
      if (!suppress_warnings) message("Error processing aspect file ", selected_aspect_file, ": ", e$message)
    })
  }
  
  # STEP 3: Now join mapping files to the combined table
  for (map_file in mapping_files) {
    if (map_file %in% names(downloads)) {
      tryCatch({
        map_dt <- get_bls_data(downloads[[map_file]])
        
        # Remove unwanted columns from mapping file
        map_dt <- map_dt |> dplyr::select(-tidyselect::any_of(columns_to_remove))
        
        if (ncol(map_dt) == 2) {
          
          # For mapping files with exactly 2 columns, assume first is join column
          join_col <- names(map_dt)[1]
          
          if (join_col %in% names(full_dt)) {
            if (!suppress_warnings) message("Joining mapping file ", map_file, " on column: ", join_col)
            full_dt <- left_join(full_dt, map_dt, by = join_col)
            processing_steps <- c(processing_steps, paste0("joined_mapping_", gsub("\\.", "_", map_file)))
          } else {
            if (!suppress_warnings) message("Skipping mapping file ", map_file, " - join column '", join_col, "' not found in data")
          }
          
        } else {
          
          # For mapping files with >2 columns, use all except last as potential join columns
          potential_join_cols <- names(map_dt)[1:(ncol(map_dt) - 1)]
          join_cols <- intersect(potential_join_cols, names(full_dt))
          
          if (length(join_cols) > 0) {
            if (!suppress_warnings) message("Joining mapping file ", map_file, " on column(s): ", paste(join_cols, collapse = ", "))
            full_dt <- left_join(full_dt, map_dt, by = join_cols)
            processing_steps <- c(processing_steps, paste0("joined_mapping_", gsub("\\.", "_", map_file)))
          } else {
            if (!suppress_warnings) message("Skipping mapping file ", map_file, " - no join columns found in data")
          }
          
        }
        
      }, error = function(e) {
        if (!suppress_warnings) message("Error processing mapping file ", map_file, ": ", e$message)
      })
    }
  }
  
  # STEP 4: Apply table simplification if requested
  if (simplify_table) {
    if (!suppress_warnings) message("Simplifying table structure...")
    
    full_dt <- full_dt |>
      dplyr::mutate(value = as.numeric(value),
                    period_type_code = substr(period,1,1),
                    date = case_when(
                      period %in% c("M13", "Q05") ~ lubridate::ym(paste0(year,"-01")),
                      period_type_code == "Q" ~ lubridate::yq(paste(year, "Q", substr(period,3,3))),
                      TRUE ~ lubridate::ym(paste0(year,period))
                    )
      ) |>
      dplyr::select(-tidyselect::contains("_code"))
    
    processing_steps <- c(processing_steps, "simplified_table")
  }
  
  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = full_dt,
    downloads = downloads,
    data_type = paste0("BLS-", toupper(database_code)),
    processing_steps = processing_steps
  )
  
  # Print summary unless suppressed
  if (!suppress_warnings) {
    if (has_bls_issues(bls_collection)) {
      message("\n")
      print_bls_warnings(bls_collection, detailed = FALSE)
      message("\nUse print_bls_warnings(result, detailed = TRUE) for detailed diagnostics\n")
    } else {
      message("\nDownload completed successfully with no issues detected.\n")
    }
  }
  
  # Return based on return_full parameter
  if (return_full) {
    return(list(
      bls_collection = bls_collection,
      full_file = get_bls_data(bls_collection),
      data = data_dt,
      series = series_dt,
      aspect = if (!is.null(selected_aspect_file)) get_bls_data(downloads[[selected_aspect_file]]) else NULL,
      mapping_files = mapping_files,
      file_table = file_table,
      downloads = downloads
    ))
  } else {
    return(bls_collection)
  }
}

#' Get QCEW Data Slices
#'
#' This function pulls data from the BLS QCEW Open Data Access CSV Data Slices.
#' It iterates over specified years and quarters (or annual data) to retrieve
#' industry-specific or area-specific data tables and merges them into a single data.table.
#' Optionally, it joins internal package lookup tables for industry and area descriptions.
#'
#' @param period_type Character. Either "quarter" or "year". Defaults to "quarter".
#' @param year_start Numeric. The first year to retrieve data for. Defaults to the year
#'   of the date 6 months prior to the current system date.
#' @param year_end Numeric. The last year to retrieve data for. Defaults to the year
#'   of the date 6 months prior to the current system date.
#' @param industry_code Character. The NAICS industry code (e.g., "10", "31-33").
#'   Constructs a URL for an Industry Data Slice. Mutually exclusive with `area_code`.
#' @param area_code Character. The QCEW area code (e.g., "US000", "32000", "C2982").
#'   Constructs a URL for an Area Data Slice. Mutually exclusive with `industry_code`.
#' @param add_lookups Logical. If \code{TRUE}, joins the package's \code{ind_lookup} and
#'   \code{area_lookup} tables to the results to provide descriptive labels. Defaults to \code{TRUE}.
#' @param silently Logical. If \code{TRUE}, suppresses status messages about the URLs being accessed.
#'   Defaults to \code{FALSE}.
#'
#' @return A combined data.table containing the requested QCEW data, optionally merged with lookup columns
#'   and a calculated \code{date} column.  The data layout is different for quarterly or annual data files set by the `period_type` argument.
#'   
#'   For Quarterly files:
#'   \itemize{
#'    \item area_fips - Character. Area code of row.  Included `area_lookup` data file contains mapping information.
#'    \item industry_code - Character. NAICS, Supersector, Cluster, or Total All Industries code. Numeric characters as a string to preserve examining the structure heirarchy.
#'    \item own_code - Integer.  Values of 0-5 to designate ownership. See definitions at \url{https://www.bls.gov/cew/classifications/ownerships/ownership-titles.htm}
#'    \item agglvl_code - Integer.   Two digit code identifying the level of aggregation.  See definitons at \url{https://www.bls.gov/cew/classifications/aggregation/agg-level-titles.htm}
#'    \item size_code - Integer. Single-digit code representing the size of establishments. See definitions at \url{https://www.bls.gov/cew/classifications/size/size-titles.htm}
#'    \item year Integer.  Four-digit calendar year for the returned data.
#'    \item qtr Integer. The calendar quarter of the data. 
#'    \item disclosure_code Character.  Values are either a blank string on "N".  Values of N do not disclose employment or wages to maintain confidentiality.
#'    \item qtryly_estabs Integer.  The number of business establishments (worksites) for the industry in the area in the quarter.
#'    \item month1_emplvl Integer. Employment in the first month of the quarter (January, April, July, or October).
#'    \item month2_emplvl Integer. Employment in the second month of the quarter (February, May, August, November).
#'    \item month3_emplvl Integer. Employment in the third month of the quarter (March, June, September, December).
#'    \item total_qtrly_wages Ingeger64. Total wages paid during the quarter.
#'    \item taxable_qtrly_wages Ingeger64. Wages subject to unemployment insurance (UI) taxes during the quarter.  Note - wages subject to UI vary by state and will follow different seasonal patterns as a result.
#'    \item qtrly_contributions Integer. UI taxes (Contributions) paid by employers for this quarter. Note - UI tax policy varies by state.
#'    \item avg_wkly_wage Integer. Average weekly wage during the quarter (Total wages divided by average employment, divided by 13).
#'   
#'    \item lq_disclosure_code Character.  Blank or "N".  Values of "N" will suppress location quotient data for confidentiality.
#'    \item lq_qtrly_estabs Numeric. Location quotient of establishments relative to the U.S.
#'    \item lq_month1_emplvl Numeric. Location quotient of month 1 employment relative to the U.S.
#'    \item lq_month2_emplvl Numeric. Location quotient of month 2 employment relative to the U.S.
#'    \item lq_month3_emplvl Numeric. Location quotient of month 3 employment relative to the U.S.
#'    \item lq_total_qtrly_wages Numeric. Location quotient of total wages relative to the U.S.
#'    \item lq_taxable_qtrly_wages Numeric. Location quotient of taxable quarterly wages relative to the U.S.
#'    \item lq_qtrly_contributions Numeric. Location quotient of quarterly UI taxes paid relative to the U.S.
#'    \item lq_avg_wkly_wage Numeric. Location quotient of average weekly wages relative to the U.S.
#'    
#'    \item oty_disclosure_code Character.  Blank or "N".  Values of "N" will suppress over-the-year data for confidentiality.
#'    \item oty_qtrly_estabs_chg Numeric. Over-the-year change in establishments.
#'    \item oty_qtrly_estabs_pct_chg Numeric. Over-the-year percent change in establishments.
#'    \item oty_month1_emplvl_chg Numeric. Over-the-year change in month 1 employment.
#'    \item oty_month1_emplvl_pct_chg Numeric. Over-the-year percent change in month 1 employment.
#'    \item oty_month2_emplvl_chg Numeric. Over-the-year change in month 2 employment.
#'    \item oty_month2_emplvl_pct_chg Numeric. Over-the-year percent change in month 2 employment.
#'    \item oty_month3_emplvl_chg Numeric. Over-the-year change in month 3 employment.
#'    \item oty_month3_emplvl_pct_chg Numeric. Over-the-year percent change in month 3 employment.
#'    \item oty_total_qtrly_wages_chg Numeric. Over-the-year change in total wages.
#'    \item oty_total_qtrly_wages_pct_chg Numeric. Over-the-year percent change in total wages.
#'    \item oty_taxable_qtrly_wages_chg Numeric. Over-the-year change in taxable quarterly wages.
#'    \item oty_taxable_qtrly_wages_pct_chg Numeric. Over-the-year percent change in taxable quarterly wages.
#'    \item oty_qtrly_contributions_chg Numeric. Over-the-year change in quarterly UI taxes paid.
#'    \item oty_qtrly_contributions_pct_chg Numeric. Over-the-year percent change in quarterly UI taxes paid.
#'    \item oty_avg_wkly_wage_chg Numeric. Over-the-year change in average weekly wages.
#'    \item oty_avg_wkly_wage_pct_chg Numeric. Over-the-year percent change in average weekly wages.
#'    
#'    \item date Date. Calculated calendar date based on year and quarter.  Reflects first day of the quarter.
#'    \item industry_title Character. Added based on industry_code
#'    \item ind_level Character. Description of the level of aggregation based on the industry_code.
#'    \item naics_2d Character. First two characters in the industry_code, useful for identifying industries.
#'    \item sector Character. Similar to naics_2d, but for industries like Manufacturing which have multiple two digit NAICS codes, this will span those groupings, for example "31-33"
#'    \item vintage_start. Integer. Calendar year of the earliest vintage for this industry_code. NAICS codes are updated every 5 years.  When using this industry codes from before this date, these titles may not exist or may be incorrect.
#'    \item vintage_end. Integer. Calendar year of the last year this industry code was used.  Years after this point should not contain this industry code.  Set to 3000 for current data.
#'    \item area_title Character.  Area description based on area_fips as provided by the BLS.
#'    \item area_type Character. Description of the type of area based on the area_title. More consistent naming and grouping than BLS data.
#'    \item stfips Character. The two-digit FIPS code of the state containing a given area. Set to "00" for multi-state regions.
#'    \item specified_region. Either a two-character US Postal Service abbreviation for the state containing an area or a hyphenated list of such codes for multi-state areas.
#'   }
#'   
#'   For Annual files:
#'   \itemize{
#'    \item area_fips - Character. Area code of row.  Included `area_lookup` data file contains mapping information.
#'    \item industry_code - Character. NAICS, Supersector, Cluster, or Total All Industries code. Numeric characters as a string to preserve examining the structure heirarchy.
#'    \item own_code - Integer.  Values of 0-5 to designate ownership. See definitions at \url{https://www.bls.gov/cew/classifications/ownerships/ownership-titles.htm}
#'    \item agglvl_code - Integer.   Two digit code identifying the level of aggregation.  See definitons at \url{https://www.bls.gov/cew/classifications/aggregation/agg-level-titles.htm}
#'    \item size_code - Integer. Single-digit code representing the size of establishments. See definitions at \url{https://www.bls.gov/cew/classifications/size/size-titles.htm}
#'    \item year Integer.  Four-digit calendar year for the returned data.
#'    \item qtr Character. Set to "A" to represent annual data.
#'    \item disclosure_code Character.  Values are either a blank string on "N".  Values of N do not disclose employment or wages to maintain confidentiality.
#'    \item annual_avg_estabs Integer.  The average number of business establishments (worksites) for the industry in the area for the year.
#'    \item annual_avg_emplvl Integer. The average monthly employment level in a given year.
#'    \item total_annual_wages Ingeger64. Total wages paid during the year.
#'    \item taxable_annual_wages Ingeger64. Wages subject to unemployment insurance (UI) taxes during the year.  Note - wages subject to UI vary by state and will follow different seasonal patterns as a result.
#'    \item annual_contributions Integer. UI taxes (Contributions) paid by employers for this year. Note - UI tax policy varies by state.
#'    \item annual_avg_wkly_wage Integer. Average weekly wage during the year (Total wages divided by average employment, divided by 52).
#'    \item avg_annual_pay Integer. Average annual pay during the year.
#'   
#'    \item lq_disclosure_code Character.  Blank or "N".  Values of "N" will suppress location quotient data for confidentiality.
#'    \item lq_annual_avg_estabs Numeric. Location quotient of establishments relative to the U.S.
#'    \item lq_annual_avg_emplvl Numeric. Location quotient of annual employment relative to the U.S.
#'    \item lq_total_annual_wages Numeric. Location quotient of total wages relative to the U.S.
#'    \item lq_taxable_annual_wages Numeric. Location quotient of taxable annual wages relative to the U.S.
#'    \item lq_annual_contributions Numeric. Location quotient of annual UI taxes paid relative to the U.S.
#'    \item lq_annual_avg_wkly_wage Numeric. Location quotient of average weekly wages relative to the U.S.
#'    \item lq_avg_annual_pay Numeric. Location quotient of average annual pay relative to the U.S.
#'    
#'    \item oty_disclosure_code Character.  Blank or "N".  Values of "N" will suppress over-the-year data for confidentiality.
#'    \item oty_annual_avg_estabs_chg Integer. Over-the-year change in establishments.
#'    \item oty_annual_avg_estabs_pct_chg Numeric. Over-the-year percent change in establishments.
#'    \item oty_annual_avg_emplvl_chg Integer. Over-the-year change in average annual employment.
#'    \item oty_annual_avg_emplvl_pct_chg Numeric. Over-the-year percent change in average annual employment.
#'    \item oty_total_annual_wages_chg Integer. Over-the-year change in total wages.
#'    \item oty_total_annual_wages_pct_chg Numeric. Over-the-year percent change in total wages.
#'    \item oty_taxable_annual_wages_chg Integer. Over-the-year change in taxable annual wages.
#'    \item oty_taxable_annual_wages_pct_chg Numeric. Over-the-year percent change in taxable annual wages.
#'    \item oty_annual_contributions_chg Integer. Over-the-year change in annual UI taxes paid.
#'    \item oty_annual_contributions_pct_chg Numeric. Over-the-year percent change in annual UI taxes paid.
#'    \item oty_annual_avg_wkly_wage_chg Integer. Over-the-year change in average weekly wages.
#'    \item oty_annual_avg_wkly_wage_pct_chg Numeric. Over-the-year percent change in average weekly wages.
#'    \item oty_avg_annual_pay_chg Integer. Over-the-year change in average annual pay.
#'    \item oty_avg_annual_pay_pct_chg Numeric. Over-the-year percent change in average annual pay.
#'    
#'    \item date Date. Calculated calendar date based on year and quarter.  Reflects first day of the quarter.
#'    \item industry_title Character. Added based on industry_code
#'    \item ind_level Character. Description of the level of aggregation based on the industry_code.
#'    \item naics_2d Character. First two characters in the industry_code, useful for identifying industries.
#'    \item sector Character. Similar to naics_2d, but for industries like Manufacturing which have multiple two digit NAICS codes, this will span those groupings, for example "31-33"
#'    \item vintage_start. Integer. Calendar year of the earliest vintage for this industry_code. NAICS codes are updated every 5 years.  When using this industry codes from before this date, these titles may not exist or may be incorrect.
#'    \item vintage_end. Integer. Calendar year of the last year this industry code was used.  Years after this point should not contain this industry code.  Set to 3000 for current data.
#'    \item area_title Character.  Area description based on area_fips as provided by the BLS.
#'    \item area_type Character. Description of the type of area based on the area_title. More consistent naming and grouping than BLS data.
#'    \item stfips Character. The two-digit FIPS code of the state containing a given area. Set to "00" for multi-state regions.
#'    \item specified_region. Either a two-character US Postal Service abbreviation for the state containing an area or a hyphenated list of such codes for multi-state areas.
#'   }
#'   
#'   
#' @importFrom data.table rbindlist fread :=
#' @importFrom stringr str_replace_all
#' @export
#'
#' @examples
#' \donttest{
#' # Get quarterly data for "Total, all industries" (Code 10)
#' # Includes industry/area descriptions and a date column by default
#' dt_default <- get_qcew(industry_code = "10")
#'
#' # Get annual data for Nevada (Code 32000) for 2023 without lookups or messages
#' dt_year <- get_qcew(period_type = "year",
#'                     year_start = 2023,
#'                     year_end = 2023,
#'                     area_code = "32000",
#'                     add_lookups = FALSE,
#'                     silently = TRUE)
#' }
get_qcew <- function(period_type = "quarter",
                     year_start = NULL,
                     year_end = NULL,
                     industry_code = NULL,
                     area_code = NULL,
                     add_lookups = TRUE,
                     silently = FALSE) {
  
  # --- Input Validation & Defaults ---
  
  # Calculate default year (System Date - 6 Months) if not provided
  if (is.null(year_start) || is.null(year_end)) {
    # safe base R method to subtract 6 months
    default_date <- seq(Sys.Date(), by = "-6 months", length.out = 2)[2]
    default_year <- as.numeric(format(default_date, "%Y"))
    
    if (is.null(year_start)) year_start <- default_year
    if (is.null(year_end)) year_end <- default_year
  }
  
  # Check for pre-2014 data availability
  if (year_start < 2014) {
    warning("Data prior to 2014 is not available via these data slices. URLs may fail.")
  }
  
  if (!period_type %in% c("quarter", "year")) {
    stop("period_type must be either 'quarter' or 'year'.")
  }
  
  if (is.null(industry_code) && is.null(area_code)) {
    stop("You must provide either an industry_code or an area_code.")
  }
  
  if (!is.null(industry_code) && !is.null(area_code)) {
    warning("Please provide only one: industry_code OR area_code, not both. Defaulting to use industry_code for all areas.")
  }
  
  # --- Setup Iteration Parameters ---
  years <- seq(from = year_start, to = year_end)
  
  # Define quarter codes: 1-4 for quarterly, 'a' for annual
  if (period_type == "quarter") {
    quarters <- as.character(1:4)
  } else {
    quarters <- "a"
  }
  
  # Initialize list to store data tables
  data_list <- list()
  
  # Base URL for BLS QCEW Data Slices
  base_url <- "https://data.bls.gov/cew/data/api"
  
  # --- Iteration Loop ---
  for (yr in years) {
    for (qtr in quarters) {
      
      # Construct the dynamic part of the URL
      if (!is.null(industry_code)) {
        clean_code <- stringr::str_replace_all(industry_code, "-", "_")
        url_path <- paste(yr, qtr, "industry", paste0(clean_code, ".csv"), sep = "/")
      } else {
        url_path <- paste(yr, qtr, "area", paste0(area_code, ".csv"), sep = "/")
      }
      
      full_url <- paste(base_url, url_path, sep = "/")
      
      if (!silently) {
        message("Accessing: ", full_url)
      }
      
      # Attempt to fetch data
      tryCatch({
        # suppressWarnings hides the messy download.file output
        temp_dt <- suppressWarnings(
          data.table::fread(full_url, 
                            showProgress = FALSE, 
                            colClasses = c("industry_code" = "character",
                                           "area_fips" = "character"))
        )
        
        data_list[[length(data_list) + 1]] <- temp_dt
        
      }, error = function(e) {
        # Determine the period label for the clean warning
        period_label <- if(qtr == "a") "Annual" else paste0("Q", qtr)
        
        # Check if the error is a missing file (Status 404)
        if (grepl("cannot open URL", e$message)) {
          warning(paste(yr, period_label, "is not found (Status 404)"), call. = FALSE)
        } else {
          # Provide a clean warning for other types of errors
          warning(paste("Error fetching", yr, period_label, ":", e$message), call. = FALSE)
        }
      })
    }
  }
  
  # --- Aggregation ---
  if (length(data_list) == 0) {
    warning("No data was retrieved. Please check your parameters and internet connection.")
    return(NULL)
  }
  
  # Fast bind of all fetched tables
  final_dt <- data.table::rbindlist(data_list, use.names = TRUE, fill = TRUE)
  
  # --- Add Date Column ---
  # We use the 'year' and 'qtr' columns that come directly from the BLS CSVs
  if (period_type == "quarter") {
    # Calculate month: Q1->1, Q2->4, Q3->7, Q4->10 using math: (qtr*3) - 2
    final_dt[, temp_month := (as.numeric(qtr) * 3) - 2]
    final_dt[, date := as.Date(paste(year, temp_month, "01", sep = "-"))]
    final_dt[, temp_month := NULL] # Clean up helper column
  } else {
    # For annual, default to January 1st
    final_dt[, date := as.Date(paste(year, "01", "01", sep = "-"))]
  }
  
  # --- Lookups Merge ---
  if (add_lookups) {
    # Check if lookups are available in the package namespace
    if (exists("ind_lookup") && exists("area_lookup")) {
      
      # Left join industry lookup
      if ("industry_code" %in% names(final_dt)) {
        final_dt <- merge(final_dt, ind_lookup, by = "industry_code", all.x = TRUE)
      }
      
      # Left join area lookup
      if ("area_fips" %in% names(final_dt)) {
        final_dt <- merge(final_dt, area_lookup, by = "area_fips", all.x = TRUE)
      }
      
    } else {
      warning("add_lookups = TRUE, but 'ind_lookup' or 'area_lookup' were not found in the environment. Returning data without lookups.")
    }
  }
  
  return(final_dt)
}

#' Get National Current Employment Statistics (CES) Data from BLS
#'
#' This function downloads and processes national Current Employment Statistics (CES)
#' data from the Bureau of Labor Statistics (BLS). It retrieves multiple related
#' datasets and joins them together to create a comprehensive employment statistics
#' dataset with industry classifications, data types, and time period information.
#'
#' @param dataset_filter Character string specifying which dataset to download.
#'   Options include:
#'   \itemize{
#'     \item "all_data" (default) - Complete dataset with all series
#'     \item "current_seasonally_adjusted" - Only seasonally adjusted all-employee series
#'     \item "real_earnings_all_employees" - Real earnings data for all employees
#'     \item "real_earnings_production" - Real earnings data for production employees
#'   }
#' @param monthly_only Logical. If TRUE (default), excludes annual averages
#'   (period "M13") and returns only monthly data. If FALSE, includes all
#'   periods including annual averages.
#' @param simplify_table Logical. If TRUE (default), removes several metadata
#'   columns (series_title, begin_year, begin_period, end_year, end_period,
#'   naics_code, publishing_status, display_level, selectable, sort_sequence)
#'   and adds a formatted date column. If FALSE, returns the full dataset
#'   with all available columns.
#' @param suppress_warnings Logical. If TRUE (default), suppresses download warnings
#'   and diagnostics. If FALSE, displays warning output and diagnostic information.
#' @param return_diagnostics Logical. If TRUE, returns a bls_data_collection object
#'   with full diagnostics. If FALSE (default), returns just the data table.
#' @param cache Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.
#'
#' @return By default, returns a data.table with CES data. If return_diagnostics = TRUE,
#'   returns a bls_data_collection object containing data and comprehensive diagnostics.
#'
#' @details
#' The function can download one of four specialized national CES datasets based on
#' the dataset_filter parameter:
#' \itemize{
#'   \item all_data: Complete dataset (ce.data.0.AllCESSeries) - contains entire
#'     history of all series currently published by the CES program
#'   \item current_seasonally_adjusted: (ce.data.01a.CurrentSeasAE) - contains
#'     every seasonally adjusted all employee series and complete history
#'   \item real_earnings_all_employees: (ce.data.02b.AllRealEarningsAE) - contains
#'     real earnings data (1982-84 dollars) for all employees
#'   \item real_earnings_production: (ce.data.03c.AllRealEarningsPE) - contains
#'     real earnings data (1982-84 dollars) for production/nonsupervisory employees
#' }
#'
#' Additional metadata files are always downloaded and joined:
#' \itemize{
#'   \item ce.series - Series metadata
#'   \item ce.industry - Industry classifications
#'   \item ce.datatype - Data type definitions
#'   \item ce.period - Time period definitions
#'   \item ce.supersector - Supersector classifications
#' }
#'
#' These datasets are joined together to provide context and labels for the
#' employment statistics. The function uses the enhanced `download_bls_files()`
#' helper function for robust downloads with diagnostic reporting.
#'
#' Performance Note: Using specialized datasets (other than "all_data") can
#' significantly reduce download time and file size while still providing
#' comprehensive employment statistics.
#'
#' @note
#' This function requires the following packages: dplyr, data.table, httr, and
#' lubridate (for date formatting when simplify_table=TRUE). The `fread_bls()`
#' and `create_bls_object()` helper functions must be available in your environment.
#'
#' @examples
#' \dontrun{
#' # Get complete monthly CES data with simplified table structure (default)
#' ces_monthly <- get_national_ces()
#'
#' # Get only seasonally adjusted data (faster download)
#' ces_seasonal <- get_national_ces(dataset_filter = "current_seasonally_adjusted")
#'
#' # Get real earnings data for all employees
#' ces_real_earnings <- get_national_ces(dataset_filter = "real_earnings_all_employees")
#'
#' # Get all data including annual averages with full metadata
#' ces_full <- get_national_ces(dataset_filter = "all_data",
#'                              monthly_only = FALSE, simplify_table = FALSE)
#'
#' # Get data with warnings and diagnostic information displayed
#' ces_with_warnings <- get_national_ces(suppress_warnings = FALSE)
#'
#' # Get full diagnostic object if needed
#' data_with_diagnostics <- get_national_ces(return_diagnostics = TRUE)
#' print_bls_warnings(data_with_diagnostics)
#' }
#'
#'
#' @seealso
#' Please visit the Bureau of Labor Statistics at https://www.bls.gov/ces/ for more information about CES data
#'
#' @export
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr select
#' @importFrom lubridate ym
get_national_ces <- function(dataset_filter = "all_data", monthly_only = TRUE,
                             simplify_table = TRUE, suppress_warnings = TRUE,
                             return_diagnostics = FALSE, cache = check_bls_cache_env()) {

  # Validate dataset_filter parameter
  valid_filters <- c("all_data", "current_seasonally_adjusted",
                     "real_earnings_all_employees", "real_earnings_production")
  if (!dataset_filter %in% valid_filters) {
    stop("Invalid dataset_filter. Must be one of: ", paste(valid_filters, collapse = ", "))
  }

  # Define dataset-specific URLs
  if (dataset_filter == "all_data") {
    data_url <- "https://download.bls.gov/pub/time.series/ce/ce.data.0.AllCESSeries"
    dataset_name <- "Complete national CES dataset"
  } else if (dataset_filter == "current_seasonally_adjusted") {
    data_url <- "https://download.bls.gov/pub/time.series/ce/ce.data.01a.CurrentSeasAE"
    dataset_name <- "Seasonally adjusted all-employee series"
  } else if (dataset_filter == "real_earnings_all_employees") {
    data_url <- "https://download.bls.gov/pub/time.series/ce/ce.data.02b.AllRealEarningsAE"
    dataset_name <- "Real earnings for all employees"
  } else if (dataset_filter == "real_earnings_production") {
    data_url <- "https://download.bls.gov/pub/time.series/ce/ce.data.03c.AllRealEarningsPE"
    dataset_name <- "Real earnings for production employees"
  }

  # Define URLs for all CES datasets
  ces_urls <- c(
    "data" = data_url,
    "series" = "https://download.bls.gov/pub/time.series/ce/ce.series",
    "industry" = "https://download.bls.gov/pub/time.series/ce/ce.industry",
    "period" = "https://download.bls.gov/pub/time.series/ce/ce.period",
    "datatype" = "https://download.bls.gov/pub/time.series/ce/ce.datatype",
    "supersector" = "https://download.bls.gov/pub/time.series/ce/ce.supersector"
  )

  # Download all files
  message("Downloading national CES datasets (", dataset_name, ")...")
  downloads <- download_bls_files(ces_urls, suppress_warnings = suppress_warnings, cache = cache)
  
  # Exit function if download failed.
  if(is.null(downloads) | length(downloads) == 0){
    stop("Download of BLS data failed.  Please run with suppress_warnings = FALSE for additional status messages. Consider setting the BLS_USER_AGENT environment variable to your email address to avoid Status 403 errors from BLS.")
  }

  # Extract data from each download
  ces_data <- get_bls_data(downloads[["data"]])
  ces_series <- get_bls_data(downloads[["series"]])
  ces_industry <- get_bls_data(downloads[["industry"]])
  ces_period <- get_bls_data(downloads[["period"]])
  ces_datatype <- get_bls_data(downloads[["datatype"]])
  ces_supersector <- get_bls_data(downloads[["supersector"]])

  # Track processing steps
  processing_steps <- character(0)

  # Join all datasets together
  message("Joining CES datasets...")
  ces_full <- ces_data |>
    dplyr::select(-footnote_codes) |>
    dplyr::left_join(ces_series, by = "series_id") |>
    dplyr::select(-footnote_codes) |>
    dplyr::left_join(ces_industry, by = "industry_code") |>
    dplyr::left_join(ces_period, by = "period") |>
    dplyr::left_join(ces_datatype, by = "data_type_code") |>
    dplyr::left_join(ces_supersector, by = "supersector_code")

  processing_steps <- c(processing_steps, "joined_all_datasets")

  # Filter to monthly data only if requested
  if (monthly_only) {
    ces_full <- ces_full |>
      dplyr::filter(period != "M13")
    processing_steps <- c(processing_steps, "filtered_monthly_only")
  }

  # Simplify table structure if requested
  if (simplify_table) {
    ces_full <- ces_full |>
      dplyr::select(-c(series_title, begin_year, begin_period, end_year, end_period,
                       naics_code, publishing_status, display_level, selectable, sort_sequence)) |>
      dplyr::mutate(date = lubridate::ym(paste0(year, period)))
    processing_steps <- c(processing_steps, "simplified_table", "added_date_column")
  }

  # Create the BLS data collection object
  bls_collection <- create_bls_object(
    data = ces_full,
    downloads = downloads,
    data_type = paste("National CES:", dataset_name),
    processing_steps = processing_steps
  )

  # Display warnings if requested
  if (!suppress_warnings) {
    print_bls_warnings(bls_collection)
  }

  message("National CES data download complete!")
  message("Dataset: ", dataset_name)
  message("Final dataset dimensions: ", paste(dim(ces_full), collapse = " x "))

  # Return either the collection object or just the data
  if (return_diagnostics) {
    return(bls_collection)
  } else {
    return(ces_full)
  }
}

#' Download BLS Time Series Data
#'
#' This function downloads a tab-delimited BLS flat file, incorporating 
#' diagnostic information about the file and returning an object with the
#' bls_data class that can be used in the BLSloadR package.
#'
#' @param url Character string. URL to the BLS flat file
#' @param verbose Logical. If TRUE, prints additional messages during file read and processing.
#' @param cache Logical. If TRUE, uses local persistent caching.
#' @return A named list with the data and diagnostics.
#' @export
#' @importFrom httr GET stop_for_status content add_headers
#' @importFrom data.table fread
fread_bls <- function(url, verbose = FALSE, cache = check_bls_cache_env()) {
  
  # --- 1. DATA ACQUISITION ---
  if (cache) {
    # Uses the smart download logic to check headers/mtime
    temp_file <- smart_bls_download(url, verbose = verbose)
  } else {
    headers <- get_bls_headers()
    
    # Perform request and catch transport-level failures gracefully
    response <- tryCatch(
      httr::GET(url, httr::add_headers(.headers = headers)),
      error = function(e) {
        if (verbose) message("Network/transport error: ", conditionMessage(e))
        return(NULL)
      }
    )
    
    # If transport failed, exit early
    if (is.null(response)) {
      return(NULL)
    }
    
    status <- httr::status_code(response)
    
    # For any non-2xx status, fail gracefully and return NULL
    if (status < 200 || status >= 300) {
      # Human-readable reason (e.g., "Client error", "Server error")
      hs <- httr::http_status(response)
      
      # Capture and clean server message (strip HTML, normalize spaces)
      error_body <- httr::content(response, as = "text", encoding = "UTF-8")
      clean_error <- gsub("<.*?>", "", error_body)
      clean_error <- trimws(gsub("\\s+", " ", clean_error))
      clean_error <- substr(clean_error, 1, 500)
      
      # Provide a short hint by status code
      hint <- switch(
        as.character(status),
        "401" = "Unauthorized.",
        "403" = "Forbidden.",
        "404" = "Not found.",
        "429" = "Rate limited.",
        {
          if (status >= 500) "Server error. Consider retrying with backoff."
          else "Client error. Inspect request headers and URL."
        }
      )
      
      if (verbose) {
        message(
          sprintf(
            "%s (%d). %s%s",
            hs$message %||% hs$reason %||% "HTTP error",
            status,
            if (nzchar(clean_error)) paste0(" Server message: ", clean_error) else " No server message provided.",
            if (nzchar(hint)) paste0(" Brief code description: ", hint) else ""
          )
        )
      }
      
      return(NULL)
    }
    
    
    raw_data <- httr::content(response, as = "raw")
    temp_file <- tempfile(fileext = ".txt")
    writeBin(raw_data, temp_file)
  }
  
  # --- 2. INITIAL DIAGNOSTIC PASS ---
  # Read as-is to check for phantom columns
  initial_data <- data.table::fread(
    temp_file, 
    sep = "\t", 
    colClasses = "character", 
    header = TRUE, 
    fill = TRUE, 
    showProgress = FALSE
  )
  
  phantom_cols <- sapply(initial_data, function(col) {
    all(is.na(col) | col == "" | grepl("^\\s*$", col))
  })
  
  has_phantoms <- sum(phantom_cols) > 0
  
  if (verbose) {
    message("Initial data dimensions: ", nrow(initial_data), " x ", ncol(initial_data))
    message("Phantom columns detected: ", sum(phantom_cols))
  }
  
  # --- 3. VECTORIZED CLEANING (Only if needed) ---
  if (has_phantoms) {
    # Read file back as raw to perform vectorized string replacement
    raw_bytes <- readBin(temp_file, "raw", n = file.info(temp_file)$size)
    text_data <- rawToChar(raw_bytes)
    
    # VECTORIZED REPLACEMENT: Faster than row-by-row sapply
    # Replaces tab + whitespace + tab with single tab across the entire file at once
    cleaned_data <- gsub("\t\\s+\t", "\t", text_data, perl = TRUE)
    
    # Overwrite the file with cleaned data
    writeLines(cleaned_data, temp_file, sep = "")
    
    if (verbose) message("Applied vectorized tab cleaning.")
    
    # Re-read the cleaned data (Necessary because the structure changed)
    return_data <- data.table::fread(
      temp_file, 
      sep = "\t", 
      colClasses = "character", 
      header = TRUE, 
      fill = TRUE,
      showProgress = FALSE
    )
  } else {
    # If no cleaning was needed, return_data is just initial_data
    return_data <- initial_data
  }
  
  # --- 4. HEADER & COLUMN MANAGEMENT ---
  # Extract and clean header names from the file
  header_line <- readLines(temp_file, n = 1)
  header_names <- trimws(strsplit(header_line, "\t", fixed = TRUE)[[1]])
  
  n_header_cols <- length(header_names)
  n_data_cols <- ncol(return_data)
  
  if (n_header_cols != n_data_cols) {
    if (verbose) warning("Column count mismatch! Headers: ", n_header_cols, " Data: ", n_data_cols)
    if (n_data_cols > n_header_cols) {
      header_names <- c(header_names, paste0("EXTRA_COL_", 1:(n_data_cols - n_header_cols)))
    } else {
      header_names <- header_names[1:n_data_cols]
    }
  }
  
  # Assign names
  if (length(header_names) == ncol(return_data)) {
    names(return_data) <- header_names
  }
  
  # Final Empty Column Removal (Post-cleaning check)
  empty_cols <- sapply(return_data, function(col) {
    all(is.na(col) | col == "" | grepl("^\\s*$", col))
  })
  
  if (any(empty_cols)) {
    if (verbose) message("Removing ", sum(empty_cols), " remaining empty columns.")
    return_data <- return_data[, !empty_cols, with = FALSE]
  }
  
  # --- 5. CLEANUP & DIAGNOSTICS ---
  if (!cache) {
    unlink(temp_file)
  }
  
  diagnostics <- list(
    url = url,
    original_dimensions = c(nrow(initial_data), ncol(initial_data)),
    final_dimensions = c(nrow(return_data), ncol(return_data)),
    phantom_columns_detected = sum(phantom_cols),
    cleaning_applied = has_phantoms,
    header_data_mismatch = n_header_cols != n_data_cols,
    empty_columns_removed = sum(empty_cols),
    final_column_names = names(return_data),
    warnings = character(0)
  )
  
  result <- list(data = return_data, diagnostics = diagnostics)
  class(result) <- c("bls_data", "list")
  
  return(result)
}

#' Download BLS Excel Data
#'
#' @param url Character string. URL to the BLS .xlsx or .xls file.
#' @param verbose Logical. If TRUE, prints diagnostic messages.
#' @param ... Additional arguments passed to readxl::read_excel (e.g., sheet, range).
#' @return A data.frame or NULL if the download or read fails.
#' @export
#' @importFrom httr GET add_headers status_code http_status content
#' @importFrom readxl read_excel
#' @examples
#' \dontrun{
#' # Download BLS Alternative MEasures History
#' salt_url <- "https://www.bls.gov/lau/stalt-moave.xlsx"
#' salt_data <- read_bls_excel(salt_url, skip = 1)
#' 
#' }
#' 
read_bls_excel <- function(url, verbose = FALSE, ...) {
  # --- 1. DATA ACQUISITION ---
  headers <- get_bls_excel_headers()
  
  # Perform request and catch transport-level failures (e.g., DNS, Connection Refused)
  response <- tryCatch(
    httr::GET(url, httr::add_headers(.headers = headers)),
    error = function(e) {
      message("Network error: ", conditionMessage(e))
      return(NULL)
    }
  )
  
  if (is.null(response)) return(NULL)
  
  status <- httr::status_code(response)
  
  # --- 2. ERROR HANDLING (Always status, Detailed if Verbose) ---
  if (status < 200 || status >= 300) {
    hs <- httr::http_status(response)
    
    # Always print the basic failure status
    message(sprintf("Download failed for %s\nStatus: %d (%s)", url, status, hs$reason))
    
    # Provide full response details only if verbose is TRUE
    if (verbose) {
      # Capture and clean server message
      error_body <- tryCatch(httr::content(response, as = "text", encoding = "UTF-8"), error = function(e) "")
      clean_error <- gsub("<.*?>", "", error_body)
      clean_error <- substr(trimws(gsub("\\s+", " ", clean_error)), 1, 500)
      
      # Determine Hint
      hint <- switch(as.character(status),
                     "401" = "Unauthorized.",
                     "403" = "Forbidden. Check User-Agent or API key.",
                     "404" = "Not found.",
                     "429" = "Rate limited.",
                     if (status >= 500) "Server error. Consider retrying later." else "Client error.")
      
      message(sprintf("Hint: %s", hint))
      if (nzchar(clean_error)) message(sprintf("Server Message: %s", clean_error))
    }
    
    return(NULL)
  }
  
  # --- 3. FILE PROCESSING ---
  raw_data <- httr::content(response, as = "raw")
  temp_file <- tempfile(fileext = ".xlsx")
  writeBin(raw_data, temp_file)
  
  # Wrap the read in tryCatch for a graceful exit if the file is unreadable
  data_out <- tryCatch({
    readxl::read_excel(temp_file, ...)
  }, error = function(e) {
    message("Failed to parse Excel content: ", conditionMessage(e))
    return(NULL)
  })
  
  # Cleanup temp file
  if (file.exists(temp_file)) unlink(temp_file)
  
  return(data_out)
}

#' Display BLS Dataset Overview
#'
#' Fetches and displays the overview text file for a BLS dataset. This provides a convenient reference within the R environment without needing to manually find and review the text file on the BLS website.
#'
#' @param series_id Character string. The BLS series identifier (e.g., "ln", "cu", "ap")
#' @param display_method Character string. How to display the overview: 
#'   "viewer" (default), "console", or "popup"
#' @param base_url Character string. Base URL for BLS data (default uses official BLS site)
#'
#' @return Invisibly returns the text content. Function is called to use the viewer, console, or as a popup, depending on the 'display_method' argument.
#' 
#' @export
#' 
#' @importFrom httr GET add_headers stop_for_status content
#' @importFrom htmltools HTML
#' @importFrom htmltools htmlEscape
#' @importFrom rstudioapi isAvailable
#'
#' @examples
#' \donttest{
#' if(interactive()){
#' # Display Average Price Data overview
#' bls_overview("ap")
#' 
#' # Display consumer price index overview  
#' bls_overview("cu")
#' 
#' # Display in console instead of viewer
#' bls_overview("ap", display_method = "console")
#' }
#' }
bls_overview <- function(series_id, 
                         display_method = "viewer",
                         base_url = "https://download.bls.gov/pub/time.series") {
  
  # Validate inputs
  if (!is.character(series_id) || length(series_id) != 1) {
    stop("series_id must be a single character string")
  }
  
  display_method <- match.arg(display_method, c("viewer", "console", "popup"))
  
  # Construct URL
  url <- file.path(base_url, series_id, paste0(series_id, ".txt"))
  
  # Fetch content with proper headers (similar to fread_bls)
  tryCatch({
    headers <- get_bls_headers()
    
    response <- httr::GET(url, httr::add_headers(.headers = headers))
    httr::stop_for_status(response)
    
    content_text <- httr::content(response, as = "text", encoding = "UTF-8")
    
    # Display based on method
    switch(display_method,
           "viewer" = display_in_viewer(content_text, series_id),
           "console" = display_in_console(content_text, series_id),
           "popup" = display_in_popup(content_text, series_id)
    )
    
    invisible(content_text)
    
  }, error = function(e) {
    stop(sprintf("Could not fetch overview for series '%s'. URL: %s\nError: %s", 
                 series_id, url, e$message))
  })
}

#' Display text content in Viewer window.
#'
#' Helper function used to display content from 'bls_overview' in the HTML viewer.
#' 
#' @param content Character. Text content to display
#' @param series_id Two-letter series ID for a BLS time series to render in the display.
#'
#' @return No object returned, called to render content in HTML viewer.
display_in_viewer <- function(content, series_id) {
  if (!requireNamespace("htmltools", quietly = TRUE)) {
    stop("Package 'htmltools' is required for viewer display. Install with: install.packages('htmltools')")
  }
  
  # Create HTML content
  html_content <- htmltools::HTML(sprintf("
    <html>
    <head>
      <title>BLS Overview: %s</title>
      <style>
        body { 
          font-family: 'Courier New', monospace; 
          margin: 20px; 
          line-height: 1.4;
          background-color: #f8f9fa;
        }
        .header { 
          background-color: #007bff; 
          color: white; 
          padding: 15px; 
          margin: -20px -20px 20px -20px;
          border-radius: 0;
        }
        .content { 
          white-space: pre-wrap; 
          background-color: white;
          padding: 20px;
          border: 1px solid #dee2e6;
          border-radius: 5px;
        }
      </style>
    </head>
    <body>
      <div class='header'>
        <h1>BLS Dataset Overview: %s</h1>
        <p>Source: https://download.bls.gov/pub/time.series/%s/%s.txt</p>
      </div>
      <div class='content'>%s</div>
    </body>
    </html>
  ", toupper(series_id), toupper(series_id), series_id, series_id, 
                                          htmltools::htmlEscape(content)))
  
  # Create temporary file and display
  temp_file <- tempfile(fileext = ".html")
  writeLines(as.character(html_content), temp_file)
  
  if (rstudioapi::isAvailable()) {
    rstudioapi::viewer(temp_file)
  } else {
    utils::browseURL(temp_file)
  }
}




#' Display text content in console.
#'
#' Helper function used to display content from 'bls_overview' in the console..
#' 
#' @param content Character. Text content to display
#' @param series_id Two-letter series ID for a BLS time series to render in the display.
#'
#' @return No object returned, called to render content in console only..
display_in_console <- function(content, series_id) {
  cat(sprintf("\n=== BLS Dataset Overview: %s ===\n", toupper(series_id)))
  cat(sprintf("Source: https://download.bls.gov/pub/time.series/%s/%s.txt\n", series_id, series_id))
  cat(paste(rep("=", 50), collapse = ""), "\n\n")
  cat(content, "\n\n")
}


#' Display text content in popup window.
#'
#' Helper function used to display content from 'bls_overview' in a popup window.
#' 
#' @param content Character. Text content to display
#' @param series_id Two-letter series ID for a BLS time series to render in the display.
#'
#' @return No object returned, called to render content in popup.
display_in_popup <- function(content, series_id) {
  if (rstudioapi::isAvailable() && rstudioapi::hasFun("showDialog")) {
    # Truncate content if too long for dialog
    display_content <- content
    if (nchar(content) > 2000) {
      display_content <- paste(substr(content, 1, 2000), "\n\n[Content truncated - use viewer method for full text]")
    }
    
    rstudioapi::showDialog(
      title = paste("BLS Overview:", toupper(series_id)),
      message = display_content
    )
  } else {
    message("Popup not available, displaying in console instead")
    display_in_console(content, series_id)
  }
}

#' @title NAICS Industry Titles Lookup Table (QCEW)
#'
#' @description
#' A data frame containing the structure of the North American Industry Classification System (NAICS) titles and codes used by the U.S. Bureau of Labor Statistics (BLS), including in the Quarterly Census of Employment and Wages (QCEW) program. This data is current as of the 2022 NAICS revision.
#'
#' This table is primarily used internally by `BLSloadR` functions to add human-readable industry titles or define valid industry lookups.
#'
#' @format A data frame with 2678 rows and 7 columns:
#' \itemize{
#'   \item{industry_code - Character.  The NAICS industry code.  Items starting with 10 are aggregated values, not corresponding to a unique 6-digit NAICS.}
#'   \item{industry_title - Character. The descripion of the industry code provided by the BLS.  This title also includes the code value for clarity.}
#'   \item{ind_level - Character. A description of the level of aggregation. Values are "Total", "Cluster", or "Supersector" for the "10" code aggregations, or else the length of the NAICS code, from 2-6 digits.}
#'   \item{naics_2d - Character. The first two digits of `industry_code`, which may be helpful to filter the results.}
#'   \item{sector - Character.  Similar to `naics_2d` except that when the industry sector spans multiple two digit codes, those codes are hyphenated (e.g. Manufacturing is NAICS 31, 32, and 33, so this displays '31-33').}
#'   \item{vintage_start - Integer.  The earliest year reviewed for NAICS code use. NAICS will change every 5 years, so data from before this year will have some missing values.}
#'   \item{vintage_end - Integer. The last year that a particular code is used, if applicable. Set to 3000 for current codes.}
#' }
#'
#' @source \url{https://www.bls.gov/cew/classifications/industry/industry-titles.htm}
#'
#' @usage data(ind_lookup)
#'
#' @details
#' The NAICS structure is hierarchical. Codes are typically longer for more detailed industries.
#'
#' @examples
#' # Load the lookup table
#' data(ind_lookup)
#'
#' # Find the industry title for NAICS 51 (Information)
#' ind_lookup[ind_lookup$industry_code == "51", ]
#'
#' # Get the supersector codes
#' supersectors <- ind_lookup[ind_lookup$ind_level == "Supersector", ]
#' 
#' # Get all 3-digit NAICS codes in the Manufacturing industry
#' mfg_codes <- ind_lookup |> 
#' dplyr::filter(sector == "31-33" & ind_level == "NAICS 3-digit")
"ind_lookup"

#' @title Area Lookup Tables (QCEW)
#'
#' @description
#' A data frame containing area codes, titles, and additional geographic information about valid areas for the Quarterly Census of Employment and Wages (QCEW).
#'
#' This table is primarily used internally by `BLSloadR` functions to add human-readable area titles or define valid area lookups.
#'
#' @format A data frame with 4649 rows and 5 columns:
#' \itemize{
#'   \item{area_fips - Character.  The area FIPS code.  When all numeric characters, it represents either a state or a county definition.}
#'   \item{area_title - Character. The descripion of the area code provided by the Bureau of Labor Statistics.}
#'   \item{area_type - Character. A desription ot the type of area defined.  Values are National, State, County, County Unknown or Undefined, National Subgroup, Combined Statistical Area, Metropolitan Statistical Area, or Micropolitan Statistical Area.}
#'   \item{stfips - Character. For state or counties, the two-digit FIPS code of the associated state.  For national areas, or those areas which may span multiple states the value is "00".}
#'   \item{specified_region - Either the two-character US Postal abbreviation for a state or group of states, or "No region" for other areas.}
#' }
#'
#' @source \url{https://www.bls.gov/cew/classifications/areas/qcew-area-titles.htm}
#'
#' @usage data(area_lookup)
#'
#' @details
#' Area codes are five characters long. When all numeric characters, this is a state-county FIPS, with statewide data using a "000" as the county FIPS.  Other aggregations include various alphabetic characters to aid in the classification of different regions.
#'
#' @examples
#' # Load the lookup table
#' data(area_lookup)
#'
#' # Find the area codes for all Statewide areas
#' state_codes <- area_lookup[area_lookup$area_type == "State", ]
#' 
#' # Get all Metropolitan Statistical Area codes including Arkansas.
#' ar_codes <- area_lookup |> 
#' dplyr::filter(grepl("AR", specified_region) &
#'  area_type == "Metropolitan Statistical Area")
"area_lookup"

# R/globals.R
# Global variables declaration to avoid R CMD check warnings
# These variables are used in non-standard evaluation contexts (dplyr, data.table, etc.)

utils::globalVariables(c(
  # Common BLS series metadata columns
  "series_id",
  "series_title", 
  "begin_year",
  "begin_period",
  "end_year", 
  "end_period",
  "footnote_codes",
  "publishing_status",
  "display_level",
  "selectable",
  "sort_sequence",
  "naics_code",
  "sector_code",
  
  # Time and date related variables
  "year",
  "period", 
  "value",
  "benchmark_year",
  "quarter",
  "start quarter",
  "start year",
  "end quarter", 
  "end year",
  "unique period",
  
  # Geographic identifiers
  "fips",
  "fips_len",
  "state",
  "state_code",
  "record",
  
  # Employment and labor force measures
  "civilian_labor_force",
  "unemployed",
  "involuntary_part_time_employed",
  "marginally_attached_not_discouraged",
  "all_marginally_attached",
  "discouraged_workers",
  "job_losers",
  "not_job_losers",
  "unemployed_15+_weeks",
  "u1",
  "u2", 
  "u3",
  "u4",
  "u4b",
  "u5",
  "u5b",
  
  # Data classification codes and column names
  "data_type_code",
  "data_type_text",
  "dataelement_code",
  "measure_text",
  "ratelevel_code",
  "file_type",
  "file_name",
  "NAME",
  "oews_area_code",
  "fips_code",
  "county_code",
  "GEOID",
  "oews_area_name",
  "temp_month",
  "ind_lookup",
  "area_lookup",
  "refer",
  
  # Statistical functions (base R)
  "median",
  "quantile",
  
  # Placeholders within functions
  "result",
  "tf"
  
))

# Note: This file declares global variables used throughout the BLSloadR package
# to suppress R CMD check warnings about "no visible binding for global variable".
# These variables are column names that exist in BLS datasets and are referenced
# using non-standard evaluation in dplyr and data.table operations.

#' Read Plain Text Files from BLS Website
#'
#' Downloads and reads plain text files from the Bureau of Labor Statistics (BLS)
#' website. This is a companion function to \code{fread_bls()} that handles text
#' files rather than structured data tables. The function uses custom headers to
#' ensure reliable access to BLS resources.
#'
#' @param url A character string specifying the full URL to a text file on the
#'   BLS website (e.g., \url{https://download.bls.gov/pub/time.series/}).
#'
#' @return A character vector where each element is one line from the text file.
#'   Lines are split on newline characters (\code{\\n}).
#'
#' @details
#' This function is designed to read descriptive text files from BLS, such as
#' README files or database overview documents. It sends an HTTP GET request
#' with browser-like headers to ensure compatibility with BLS server requirements.
#'
#' The function will stop with an error if the HTTP request fails (e.g., if the
#' URL is invalid or the server is unavailable).
#'
#' @export
#'
#' @examples
#' \donttest{
#' # Read the overview file for Current Employment Statistics
#' ces_overview <- read_bls_text(
#'   "https://download.bls.gov/pub/time.series/ce/ce.txt"
#' )
#'
#' # Display the first few lines
#' head(ces_overview)
#' }
#'
#' @seealso
#' \code{\link{bls_overview}} for formatted database overviews,
#' \code{\link{load_bls_dataset}} for loading complete datasets
#'
#' @keywords internal
read_bls_text <- function(url) {
  headers <- c(
    "Accept" = "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding" = "gzip, deflate, br",
    "Accept-Language" = "en-US,en;q=0.9",
    "Connection" = "keep-alive",
    "Host" = "download.bls.gov",
    "Referer" = "https://download.bls.gov/pub/time.series/",
    "Sec-Ch-Ua" = 'Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
    "Sec-Ch-Ua-Mobile" = "?0",
    "Sec-Ch-Ua-Platform" = '"Windows"',
    "Sec-Fetch-Dest" = "document",
    "Sec-Fetch-Mode" = "navigate",
    "Sec-Fetch-Site" = "same-origin",
    "Sec-Fetch-User" = "?1",
    "Upgrade-Insecure-Requests" = "1",
    "User-Agent" = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
  )
  
  response <- httr::GET(url, httr::add_headers(.headers = headers))
  httr::stop_for_status(response)
  
  content_text <- httr::content(response, as = "text", encoding = "UTF-8")
  return(strsplit(content_text, "\n", fixed = TRUE)[[1]])
}

---
title: "Using the BLSloadR File Cache"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using the BLSloadR File Cache}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(rmarkdown.html_vignette.check_title = FALSE)
```

```{r setup}
library(BLSloadR)
```

## Overview of BLS File Structure

`BLSloadR` streamlines access to data from the U.S. Bureau of Labor Statistics (BLS). The primary source of this data is the BLS "flat files" which are published at <https://download.bls.gov/pub/time.series/>. This data is published as text files in the form of a relational database, with data organized into three main categories of file:

1.  Data files, which contain the values associated with discrete data series.
2.  Series files, which link data files to a series of data lookup codes via a series code.
3.  Lookup files, which map the series lookup codes to descriptive values.

Some databases will use multiple data files that act as slices of the larger data set. Others use an "aspect" file to add additional dimensions to the data.

## File Caching in BLSloadR

This article discusses the implementation of a file cache for `BLSloadR`, which will optionally download files from the BLS to local storage to accomplish two goals:

1.  Maintain a local copy of the data in case of network disruption.
2.  Reduce network overhead when regularly accessing BLS data, more frequently than the data is updated by BLS.

To preserve existing functionality, file caching is not enabled by default in `BLSloadR`. There are two ways to use file caching - as a one-off argument in a supported function, or by setting the environment variable in your Renviron file. When file caching is enabled, `BLSloadR` will take the following steps:

1.  View information about the remote files (size and last-modified date)
2.  Look for the corresponding file in your local cache

If the local file does not exist, or is older or a different size than the BLS file, then a new file is downloaded to the cache and then read. Otherwise, the local file will be read. Functions currently supporting a file cache are:

-   `get_ces()`

-   `get_national_ces()`

-   `get_laus()`

-   `get_oews()`

-   `get_jolts()`

### Using File Cache Case-by-Case

Supported functions in `BLSloadR` include the `cache` argument, which defaults to FALSE. Using the file cache on a case-by-case basis can be done by setting `cache=TRUE` in the function. Note that subsequent usage without setting the cache variable will ignore any cached files - in order to evaluate whether there are any cached files, the argument must be set to TRUE.

### Enable Default File Caching

In order to enable file caching by default, you can set an environment variable in your Renviron file, and add the following: `USE_BLS_CACHE="TRUE"`

When file caching is enabled by default, you can still disable use of the cache on a temporary basis by using the cache argument within the supported functions: `cache=FALSE`

### Controlling Cache Location

The BLS file cache can be controlled with the `BLS_CACHE_DIR` environment variable. If this variable is not set but the cache is used, `BLSloadR` will use the folder location given by `tools::R_user_dir("BLSloadR", which = "cache")`. You can check the cache directory with the helper function `bls_get_cache_dir()`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ces_helpers.R
\name{list_ces_industries}
\alias{list_ces_industries}
\title{List Available Industry Filters for CES Data}
\usage{
list_ces_industries(show_descriptions = FALSE)
}
\arguments{
\item{show_descriptions}{Logical. If TRUE, returns a data frame with
filter names and descriptions. If FALSE, returns just the filter names.}
}
\value{
A character vector of industry filter names, or a data frame
  with names and descriptions if show_descriptions = TRUE
}
\description{
Lists all available industry categories that can be used with the
`industry_filter` parameter in `get_ces()` function. These filters allow
you to download specific industry data instead of the complete dataset.
}
\examples{
# See all available industry filters
list_ces_industries()

# See filters with descriptions
list_ces_industries(show_descriptions = TRUE)

# Use with get_ces
# manufacturing_data <- get_ces(industry_filter = "manufacturing")  # All states
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/national_ces_helpers.R
\name{show_national_ces_options}
\alias{show_national_ces_options}
\title{Show National CES Dataset Options and Usage Examples}
\usage{
show_national_ces_options()
}
\value{
Prints formatted information to the console.
}
\description{
This function provides a comprehensive overview of the national CES dataset
filtering options available in get_national_ces(), including examples
of how to use each option.
}
\examples{
show_national_ces_options()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_laus.R
\name{get_laus}
\alias{get_laus}
\title{Download Local Area Unemployment Statistics (LAUS) Data}
\usage{
get_laus(
  geography = "state_adjusted",
  monthly_only = TRUE,
  transform = TRUE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{geography}{Character string specifying the geographic level and adjustment type.
Default is "state_adjusted". Valid options are:
\itemize{
  \item "state_current_adjusted" - Current seasonally adjusted state data
  \item "state_unadjusted" - All historical unadjusted state data
  \item "state_adjusted" - All historical seasonally adjusted state data (default)
  \item "region_unadjusted" - Unadjusted regional and division data
  \item "region_adjusted" - Seasonally adjusted regional and division data
  \item "metro" - Metropolitan statistical area data
  \item "division" - Division-level data
  \item "micro" - Micropolitan statistical area data
  \item "combined" - Combined statistical area data
  \item "county" - County-level data (large file >300MB)
  \item "city" - City and town data (large file >300MB)
  \item "1990-1994" - Comprehensive unadjusted data for 1990-1994
  \item "1995-1999" - Comprehensive unadjusted data for 1995-1999
  \item "2000-2004" - Comprehensive unadjusted data for 2000-2004
  \item "2005-2009" - Comprehensive unadjusted data for 2005-2009
  \item "2010-2014" - Comprehensive unadjusted data for 2010-2014
  \item "2015-2019" - Comprehensive unadjusted data for 2015-2019
  \item "2020-2024" - Comprehensive unadjusted data for 2020-2024
  \item "2025-2029" - Comprehensive unadjusted data for 2025-2029
  \item "ST" - Any state two-character USPS abbreviation, plus DC and PR
}}

\item{monthly_only}{Logical. If TRUE (default), excludes annual data (period M13)
and creates a date column from year and period.}

\item{transform}{Logical. If TRUE (default), converts rate and ratio measures from
percentages to proportions by dividing by 100. Unemployment rates will be expressed
as decimals (e.g., 0.05 for 5\% unemployment) rather than as whole numbers (e.g. 5).}

\item{suppress_warnings}{Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.}

\item{return_diagnostics}{Logical. If TRUE, returns a bls_data_collection object
with full diagnostics. If FALSE (default), returns just the data table.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
By default, returns a data.table with LAUS data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing LAUS data with the following key columns:
  \describe{
    \item{series_id}{BLS series identifier}
    \item{year}{Year of observation}
    \item{period}{Time period (M01-M12 for months, M13 for annual)}
    \item{value}{Employment statistic value (transformed if transform = TRUE)}
    \item{date}{Date of observation (if monthly_only = TRUE)}
    \item{area_text}{Geographic area name}
    \item{area_type_code}{Code indicating area type}
    \item{measure_text}{Type of measure (unemployment rate, labor force, employment, etc.)}
    \item{seasonal}{Seasonal adjustment status}
  }
}
\description{
This function downloads Local Area Unemployment Statistics data from the U.S. Bureau
 of Labor Statistics. Due to the large size of some LAUS datasets (county and city
 files are >300MB), users must specify which geographic level to download. The function
 provides access to both seasonally adjusted and unadjusted data at various geographic levels. 
 Additional datasets provide comprehensive non-seasonally-adjusted data for all areas broken out 
 in 5-year increments
}
\details{
The function joins data from multiple BLS files:
  \itemize{
    \item Main data file (varies by geography selection)
    \item Series definitions (la.series)
    \item Area codes and names (la.area)
    \item Measure definitions (la.measure)
  }
}
\examples{
\dontrun{
# Download state-level seasonally adjusted data (default operation)
laus_states <- get_laus()

# View unemployment rates by state for latest period
unemployment <- laus_states[grepl("unemployment rate", measure_text) & date == max(date)]

# Download unadjusted state data
laus_states_unadjusted <- get_laus(geography = "state_unadjusted")

# Get full diagnostic object if needed
laus_with_diagnostics <- get_laus(return_diagnostics = TRUE)
print_bls_warnings(laus_with_diagnostics)


}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{get_bls_summary}
\alias{get_bls_summary}
\title{Get summary information from BLS data object}
\usage{
get_bls_summary(bls_obj)
}
\arguments{
\item{bls_obj}{A bls_data_collection object}
}
\value{
List of summary information
}
\description{
This is a helper function to extract the summary element of a 'bls_data_collection' object. This containes the number of files downloaded, the number of files with potential warnings, and the total number of warnings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fread_BLS.R
\name{read_bls_excel}
\alias{read_bls_excel}
\title{Download BLS Excel Data}
\usage{
read_bls_excel(url, verbose = FALSE, ...)
}
\arguments{
\item{url}{Character string. URL to the BLS .xlsx or .xls file.}

\item{verbose}{Logical. If TRUE, prints diagnostic messages.}

\item{...}{Additional arguments passed to readxl::read_excel (e.g., sheet, range).}
}
\value{
A data.frame or NULL if the download or read fails.
}
\description{
Download BLS Excel Data
}
\examples{
\dontrun{
# Download BLS Alternative MEasures History
salt_url <- "https://www.bls.gov/lau/stalt-moave.xlsx"
salt_data <- read_bls_excel(salt_url, skip = 1)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ces_helpers.R
\name{list_ces_states}
\alias{list_ces_states}
\title{List Available States for CES Data}
\usage{
list_ces_states()
}
\value{
A character vector of available state/territory abbreviations
}
\description{
Lists all available U.S. states and territories that can be used with the
`states` parameter in `get_ces()` function.
}
\examples{
# See all available states
list_ces_states()

# Use with get_ces
# ces_data <- get_ces(states = c("MA", "NY"))  # All industries for these states
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{area_lookup}
\alias{area_lookup}
\title{Area Lookup Tables (QCEW)}
\format{
A data frame with 4649 rows and 5 columns:
\itemize{
  \item{area_fips - Character.  The area FIPS code.  When all numeric characters, it represents either a state or a county definition.}
  \item{area_title - Character. The descripion of the area code provided by the Bureau of Labor Statistics.}
  \item{area_type - Character. A desription ot the type of area defined.  Values are National, State, County, County Unknown or Undefined, National Subgroup, Combined Statistical Area, Metropolitan Statistical Area, or Micropolitan Statistical Area.}
  \item{stfips - Character. For state or counties, the two-digit FIPS code of the associated state.  For national areas, or those areas which may span multiple states the value is "00".}
  \item{specified_region - Either the two-character US Postal abbreviation for a state or group of states, or "No region" for other areas.}
}
}
\source{
\url{https://www.bls.gov/cew/classifications/areas/qcew-area-titles.htm}
}
\usage{
data(area_lookup)
}
\description{
A data frame containing area codes, titles, and additional geographic information about valid areas for the Quarterly Census of Employment and Wages (QCEW).

This table is primarily used internally by `BLSloadR` functions to add human-readable area titles or define valid area lookups.
}
\details{
Area codes are five characters long. When all numeric characters, this is a state-county FIPS, with statewide data using a "000" as the county FIPS.  Other aggregations include various alphabetic characters to aid in the classification of different regions.
}
\examples{
# Load the lookup table
data(area_lookup)

# Find the area codes for all Statewide areas
state_codes <- area_lookup[area_lookup$area_type == "State", ]

# Get all Metropolitan Statistical Area codes including Arkansas.
ar_codes <- area_lookup |> 
dplyr::filter(grepl("AR", specified_region) &
 area_type == "Metropolitan Statistical Area")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_qcew.R
\name{get_qcew}
\alias{get_qcew}
\title{Get QCEW Data Slices}
\usage{
get_qcew(
  period_type = "quarter",
  year_start = NULL,
  year_end = NULL,
  industry_code = NULL,
  area_code = NULL,
  add_lookups = TRUE,
  silently = FALSE
)
}
\arguments{
\item{period_type}{Character. Either "quarter" or "year". Defaults to "quarter".}

\item{year_start}{Numeric. The first year to retrieve data for. Defaults to the year
of the date 6 months prior to the current system date.}

\item{year_end}{Numeric. The last year to retrieve data for. Defaults to the year
of the date 6 months prior to the current system date.}

\item{industry_code}{Character. The NAICS industry code (e.g., "10", "31-33").
Constructs a URL for an Industry Data Slice. Mutually exclusive with `area_code`.}

\item{area_code}{Character. The QCEW area code (e.g., "US000", "32000", "C2982").
Constructs a URL for an Area Data Slice. Mutually exclusive with `industry_code`.}

\item{add_lookups}{Logical. If \code{TRUE}, joins the package's \code{ind_lookup} and
\code{area_lookup} tables to the results to provide descriptive labels. Defaults to \code{TRUE}.}

\item{silently}{Logical. If \code{TRUE}, suppresses status messages about the URLs being accessed.
Defaults to \code{FALSE}.}
}
\value{
A combined data.table containing the requested QCEW data, optionally merged with lookup columns
  and a calculated \code{date} column.  The data layout is different for quarterly or annual data files set by the `period_type` argument.
  
  For Quarterly files:
  \itemize{
   \item area_fips - Character. Area code of row.  Included `area_lookup` data file contains mapping information.
   \item industry_code - Character. NAICS, Supersector, Cluster, or Total All Industries code. Numeric characters as a string to preserve examining the structure heirarchy.
   \item own_code - Integer.  Values of 0-5 to designate ownership. See definitions at \url{https://www.bls.gov/cew/classifications/ownerships/ownership-titles.htm}
   \item agglvl_code - Integer.   Two digit code identifying the level of aggregation.  See definitons at \url{https://www.bls.gov/cew/classifications/aggregation/agg-level-titles.htm}
   \item size_code - Integer. Single-digit code representing the size of establishments. See definitions at \url{https://www.bls.gov/cew/classifications/size/size-titles.htm}
   \item year Integer.  Four-digit calendar year for the returned data.
   \item qtr Integer. The calendar quarter of the data. 
   \item disclosure_code Character.  Values are either a blank string on "N".  Values of N do not disclose employment or wages to maintain confidentiality.
   \item qtryly_estabs Integer.  The number of business establishments (worksites) for the industry in the area in the quarter.
   \item month1_emplvl Integer. Employment in the first month of the quarter (January, April, July, or October).
   \item month2_emplvl Integer. Employment in the second month of the quarter (February, May, August, November).
   \item month3_emplvl Integer. Employment in the third month of the quarter (March, June, September, December).
   \item total_qtrly_wages Ingeger64. Total wages paid during the quarter.
   \item taxable_qtrly_wages Ingeger64. Wages subject to unemployment insurance (UI) taxes during the quarter.  Note - wages subject to UI vary by state and will follow different seasonal patterns as a result.
   \item qtrly_contributions Integer. UI taxes (Contributions) paid by employers for this quarter. Note - UI tax policy varies by state.
   \item avg_wkly_wage Integer. Average weekly wage during the quarter (Total wages divided by average employment, divided by 13).
  
   \item lq_disclosure_code Character.  Blank or "N".  Values of "N" will suppress location quotient data for confidentiality.
   \item lq_qtrly_estabs Numeric. Location quotient of establishments relative to the U.S.
   \item lq_month1_emplvl Numeric. Location quotient of month 1 employment relative to the U.S.
   \item lq_month2_emplvl Numeric. Location quotient of month 2 employment relative to the U.S.
   \item lq_month3_emplvl Numeric. Location quotient of month 3 employment relative to the U.S.
   \item lq_total_qtrly_wages Numeric. Location quotient of total wages relative to the U.S.
   \item lq_taxable_qtrly_wages Numeric. Location quotient of taxable quarterly wages relative to the U.S.
   \item lq_qtrly_contributions Numeric. Location quotient of quarterly UI taxes paid relative to the U.S.
   \item lq_avg_wkly_wage Numeric. Location quotient of average weekly wages relative to the U.S.
   
   \item oty_disclosure_code Character.  Blank or "N".  Values of "N" will suppress over-the-year data for confidentiality.
   \item oty_qtrly_estabs_chg Numeric. Over-the-year change in establishments.
   \item oty_qtrly_estabs_pct_chg Numeric. Over-the-year percent change in establishments.
   \item oty_month1_emplvl_chg Numeric. Over-the-year change in month 1 employment.
   \item oty_month1_emplvl_pct_chg Numeric. Over-the-year percent change in month 1 employment.
   \item oty_month2_emplvl_chg Numeric. Over-the-year change in month 2 employment.
   \item oty_month2_emplvl_pct_chg Numeric. Over-the-year percent change in month 2 employment.
   \item oty_month3_emplvl_chg Numeric. Over-the-year change in month 3 employment.
   \item oty_month3_emplvl_pct_chg Numeric. Over-the-year percent change in month 3 employment.
   \item oty_total_qtrly_wages_chg Numeric. Over-the-year change in total wages.
   \item oty_total_qtrly_wages_pct_chg Numeric. Over-the-year percent change in total wages.
   \item oty_taxable_qtrly_wages_chg Numeric. Over-the-year change in taxable quarterly wages.
   \item oty_taxable_qtrly_wages_pct_chg Numeric. Over-the-year percent change in taxable quarterly wages.
   \item oty_qtrly_contributions_chg Numeric. Over-the-year change in quarterly UI taxes paid.
   \item oty_qtrly_contributions_pct_chg Numeric. Over-the-year percent change in quarterly UI taxes paid.
   \item oty_avg_wkly_wage_chg Numeric. Over-the-year change in average weekly wages.
   \item oty_avg_wkly_wage_pct_chg Numeric. Over-the-year percent change in average weekly wages.
   
   \item date Date. Calculated calendar date based on year and quarter.  Reflects first day of the quarter.
   \item industry_title Character. Added based on industry_code
   \item ind_level Character. Description of the level of aggregation based on the industry_code.
   \item naics_2d Character. First two characters in the industry_code, useful for identifying industries.
   \item sector Character. Similar to naics_2d, but for industries like Manufacturing which have multiple two digit NAICS codes, this will span those groupings, for example "31-33"
   \item vintage_start. Integer. Calendar year of the earliest vintage for this industry_code. NAICS codes are updated every 5 years.  When using this industry codes from before this date, these titles may not exist or may be incorrect.
   \item vintage_end. Integer. Calendar year of the last year this industry code was used.  Years after this point should not contain this industry code.  Set to 3000 for current data.
   \item area_title Character.  Area description based on area_fips as provided by the BLS.
   \item area_type Character. Description of the type of area based on the area_title. More consistent naming and grouping than BLS data.
   \item stfips Character. The two-digit FIPS code of the state containing a given area. Set to "00" for multi-state regions.
   \item specified_region. Either a two-character US Postal Service abbreviation for the state containing an area or a hyphenated list of such codes for multi-state areas.
  }
  
  For Annual files:
  \itemize{
   \item area_fips - Character. Area code of row.  Included `area_lookup` data file contains mapping information.
   \item industry_code - Character. NAICS, Supersector, Cluster, or Total All Industries code. Numeric characters as a string to preserve examining the structure heirarchy.
   \item own_code - Integer.  Values of 0-5 to designate ownership. See definitions at \url{https://www.bls.gov/cew/classifications/ownerships/ownership-titles.htm}
   \item agglvl_code - Integer.   Two digit code identifying the level of aggregation.  See definitons at \url{https://www.bls.gov/cew/classifications/aggregation/agg-level-titles.htm}
   \item size_code - Integer. Single-digit code representing the size of establishments. See definitions at \url{https://www.bls.gov/cew/classifications/size/size-titles.htm}
   \item year Integer.  Four-digit calendar year for the returned data.
   \item qtr Character. Set to "A" to represent annual data.
   \item disclosure_code Character.  Values are either a blank string on "N".  Values of N do not disclose employment or wages to maintain confidentiality.
   \item annual_avg_estabs Integer.  The average number of business establishments (worksites) for the industry in the area for the year.
   \item annual_avg_emplvl Integer. The average monthly employment level in a given year.
   \item total_annual_wages Ingeger64. Total wages paid during the year.
   \item taxable_annual_wages Ingeger64. Wages subject to unemployment insurance (UI) taxes during the year.  Note - wages subject to UI vary by state and will follow different seasonal patterns as a result.
   \item annual_contributions Integer. UI taxes (Contributions) paid by employers for this year. Note - UI tax policy varies by state.
   \item annual_avg_wkly_wage Integer. Average weekly wage during the year (Total wages divided by average employment, divided by 52).
   \item avg_annual_pay Integer. Average annual pay during the year.
  
   \item lq_disclosure_code Character.  Blank or "N".  Values of "N" will suppress location quotient data for confidentiality.
   \item lq_annual_avg_estabs Numeric. Location quotient of establishments relative to the U.S.
   \item lq_annual_avg_emplvl Numeric. Location quotient of annual employment relative to the U.S.
   \item lq_total_annual_wages Numeric. Location quotient of total wages relative to the U.S.
   \item lq_taxable_annual_wages Numeric. Location quotient of taxable annual wages relative to the U.S.
   \item lq_annual_contributions Numeric. Location quotient of annual UI taxes paid relative to the U.S.
   \item lq_annual_avg_wkly_wage Numeric. Location quotient of average weekly wages relative to the U.S.
   \item lq_avg_annual_pay Numeric. Location quotient of average annual pay relative to the U.S.
   
   \item oty_disclosure_code Character.  Blank or "N".  Values of "N" will suppress over-the-year data for confidentiality.
   \item oty_annual_avg_estabs_chg Integer. Over-the-year change in establishments.
   \item oty_annual_avg_estabs_pct_chg Numeric. Over-the-year percent change in establishments.
   \item oty_annual_avg_emplvl_chg Integer. Over-the-year change in average annual employment.
   \item oty_annual_avg_emplvl_pct_chg Numeric. Over-the-year percent change in average annual employment.
   \item oty_total_annual_wages_chg Integer. Over-the-year change in total wages.
   \item oty_total_annual_wages_pct_chg Numeric. Over-the-year percent change in total wages.
   \item oty_taxable_annual_wages_chg Integer. Over-the-year change in taxable annual wages.
   \item oty_taxable_annual_wages_pct_chg Numeric. Over-the-year percent change in taxable annual wages.
   \item oty_annual_contributions_chg Integer. Over-the-year change in annual UI taxes paid.
   \item oty_annual_contributions_pct_chg Numeric. Over-the-year percent change in annual UI taxes paid.
   \item oty_annual_avg_wkly_wage_chg Integer. Over-the-year change in average weekly wages.
   \item oty_annual_avg_wkly_wage_pct_chg Numeric. Over-the-year percent change in average weekly wages.
   \item oty_avg_annual_pay_chg Integer. Over-the-year change in average annual pay.
   \item oty_avg_annual_pay_pct_chg Numeric. Over-the-year percent change in average annual pay.
   
   \item date Date. Calculated calendar date based on year and quarter.  Reflects first day of the quarter.
   \item industry_title Character. Added based on industry_code
   \item ind_level Character. Description of the level of aggregation based on the industry_code.
   \item naics_2d Character. First two characters in the industry_code, useful for identifying industries.
   \item sector Character. Similar to naics_2d, but for industries like Manufacturing which have multiple two digit NAICS codes, this will span those groupings, for example "31-33"
   \item vintage_start. Integer. Calendar year of the earliest vintage for this industry_code. NAICS codes are updated every 5 years.  When using this industry codes from before this date, these titles may not exist or may be incorrect.
   \item vintage_end. Integer. Calendar year of the last year this industry code was used.  Years after this point should not contain this industry code.  Set to 3000 for current data.
   \item area_title Character.  Area description based on area_fips as provided by the BLS.
   \item area_type Character. Description of the type of area based on the area_title. More consistent naming and grouping than BLS data.
   \item stfips Character. The two-digit FIPS code of the state containing a given area. Set to "00" for multi-state regions.
   \item specified_region. Either a two-character US Postal Service abbreviation for the state containing an area or a hyphenated list of such codes for multi-state areas.
  }
}
\description{
This function pulls data from the BLS QCEW Open Data Access CSV Data Slices.
It iterates over specified years and quarters (or annual data) to retrieve
industry-specific or area-specific data tables and merges them into a single data.table.
Optionally, it joins internal package lookup tables for industry and area descriptions.
}
\examples{
\donttest{
# Get quarterly data for "Total, all industries" (Code 10)
# Includes industry/area descriptions and a date column by default
dt_default <- get_qcew(industry_code = "10")

# Get annual data for Nevada (Code 32000) for 2023 without lookups or messages
dt_year <- get_qcew(period_type = "year",
                    year_start = 2023,
                    year_end = 2023,
                    area_code = "32000",
                    add_lookups = FALSE,
                    silently = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{get_bls_excel_headers}
\alias{get_bls_excel_headers}
\title{Generate headers for BLS requests to download Excel files}
\usage{
get_bls_excel_headers(refer = "https://www.bls.gov/lau/stalt-archived.htm")
}
\arguments{
\item{refer}{The URL to use in the Referer header (default: "https://www.bls.gov/lau/stalt-archived.htm")}
}
\value{
A named character vector of HTTP headers
}
\description{
Returns a named character vector of HTTP headers required for BLS API requests.
These headers mimic a standard browser to ensure compatibility with BLS servers.
This function returns a more limited set of headers used to download an Ecel file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_oews.R
\name{get_oews}
\alias{get_oews}
\title{Download Occupational Employment and Wage Statistics (OEWS) Data}
\usage{
get_oews(
  simplify_table = TRUE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE,
  fast_read = TRUE,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{simplify_table}{Logical.  If TRUE (default), remove columns from the result that are internal BLS references or can be derived from other elements in the table.}

\item{suppress_warnings}{Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.}

\item{return_diagnostics}{Logical. If TRUE, returns a bls_data_collection object
with full diagnostics. If FALSE (default), returns just the data table.}

\item{fast_read}{Logical.  If TRUE (default), derives lookup values directly from series_id to avoid reading the series file, to speed download process. With fast_read, the data can download in 17 seconds (depending on bandwidth).  Without fast_read, the same download takes 57 seconds.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
By default, returns a data.table with OEWS data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing data and comprehensive diagnostics. The columns in the returned data frame when `simplify_table = TRUE` are listed below.  Unless otherwise specified, all data is returned as a character string to preserve the value of leading and trailing zeroes.
 \itemize{
    \item series_id - The unique OEWS series identifier.
    \item year - The year to which the estimate refers.  Because OEWS is not time series data, this is always the most recent year.
    \item value - Numeric. The value of the given data type, for the given area, in the given industry and occupation.
    \item seasonal - Whether or not the data is seasonally adjusted.
    \item areatype_code - Code representing the type of area (National ("N"), Statewide ("S"), or Local ("M")).
    \item industry_code - NAICS code of the industry.
    \item occupation_code - SOC code of the occupation. Description given by occupation_name.
    \item datatype_code - Lookup code for the data type of a given row.  Description given by datatype_name.
    \item state_code - Two-digit FIPS code for the state.
    \item area_code - The unique OEWS code for a substate area. Description given by area_name.
    \item series_title - Descriptive title of the full series ID.
    \item occupation_name - The text description of the occupation.
    \item occupation_description - More detailed description of the tasks associated with the occupation.
    \item area_name - The text description of the area.
    \item datatype_name - The text description of the type of data represented by `value`.
    
  }
}
\description{
This function downloads and joins together occupational employment and wage data
from the Bureau of Labor Statistics OEWS program. The data includes employment
and wage estimates by occupation and geographic area. Note that OEWS is a large data set (over 6 million rows), so it will require longer to download.
}
\examples{
\dontrun{
# Download current OEWS data
oews_data <- get_oews()

# View available occupations
unique(oews_data$occupation_name)

# Filter for specific occupation
software_devs <- oews_data[grepl("Software", occupation_name)]

# Get full diagnostic object if needed
oews_with_diagnostics <- get_oews(return_diagnostics = TRUE)
print_bls_warnings(oews_with_diagnostics)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{check_bls_cache_env}
\alias{check_bls_cache_env}
\title{Check if Global Caching is Enabled via Environment Variable}
\usage{
check_bls_cache_env()
}
\value{
Logical value indicating if the environment variable USE_BLS_CACHE is one of TRUE, YES, or 1
}
\description{
Check if Global Caching is Enabled via Environment Variable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bls_overview.R
\name{bls_overview}
\alias{bls_overview}
\title{Display BLS Dataset Overview}
\usage{
bls_overview(
  series_id,
  display_method = "viewer",
  base_url = "https://download.bls.gov/pub/time.series"
)
}
\arguments{
\item{series_id}{Character string. The BLS series identifier (e.g., "ln", "cu", "ap")}

\item{display_method}{Character string. How to display the overview: 
"viewer" (default), "console", or "popup"}

\item{base_url}{Character string. Base URL for BLS data (default uses official BLS site)}
}
\value{
Invisibly returns the text content. Function is called to use the viewer, console, or as a popup, depending on the 'display_method' argument.
}
\description{
Fetches and displays the overview text file for a BLS dataset. This provides a convenient reference within the R environment without needing to manually find and review the text file on the BLS website.
}
\examples{
\donttest{
if(interactive()){
# Display Average Price Data overview
bls_overview("ap")

# Display consumer price index overview  
bls_overview("cu")

# Display in console instead of viewer
bls_overview("ap", display_method = "console")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{has_bls_issues}
\alias{has_bls_issues}
\title{Check if BLS data object has potential issues with import.}
\usage{
has_bls_issues(bls_obj)
}
\arguments{
\item{bls_obj}{A BLS data object}
}
\value{
Logical indicating if there were any import issues detected.
}
\description{
Check if BLS data object has potential issues with import.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{get_bls_headers}
\alias{get_bls_headers}
\title{Generate headers for BLS requests}
\usage{
get_bls_headers(host = "download.bls.gov")
}
\arguments{
\item{host}{The URL to use in the Host header (default: "download.bls.gov")}
}
\value{
A named character vector of HTTP headers
}
\description{
Returns a named character vector of HTTP headers required for BLS API requests.
These headers mimic a standard browser to ensure compatibility with BLS servers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bls_overview.R
\name{display_in_console}
\alias{display_in_console}
\title{Display text content in console.}
\usage{
display_in_console(content, series_id)
}
\arguments{
\item{content}{Character. Text content to display}

\item{series_id}{Two-letter series ID for a BLS time series to render in the display.}
}
\value{
No object returned, called to render content in console only..
}
\description{
Helper function used to display content from 'bls_overview' in the console..
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/national_ces_helpers.R
\name{list_national_ces_options}
\alias{list_national_ces_options}
\title{List Available National CES Dataset Options}
\usage{
list_national_ces_options(show_descriptions = FALSE)
}
\arguments{
\item{show_descriptions}{Logical. If TRUE, shows detailed descriptions
of each dataset option. If FALSE (default), shows only the filter names.}
}
\value{
A data frame with dataset filter options and their descriptions.
}
\description{
This function displays the available dataset filtering options for the
get_national_ces() function, helping users understand what specialized
datasets are available for download.
}
\examples{
# Show available dataset filters
list_national_ces_options()

# Show detailed descriptions
list_national_ces_options(show_descriptions = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ind_lookup}
\alias{ind_lookup}
\title{NAICS Industry Titles Lookup Table (QCEW)}
\format{
A data frame with 2678 rows and 7 columns:
\itemize{
  \item{industry_code - Character.  The NAICS industry code.  Items starting with 10 are aggregated values, not corresponding to a unique 6-digit NAICS.}
  \item{industry_title - Character. The descripion of the industry code provided by the BLS.  This title also includes the code value for clarity.}
  \item{ind_level - Character. A description of the level of aggregation. Values are "Total", "Cluster", or "Supersector" for the "10" code aggregations, or else the length of the NAICS code, from 2-6 digits.}
  \item{naics_2d - Character. The first two digits of `industry_code`, which may be helpful to filter the results.}
  \item{sector - Character.  Similar to `naics_2d` except that when the industry sector spans multiple two digit codes, those codes are hyphenated (e.g. Manufacturing is NAICS 31, 32, and 33, so this displays '31-33').}
  \item{vintage_start - Integer.  The earliest year reviewed for NAICS code use. NAICS will change every 5 years, so data from before this year will have some missing values.}
  \item{vintage_end - Integer. The last year that a particular code is used, if applicable. Set to 3000 for current codes.}
}
}
\source{
\url{https://www.bls.gov/cew/classifications/industry/industry-titles.htm}
}
\usage{
data(ind_lookup)
}
\description{
A data frame containing the structure of the North American Industry Classification System (NAICS) titles and codes used by the U.S. Bureau of Labor Statistics (BLS), including in the Quarterly Census of Employment and Wages (QCEW) program. This data is current as of the 2022 NAICS revision.

This table is primarily used internally by `BLSloadR` functions to add human-readable industry titles or define valid industry lookups.
}
\details{
The NAICS structure is hierarchical. Codes are typically longer for more detailed industries.
}
\examples{
# Load the lookup table
data(ind_lookup)

# Find the industry title for NAICS 51 (Information)
ind_lookup[ind_lookup$industry_code == "51", ]

# Get the supersector codes
supersectors <- ind_lookup[ind_lookup$ind_level == "Supersector", ]

# Get all 3-digit NAICS codes in the Manufacturing industry
mfg_codes <- ind_lookup |> 
dplyr::filter(sector == "31-33" & ind_level == "NAICS 3-digit")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_bls_text.R
\name{read_bls_text}
\alias{read_bls_text}
\title{Read Plain Text Files from BLS Website}
\usage{
read_bls_text(url)
}
\arguments{
\item{url}{A character string specifying the full URL to a text file on the
BLS website (e.g., \url{https://download.bls.gov/pub/time.series/}).}
}
\value{
A character vector where each element is one line from the text file.
  Lines are split on newline characters (\code{\\n}).
}
\description{
Downloads and reads plain text files from the Bureau of Labor Statistics (BLS)
website. This is a companion function to \code{fread_bls()} that handles text
files rather than structured data tables. The function uses custom headers to
ensure reliable access to BLS resources.
}
\details{
This function is designed to read descriptive text files from BLS, such as
README files or database overview documents. It sends an HTTP GET request
with browser-like headers to ensure compatibility with BLS server requirements.

The function will stop with an error if the HTTP request fails (e.g., if the
URL is invalid or the server is unavailable).
}
\examples{
\donttest{
# Read the overview file for Current Employment Statistics
ces_overview <- read_bls_text(
  "https://download.bls.gov/pub/time.series/ce/ce.txt"
)

# Display the first few lines
head(ces_overview)
}

}
\seealso{
\code{\link{bls_overview}} for formatted database overviews,
\code{\link{load_bls_dataset}} for loading complete datasets
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{get_bls_data}
\alias{get_bls_data}
\title{Extract data from BLS data object}
\usage{
get_bls_data(bls_obj)
}
\arguments{
\item{bls_obj}{A bls_data_collection object or raw data}
}
\value{
The data component of a 'bls_data_collection' object as a data frame.
}
\description{
This is a helper function to extract the data element of a 'bls_data_collection' object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_bls_dataset.R
\name{load_bls_dataset}
\alias{load_bls_dataset}
\title{Generic BLS Dataset Download}
\usage{
load_bls_dataset(
  database_code,
  return_full = FALSE,
  simplify_table = TRUE,
  suppress_warnings = FALSE,
  which_data = NULL,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{database_code}{This is the two digit character identifier for the desired database.
Some Valid options are:
\itemize{
  \item "ce" - National Current Employment Statistics Data
  \item "sm" - State and Metro area Current Employment Statistics Data
  \item "mp" - Major Sector Total Factor Productivity
  \item "ci" - Employment Cost Index
  \item "eb" - Employee Benefits Survey
}}

\item{return_full}{This argument defaults to FALSE. If set to TRUE it will return
a list of the elements of data retrieved from the BLS separating the data, series, and
mapping values downloaded.}

\item{simplify_table}{This parameter defaults to TRUE. When TRUE it will remove all
columns from the date with "_code" in the column name, as well as a series of internal
identifiers which provide general information about the series but which are not needed for
performing time series analysis. This parameter also converts the column "value" to numeric
and generates a date column from the year and period columns in the data.}

\item{suppress_warnings}{Logical. If TRUE, suppress individual download warnings during processing.}

\item{which_data}{Character string or NULL. Defaults to NULL.
\itemize{
  \item "all" - Automatically selects the data file containing ".1.All" (e.g., "bd.data.1.AllItems" or "le.data.1.AllData").
  \item "current" - Automatically selects the data file containing "Current" (e.g., "ce.data.0.Current").
  \item NULL - Default behavior. Prompts the user to select a file if multiple exist, or selects the single available file.
}
If the requested pattern is not found, the function falls back to the default behavior, prompting the user to select a file.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
This function will return either a bls_data_collection object (if return_full is FALSE or not provided)
 or a named list of the returned data including the bls_data_collection object.
}
\description{
This function generalizes a method to download all BLS data for a given time series database.
These files are accessed from https://download.bls.gov/pub/time.series/ and several datasets
are available. A summary of an identified database can be generated using the `bls_overiew()`
function. When multiple potential data files exist (common in large data sets), the function
will prompt for an input of which file to use.
}
\examples{
\dontrun{
# Import All Data
fm_import <- load_bls_dataset("fm", which_data = "all")

# Get $data element
fm_data <- fm_import$data

# Filter to a Series
# Families with Children Under 6 and No Employed Parent

u6_no_emp <- fm_data |> 
  dplyr::filter(series_title == "Total families with children under 6 - with no parent employed") |> 
  dplyr:: select(year, value, fchld_text, fhlf_text, tdat_text)


head(u6_no_emp)
}

\dontrun{
# Examples requiring manual intervention in the console
# Download Employer Cost Index Data
cost_index <- load_bls_dataset("ci")

# Download separated data, series, and mapping columns
benefits <- load_bls_dataset("eb", return_full = TRUE)

# Download data without removing excess columns and value conversions
productivity <- load_bls_dataset("mp", simplify_table = FALSE)

# Check for download issues
if (has_bls_issues(cost_index)) {
  print_bls_warnings(cost_index, detailed = TRUE)
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_national_ces.R
\name{get_national_ces}
\alias{get_national_ces}
\title{Get National Current Employment Statistics (CES) Data from BLS}
\usage{
get_national_ces(
  dataset_filter = "all_data",
  monthly_only = TRUE,
  simplify_table = TRUE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{dataset_filter}{Character string specifying which dataset to download.
Options include:
\itemize{
  \item "all_data" (default) - Complete dataset with all series
  \item "current_seasonally_adjusted" - Only seasonally adjusted all-employee series
  \item "real_earnings_all_employees" - Real earnings data for all employees
  \item "real_earnings_production" - Real earnings data for production employees
}}

\item{monthly_only}{Logical. If TRUE (default), excludes annual averages
(period "M13") and returns only monthly data. If FALSE, includes all
periods including annual averages.}

\item{simplify_table}{Logical. If TRUE (default), removes several metadata
columns (series_title, begin_year, begin_period, end_year, end_period,
naics_code, publishing_status, display_level, selectable, sort_sequence)
and adds a formatted date column. If FALSE, returns the full dataset
with all available columns.}

\item{suppress_warnings}{Logical. If TRUE (default), suppresses download warnings
and diagnostics. If FALSE, displays warning output and diagnostic information.}

\item{return_diagnostics}{Logical. If TRUE, returns a bls_data_collection object
with full diagnostics. If FALSE (default), returns just the data table.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
By default, returns a data.table with CES data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing data and comprehensive diagnostics.
}
\description{
This function downloads and processes national Current Employment Statistics (CES)
data from the Bureau of Labor Statistics (BLS). It retrieves multiple related
datasets and joins them together to create a comprehensive employment statistics
dataset with industry classifications, data types, and time period information.
}
\details{
The function can download one of four specialized national CES datasets based on
the dataset_filter parameter:
\itemize{
  \item all_data: Complete dataset (ce.data.0.AllCESSeries) - contains entire
    history of all series currently published by the CES program
  \item current_seasonally_adjusted: (ce.data.01a.CurrentSeasAE) - contains
    every seasonally adjusted all employee series and complete history
  \item real_earnings_all_employees: (ce.data.02b.AllRealEarningsAE) - contains
    real earnings data (1982-84 dollars) for all employees
  \item real_earnings_production: (ce.data.03c.AllRealEarningsPE) - contains
    real earnings data (1982-84 dollars) for production/nonsupervisory employees
}

Additional metadata files are always downloaded and joined:
\itemize{
  \item ce.series - Series metadata
  \item ce.industry - Industry classifications
  \item ce.datatype - Data type definitions
  \item ce.period - Time period definitions
  \item ce.supersector - Supersector classifications
}

These datasets are joined together to provide context and labels for the
employment statistics. The function uses the enhanced `download_bls_files()`
helper function for robust downloads with diagnostic reporting.

Performance Note: Using specialized datasets (other than "all_data") can
significantly reduce download time and file size while still providing
comprehensive employment statistics.
}
\note{
This function requires the following packages: dplyr, data.table, httr, and
lubridate (for date formatting when simplify_table=TRUE). The `fread_bls()`
and `create_bls_object()` helper functions must be available in your environment.
}
\examples{
\dontrun{
# Get complete monthly CES data with simplified table structure (default)
ces_monthly <- get_national_ces()

# Get only seasonally adjusted data (faster download)
ces_seasonal <- get_national_ces(dataset_filter = "current_seasonally_adjusted")

# Get real earnings data for all employees
ces_real_earnings <- get_national_ces(dataset_filter = "real_earnings_all_employees")

# Get all data including annual averages with full metadata
ces_full <- get_national_ces(dataset_filter = "all_data",
                             monthly_only = FALSE, simplify_table = FALSE)

# Get data with warnings and diagnostic information displayed
ces_with_warnings <- get_national_ces(suppress_warnings = FALSE)

# Get full diagnostic object if needed
data_with_diagnostics <- get_national_ces(return_diagnostics = TRUE)
print_bls_warnings(data_with_diagnostics)
}


}
\seealso{
Please visit the Bureau of Labor Statistics at https://www.bls.gov/ces/ for more information about CES data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{bls_get_cache_dir}
\alias{bls_get_cache_dir}
\title{Get the current BLSloadR Cache Directory}
\usage{
bls_get_cache_dir()
}
\value{
A character string of the cache path.
}
\description{
Displays the path currently being used for caching. This will check the
`BLS_CACHE_DIR` environment variable, falling back to the default
system cache directory if the variable is not set.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{print_bls_warnings}
\alias{print_bls_warnings}
\title{Print warnings for BLS data object}
\usage{
print_bls_warnings(bls_obj, detailed = FALSE, silent = FALSE)
}
\arguments{
\item{bls_obj}{A bls_data_collection object}

\item{detailed}{Logical. If TRUE, shows detailed diagnostics for each file}

\item{silent}{Logical. If TRUE, suppress console output}
}
\value{
Character vector of warnings (invisibly)
}
\description{
Print warnings for BLS data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{create_bls_object}
\alias{create_bls_object}
\title{Create a BLS data object with diagnostics}
\usage{
create_bls_object(
  data,
  downloads,
  data_type = "BLS",
  processing_steps = character(0)
)
}
\arguments{
\item{data}{The processed data (data.table/data.frame)}

\item{downloads}{List of download results from fread_bls()}

\item{data_type}{Character string describing the type of BLS data (e.g., "CES", "JOLTS", "CPS")}

\item{processing_steps}{Character vector describing processing steps applied}
}
\value{
A bls_data_collection object
}
\description{
This is a helper function to create a list with the additional class 'bls_data_collection' containing data downloaded form the U.S. Bureau of Labor Statistics as well as diagnostic details about the download. It is used invisibly in the package to bundle information about file downloads.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_salt.R
\name{get_salt}
\alias{get_salt}
\title{Download State Alternative Labor Market Measures (SALT) Data}
\usage{
get_salt(
  only_states = TRUE,
  geometry = FALSE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE
)
}
\arguments{
\item{only_states}{Logical. If TRUE (default), includes only state-level data.
If FALSE, includes sub-state areas like New York City where available.}

\item{geometry}{Logical. If TRUE, uses tigris::states() to download shapefiles for the states 
to include in the data. If FALSE (default), only returns data table.}

\item{suppress_warnings}{Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.}

\item{return_diagnostics}{Logical. If TRUE, returns a bls_data_collection object
with full diagnostics. If FALSE (default), returns just the data table.}
}
\value{
By default, returns a data.table with Alternative Measures of Labor Underutilization data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing data and comprehensive diagnostics.
  The function also adds derived measures and quartile comparisons across states.
}
\description{
This function downloads detailed alternative unemployment measures data from BLS,
including U-1 through U-6 measures. The data provides a more comprehensive view
of labor market conditions beyond the standard unemployment rate (U-3).
}
\examples{
\dontrun{
# Download state-level SALT data
salt_data <- get_salt()

# Include sub-state areas
salt_all <- get_salt(only_states = FALSE)
 
# Download SALT with geometry included
get_salt(geometry = TRUE)

# Get full diagnostic object if needed
data_with_diagnostics <- get_salt(return_diagnostics = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ces.R
\name{get_ces}
\alias{get_ces}
\title{Download Current Employment Statistics (CES) Data}
\usage{
get_ces(
  states = NULL,
  industry_filter = NULL,
  current_year_only = FALSE,
  transform = TRUE,
  monthly_only = TRUE,
  simplify_table = TRUE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{states}{Character vector of state abbreviations to download (e.g., c("MA", "NY", "CA")).
If specified, downloads only these states (all industries, all years).
Cannot be combined with industry_filter or current_year_only.
Use `list_ces_states()` to see all available states.}

\item{industry_filter}{Character string specifying industry category to download.
If specified, downloads this industry for all states (2007-present).
Cannot be combined with states or current_year_only.
Use `list_ces_industries()` to see all available industry filters.}

\item{current_year_only}{Logical. If TRUE, downloads the current year file
which contains all states and industries for recent years (2006-present).
Cannot be combined with states or industry_filter. If FALSE (default), uses other parameters.}

\item{transform}{Logical. If TRUE (default), converts employment values from thousands
to actual counts by multiplying by 1000 for specific data types (codes 1, 6, 26)
and removes ", In Thousands" from data type labels.}

\item{monthly_only}{Logical. If TRUE (default), filters out annual data (period M13).}

\item{simplify_table}{Logical. If TRUE (default), removes excess columns and creates
a date column from Year and Period in the original data.}

\item{suppress_warnings}{Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.}

\item{return_diagnostics}{Logical. If FALSE (default), returns only the data. If TRUE,
returns the full bls_data_collection object with diagnostics.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
By default, returns a data.table with CES data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing data and comprehensive diagnostics.
}
\description{
This function downloads Current Employment Statistics data from the Bureau of Labor Statistics.
The data includes national, regional, state, and substate employment statistics.
By default, all available areas, data types, and periods are included.
}
\details{
**Performance Notes:** The default behavior downloads a very large file (~500MB+) containing
all states and industries, which can take several minutes. For faster downloads, consider:
\itemize{
  \item Use \code{states = c("MA", "NY")} to download only specific states
  \item Use \code{industry_filter = "total_nonfarm"} for summary employment data only
  \item Use \code{current_year_only = TRUE} for recent data only (2006-present)
}

**State Codes:** Use standard two-letter state abbreviations (e.g., "MA", "CA", "NY").
Puerto Rico = "PR", Virgin Islands = "VI", District of Columbia = "DC".

**Industry Filters:** Available options include:
\itemize{
  \item "total_nonfarm" - Total non-farm employment summary
  \item "total_private" - Private sector totals (2007-present)
  \item "manufacturing" - Manufacturing sector (2007-present)
  \item "construction" - Construction sector (2007-present)
  \item "retail_trade" - Retail trade sector (2007-present)
  \item "government" - Government sector (2007-present)
  \item And others - see BLS documentation for full list
}
}
\examples{
\dontrun{
# Fast download: Massachusetts and Connecticut data only (all industries)
ces_states <- get_ces(states = c("MA", "CT"))

# Fast download: Manufacturing data for all states
ces_manufacturing <- get_ces(industry_filter = "manufacturing")

# Fast download: Current year data for all states and industries
ces_current <- get_ces(current_year_only = TRUE)

# Complete dataset (slower - all states, industries, and years)
ces_all <- get_ces()

# Download with full diagnostics if needed
ces_result <- get_ces(states = "MA", return_diagnostics = TRUE)
ces_data <- get_bls_data(ces_result)

# Check for download issues
if (has_bls_issues(ces_result)) {
  print_bls_warnings(ces_result)
}
}
}
\seealso{
\code{\link{list_ces_states}()} to see available states,
\code{\link{list_ces_industries}()} to see available industry filters,
\code{\link{show_ces_options}()} for a comprehensive overview of filtering options.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{smart_bls_download}
\alias{smart_bls_download}
\title{Check and Download BLS File with Local Caching}
\usage{
smart_bls_download(url, cache_dir = NULL, verbose = FALSE)
}
\arguments{
\item{url}{A character string representing the URL of the BLS file (e.g., a `.txt` or `.gz` file from download.bls.gov).}

\item{cache_dir}{A character string specifying the local directory to store cached files. May also be set with the  enviroment variable `BLS_CACHE_DIR`
Defaults to a persistent user data directory managed by \code{tools::R_user_dir}.}

\item{verbose}{Logical. Defaults to FALSE.  If TRUE, returns status messages for download.}
}
\value{
A character string containing the local path to the downloaded (or cached) file.
}
\description{
This function manages the downloading of files from the BLS server with a 
local caching layer. It uses HTTP HEAD requests to compare the server's 
`Content-Length` and `Last-Modified` headers with local file attributes. 
The file is only downloaded if it does not exist locally, or if the remote 
version is newer or a different size.
}
\details{
The function uses a specific set of browser-like headers to ensure compatibility 
with BLS server security policies. Upon a successful download, the local file's 
modification time is synchronized with the server's `Last-Modified` header using 
\code{Sys.setFileTime} to ensure accurate future comparisons.
}
\examples{
\dontrun{
url <- "https://download.bls.gov/pub/time.series/ce/ce.data.0.AllCESSeries"
local_path <- smart_bls_download(url)
data <- data.table::fread(local_path)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ces_helpers.R
\name{show_ces_options}
\alias{show_ces_options}
\title{Show CES Data Filtering Options}
\usage{
show_ces_options()
}
\description{
Displays a comprehensive overview of all filtering options available
for the `get_ces()` function, including states, industries, and usage examples.
}
\examples{
# See all filtering options
show_ces_options()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_jolts.R
\name{get_jolts}
\alias{get_jolts}
\title{Download Job Openings and Labor Turnover Survey (JOLTS) Data}
\usage{
get_jolts(
  monthly_only = TRUE,
  remove_regions = TRUE,
  remove_national = TRUE,
  suppress_warnings = TRUE,
  return_diagnostics = FALSE,
  cache = check_bls_cache_env()
)
}
\arguments{
\item{monthly_only}{Logical. If TRUE (default), excludes annual data (period M13)
and includes only monthly observations.}

\item{remove_regions}{Logical. If TRUE (default), excludes regional aggregates
(Midwest, Northeast, South, West) identified by state codes MW, NE, SO, WE.}

\item{remove_national}{Logical. If TRUE (default), excludes national-level data
(state code 00). Set to FALSE to include national data with industry and
size class breakdowns.}

\item{suppress_warnings}{Logical. If TRUE (default), suppress individual download warnings and diagnostic messages
for cleaner output during batch processing. If FALSE, returns the data and prints warnings and messages to the console.}

\item{return_diagnostics}{Logical. If TRUE, returns a bls_data_collection object
with full diagnostics. If FALSE (default), returns just the data table.}

\item{cache}{Logical.  Uses USE_BLS_CACHE environment variable, or defaults to FALSE. If TRUE, will download a cached file from BLS server and update cache if BLS server indicates an updated file.}
}
\value{
By default, returns a data.table with JOLTS data. If return_diagnostics = TRUE,
  returns a bls_data_collection object containing JOLTS data with the following key columns:
  \describe{
    \item{series_id}{BLS series identifier}
    \item{year}{Year of observation}
    \item{period}{Time period (M01-M12 for months)}
    \item{value}{JOLTS statistic value (transformed based on data type)}
    \item{date}{Date of observation}
    \item{state_text}{State name}
    \item{dataelement_text}{Type of JOLTS measure (job openings, hires, separations, etc.)}
    \item{area_text}{Geographic area description}
    \item{sizeclass_text}{Establishment size class}
    \item{industry_text}{Industry classification}
    \item{ratelevel_code}{Whether the value is a "Level" (count) or "Rate" (percentage)}
    \item{periodname}{Month name}
  }
}
\description{
This function downloads Job Openings and Labor Turnover data from the U.S. Bureau
of Labor Statistics. JOLTS data provides insights into job market dynamics including
job openings, hires, separations, quits, and layoffs. Data is available at national,
regional, and state levels with various industry and size class breakdowns.
}
\details{
The function performs several data transformations:
  \itemize{
    \item Converts rate values to proportions (divides by 100) except for Unemployed to Job Opening ratio.
    \item Converts level values to actual counts (multiplies by 1000)
    \item Creates a proper date column from year and period
    \item Adds readable month names
  }
}
\examples{
\dontrun{
# Download state-level JOLTS data (default - returns data directly)
jolts_data <- get_jolts()

# Include national data with industry breakdowns
jolts_national <- get_jolts(remove_national = FALSE)

# Get full diagnostic object if needed
jolts_with_diagnostics <- get_jolts(return_diagnostics = TRUE)
print_bls_warnings(jolts_with_diagnostics)

# View job openings by state for latest period
job_openings <- jolts_data[dataelement_text == "Job openings" & 
                          date == max(date)]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{download_bls_files}
\alias{download_bls_files}
\title{Helper function for downloading and tracking BLS files}
\usage{
download_bls_files(urls, suppress_warnings = TRUE, cache = FALSE)
}
\arguments{
\item{urls}{Named character vector of URLs to download}

\item{suppress_warnings}{Logical. If TRUE, suppress individual download warnings}

\item{cache}{Logical. If TRUE, download and cache local copy of files.}
}
\value{
Named list of bls_data objects
}
\description{
This function is used to pass multiple URLs at the Bureau of Labor Statistics into 'fread_bls()'
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fread_BLS.R
\name{fread_bls}
\alias{fread_bls}
\title{Download BLS Time Series Data}
\usage{
fread_bls(url, verbose = FALSE, cache = check_bls_cache_env())
}
\arguments{
\item{url}{Character string. URL to the BLS flat file}

\item{verbose}{Logical. If TRUE, prints additional messages during file read and processing.}

\item{cache}{Logical. If TRUE, uses local persistent caching.}
}
\value{
A named list with the data and diagnostics.
}
\description{
This function downloads a tab-delimited BLS flat file, incorporating 
diagnostic information about the file and returning an object with the
bls_data class that can be used in the BLSloadR package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_oews.R
\name{get_oews_areas}
\alias{get_oews_areas}
\title{Download OEWS Area Definitions}
\usage{
get_oews_areas(ref_year, silent = TRUE, geometry = TRUE)
}
\arguments{
\item{ref_year}{Four-digit year (converted to integer). The year for which to retrieve OEWS area definitions. Valid values are 2024 through current release year. Prior years included Township codes, which change the structure of the file.}

\item{silent}{Logical. If TRUE (default), suppress console output}

\item{geometry}{Logical.  If TRUE (default), downloads shapefiles for OEWS area definitions using `tigris::counties()` and `tigris::shift_geometry()` to render Alaska, Hawaii, and Puerto Rico with a focus on the area of the continental United States.}
}
\value{
Data table which maps individual counties to OEWS area definitions.
  \itemize{
    \item fips_code - The State FIPS code
    \item state_name - The state name
    \item state_abb - The state two-character postal abbreviation
    \item oews_area_code - The OEWS area code defining the metropolitan area or nonmetropolitan area the county belongs to.
    \item oews_area_name - The OEWS area name
    \item county_code - The FIPS code for the county
    \item county_name - The county name
    }
}
\description{
Download OEWS Area Definitions
}
\examples{
\dontrun{
 # Get OEWS area definitions without shapefiles and with processing messages.
 test <- get_oews_areas(ref_year = 2024, geometry = FALSE, silent = FALSE)
 
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_helpers.R
\name{get_bls_diagnostics}
\alias{get_bls_diagnostics}
\title{Get download diagnostics from BLS data object}
\usage{
get_bls_diagnostics(bls_obj)
}
\arguments{
\item{bls_obj}{A bls_data_collection object}
}
\value{
List of download diagnostics from a bls_data_collection object.
}
\description{
This is a helper function to extract the download diagnostics element of a 'bls_data_collection' object.
}
